#
# This TCL code draws pictures of the SNAP focal plane.
# 
# MWR  8/12/2003
#    


if { [catch { set debug } ] == 1 } {
  set debug 2
}

source param.tcl


###########################################################################
# Read information from the high-level and low-level focal plane descriptions,
#   and draw a picture which we export as Postscript
#
# RETURNS
#    0           if all goes well
#    1           if there's a problem
#    
proc draw_focal { focal_param } {
  global debug

  if { $debug > 0 } {
    puts "entering draw_focal"
  }

  # get information from the high-level param file
  if { [file exists $focal_param] != 1 } {
    puts stderr "draw_focal: can't open param file $focal_param"
    return 1
  }
  set output_dir [get_param $focal_param "output_dir"]
  set output_file [get_param $focal_param "output_file"]
  set param_dir [get_param $focal_param "param_dir"]
  set foc_output_file [get_param $focal_param "foc_output_file"]
  set foc_annulus_inner [get_param $focal_param "foc_annulus_inner"]
  set foc_annulus_outer [get_param $focal_param "foc_annulus_outer"]
  set foc_label_chip    [get_param $focal_param "foc_label_chip"]
  set foc_draw_filters  [get_param $focal_param "foc_draw_filters"]
  set foc_draw_magnify  [get_param $focal_param "foc_draw_magnify"]


  # verify that graphics are enabled.  If not, print an error 
  #   message and continue ...
  if { [is_graphics_enabled] == 0 } {
    puts stderr "draw_focal: graphics are not enabled, so skip them"
    return 0
  }

  # now get information on the chips from the low-level param file
  set low_param [format "%s/%s" $param_dir $foc_output_file]
  set num_chips [get_param $low_param "num_chips"]
  set i_chip 0
  while { $i_chip < $num_chips } {
    set param_name [format "chip_%03d" $i_chip]
    set chip_info($i_chip) [get_param $low_param $param_name]
    if { $debug > 1 } {
      puts " draw_focal: chip_info( $i_chip ) is ..$chip_info($i_chip).."
    }
    incr i_chip
  }


  # time to make the picture
  if { [draw_pieces $foc_annulus_inner $foc_annulus_outer \
                       $foc_label_chip $foc_draw_filters $foc_draw_magnify \
                       $num_chips chip_info ] != 0 } {
    puts stderr "draw_focal: draw_pieces fails"
    return 1
  }


  # and save output in a Postscript file
  if { [export_canvas] != 0 } {
    puts stderr "draw_focal: export_canvas fails"
    return 1
  }


  return 0
}



#############################################################################
# PROCEDURE: draw_pieces
#
# DESCRIPTION: Make a drawing of the focal plane, including the 
#              annuli and each chip.
#
#              Args:  
#                     inner_ring                radius of inner ring (mm)
#                     outer_ring                radius of outer ring (mm)
#                     label_chips               if 1, label chips with ID
#                     draw_filters              if 1, draw filters, too
#                     magnify                   enlargement factor in the
#                                                   drawing
#                     num_chips                 number of chips on plane
#                     chip_info                 array, one per chip,
#                                                   each elem is a list 
#
#              (must eventually pass a file name for Postscript output file)
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc draw_pieces { inner_ring outer_ring label_chips draw_filters magnify \
                                                    num_chips chip_info } {
  global debug
  upvar $chip_info c_info

  if { $debug > 0 } {
    puts "entering draw_pieces"
  }

  # make a canvas big enough to hold the outer ring, with room to spare
  set canvas_size [expr int(ceil($outer_ring*2.2))]
  set canvas_size [expr $canvas_size*$magnify]
  init_canvas $canvas_size $magnify

  if { $debug > 2 } {
    puts "  draw_pieces: inner_ring $inner_ring"
  }
  if { [draw_circle $inner_ring 0 0] != 0 } {
    puts stderr "draw_pieces: draw_circle fails for inner ring"
    return 1
  }
  if { $debug > 2 } {
    puts "  draw_pieces: outer_ring $inner_ring"
  }
  if { [draw_circle $outer_ring 0 0] != 0 } {
    puts stderr "draw_pieces: draw_circle fails for outer ring"
    return 1
  }


if { 1 } {
  set i_chip 0
  while { $i_chip < $num_chips } {
    set ll $c_info($i_chip)
    set clist [lindex $ll 0]
    if { $debug > 2 } {
      puts "   draw_pieces: info $i_chip ..$clist.. "
    }
   
    # what color should the chip be?
    set type [lindex $clist 1]
    if { [string compare $type "vis"] == 0 } {
      set color "green"
    } else {
      if { [string compare $type "ir"] == 0 } {
        set color "red"
      } else {
        puts stderr "draw_pieces: invalid chip type ..$type.. "
        return 1
      }
    }
      
    # get the four corners of this chip
    foreach corner_index { 0 1 2 3 } {
      set cx($corner_index) [lindex $clist [expr ($corner_index*2)+7] ]
      set cy($corner_index) [lindex $clist [expr ($corner_index*2)+8] ]
    }

    # draw this chip
    if { [draw_rectangle $cx(0) $cy(0) $cx(1) $cy(1) $cx(2) $cy(2) \
                                      $cx(3) $cy(3) $color] != 0 } {
      puts stderr "draw_pieces: draw_rectangle fails for i_chip $i_chip"
      return 1
    }

    # if desired, draw the filters on top of this chip
    if { $draw_filters == 1 } {
      if { [get_filters_for_chip $i_chip flist] != 0 } {
        puts stderr "draw_pieces: get_filters_for_chip fails"
      }
      if { $debug > 2 } {
        puts "draw_pieces: i_chip $i_chip flist is $flist"
      }
      foreach filter_pair $flist {
        set cl [get_corners $clist $filter_pair]
        set color [get_filter_color [lindex $filter_pair 1]]
        if { [draw_rectangle [lindex $cl 0] [lindex $cl 1] \
                             [lindex $cl 2] [lindex $cl 3] \
                             [lindex $cl 4] [lindex $cl 5] \
                             [lindex $cl 6] [lindex $cl 7] \
                                      $color $color] != 0 } {
          puts stderr "draw_pieces: draw_rectangle fails for i_chip $i_chip"
          return 1
        }
      }
    }

    # if desired, add a label near chip center
    if { $label_chips == 1 } {
      set center_x [lindex $clist 5]
      set center_y [lindex $clist 6]
      set label_text $i_chip
      if { [draw_text $center_x $center_y $label_text] != 0 } {
        puts stderr "draw_pieces: draw_text fails for i_chip $i_chip"
        return 1
      }
    }


    update
        
      
    incr i_chip
  }
}
 

  # this flushes all the graphical output to the canvas ...
  update

  return 0
}


#############################################################################
# PROCEDURE: init_canvas
#
# DESCRIPTION: Create a square canvas (with name ".can") of the given size.
#              Pack it, so that the window appears.
#
#              We destroy the canvas first, just in case it already exists.
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc init_canvas { size magnify } {
  global debug
  
  if { $debug > 0 } {
    puts "entering init_canvas"
  }

  set canvas_name ".can"
  if { [set_canvas_info $canvas_name $size $magnify ] != 0 } {
    puts stderr "init_canvas: set_canvas_info fails"
  }
  destroy $canvas_name
  canvas $canvas_name -height $size -width $size 
  pack $canvas_name

  return 0
}

#############################################################################
# PROCEDURE: draw_text 
#
# DESCRIPTION: Draw text at the given (x, y) position
# 
#              Args:
#                       x               xpos of text start (mm on focal plane)
#                       y               ypos ditto
#                       text_label      a string with text to draw
#                       color           color of text (default black)
#                       font            name of font to use (default {})
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc draw_text { x y text_label { color black } { font {} } } {
  global debug
  
  if { $debug > 0 } {
    puts "entering draw_text"
  }

  # get canvas info
  if { [get_canvas_info c_name c_size c_magnify] != 0 } {
    puts stderr "draw_text: get_canvas_info fails"
  }
  
  # transform the coordinates to canvas coords
  if { [focplane_to_canvas $x $y c_x c_y] != 0 } {
    puts stderr "draw_text: focplane_to_canvas fails on x, y"
    return 1
  }

  # now draw the text on the canvas
  set cmd "$c_name create text $c_x $c_y -text \"$text_label\" -fill $color"
  if { $font != {} } {
    set cmd [format "%s -text %s " $cmd $font]
  }
  eval $cmd

  return 0
}



#############################################################################
# PROCEDURE: draw_circle 
#
# DESCRIPTION: Draw a circle of the given radius at the given (x, y) position
#              in the focal plane (so x and y are in mm from optical center)
#              Required args are
#
#                      radius  xcenter  ycenter
#
#              If a fourth arg is given, use it as the "fill" parameter.
#              By default, circles are not filled
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc draw_circle { radius xcenter ycenter { fill_color {} } } {
  global debug
  
  if { $debug > 0 } {
    puts "entering draw_circle"
  }

  # get canvas info
  if { [get_canvas_info c_name c_size c_magnify] != 0 } {
    puts stderr "draw_circle: get_canvas_info fails"
  }
  
  # calculate the bounding box of this circle, in focal plane coords
  set x1 [expr $xcenter-$radius]
  set y1 [expr $ycenter-$radius]
  set x2 [expr $xcenter+$radius]
  set y2 [expr $ycenter+$radius]

  # transform the coordinates to canvas coords
  if { [focplane_to_canvas $x1 $y1 c_x1 c_y1] != 0 } {
    puts stderr "draw_circle: focplane_to_canvas fails on x1, y1"
    return 1
  }
  if { [focplane_to_canvas $x2 $y2 c_x2 c_y2] != 0 } {
    puts stderr "draw_circle: focplane_to_canvas fails on x2, y2"
    return 1
  }

  # now draw the oval on the canvas
  if { $fill_color == {} } {
    $c_name create oval $c_x1 $c_y1 $c_x2 $c_y2 
  } else {
    $c_name create oval $c_x1 $c_y1 $c_x2 $c_y2 -fill $fill_color
  }

  return 0
}


#############################################################################
# PROCEDURE: export_canvas
#
# DESCRIPTION: Export the canvas to a Postscript file.  We get the name
#              of the file from a variable which was initialized via
#              the "init_draw_focal" proc.
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc export_canvas { } {
  global debug
  
  if { $debug > 0 } {
    puts "entering export_canvas"
  }

  # flush all graphical output first
  update

  # get the output file name
  if { [get_canvas_export_name export_name] != 0 } {
    puts stderr "export_canvas: get_canvas_export_name fails"
    return 1
  }

  # figure out how big the postscript file will be
  #   For our purposes, a rule that the size is 7.5 inches times
  #   the magnification factor is good enough
  get_canvas_info c_name c_size c_magnify
  set height [expr $c_magnify*7.5]
  set height_str [format "%.1fi" $height]
  set width [expr $c_magnify*7.5]
  set width_str [format "%.1fi" $width]
  
  # now write to the given file name
  .can postscript -file "$export_name" -height $height_str -width $width_str

  return 0
}




#############################################################################
# PROCEDURE: draw_rectangle 
#
# DESCRIPTION: Draw a rectangle, given as arguments FOCAL PLANE coordinates
#                   x1 y1 
#                   x2 y2 
#                   x3 y3 
#                   x4 y4 
#              There are optional args following these for line color
#              and fill style
#
#              We convert focal plane coords to canvas coords before drawing.
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc draw_rectangle { x1 y1 x2 y2 x3 y3 x4 y4 { outline black } { fill {} } } {
  global debug
  
  if { $debug > 0 } {
    puts "entering draw_rectangle"
  }

  # get canvas info
  if { [get_canvas_info c_name c_size c_magnify] != 0 } {
    puts stderr "draw_circle: get_canvas_info fails"
  }
  
  # convert coords to canvas coord system
  if { [focplane_to_canvas $x1 $y1 c_x1 c_y1] != 0 } {
    puts stderr "draw_rectangle: focplane to canvas fails on x1, y1"
    return 1
  }
  if { [focplane_to_canvas $x2 $y2 c_x2 c_y2] != 0 } {
    puts stderr "draw_rectangle: focplane to canvas fails on x2, y2"
    return 1
  }
  if { [focplane_to_canvas $x3 $y3 c_x3 c_y3] != 0 } {
    puts stderr "draw_rectangle: focplane to canvas fails on x3, y3"
    return 1
  }
  if { [focplane_to_canvas $x4 $y4 c_x4 c_y4] != 0 } {
    puts stderr "draw_rectangle: focplane to canvas fails on x4, y4"
    return 1
  }

  # now draw the oval on the canvas
  $c_name create rectangle $c_x1 $c_y1 $c_x3 $c_y3 \
                                -outline $outline -fill $fill  

  return 0
}


#############################################################################
# PROCEDURE: draw_star_radec 
#
# DESCRIPTION: Draw a symbol for a star at the given (RA, Dec) coords
#              and with the given magnitude
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc draw_star { ra dec mag } {
  global debug
  
  if { $debug > 0 } {
    puts "entering draw_star"
  }

  # get canvas info
  if { [get_canvas_info c_name c_size c_magnify] != 0 } {
    puts stderr "draw_star: get_canvas_info fails"
  }
  
  # transform the coordinates to focal plane (x, y)
  if { [radec_to_focxy $ra $dec x y] != 0 } {
    puts stderr "draw_star: radec_to_focxy fails on $ra $dec"
    return 1
  }
  if { $debug > 0 } {
    puts "draw_star: x $x   y $y "
  }

  # Now, we draw stars as circles.  We start with a very small filled circle
  #   for a faint star, then increase the circle size up to some limit M1.
  #   For stars brighter than M1, we use the same succession of filled circles,
  #   but surround the filled circle with an empty circle.  That goes 
  #   up to magnitude M2.  
  #   For stars brighter than M2, we return to the succession of filled
  #   circles, but now surround with _two_ empty circles.  And so forth.
  #
  # The "circle_list" arg returned by this routine will have a list of
  #   circle radii.  The first one is drawn filled, the others empty.
  if { [mag_to_circles $mag circle_list] != 0 } {
    puts stderr "draw_star: mag_to_circles fails on $mag "
    return 1
  }

  # now draw the symbols on the canvas
  if { $debug > 0 } {
    puts "draw_star: about to draw_star_symbol $x $y $circle_list "
  }
  if { [draw_star_symbol $x $y $circle_list] != 0 } {
    puts stderr "draw_star: draw_star_symbol fails"
    return 1
  }

  # update the canvas?
  update

  return 0
}


#############################################################################
# PROCEDURE: mag_to_circles
#
# DESCRIPTION: Given a magnitude, figure out the way we will draw it.
#              The basic idea is to start with a filled circle which
#              grows larger with brightness, then switch to a succession
#              of the same symbols, but surrounded by empty circles,
#              for brighter magnitude ranges.
#
#              Example:   mags  25 - 20      filled circle
#                               20 - 15      filled circle w/ 1 ring
#                               15 - 10      filled circle w/ 2 rings
#                          etc.
#
#              Place into the output arg "o_circle_list" a list of 
#              circle radii.  The first one is always filled, the others
#              are always just rings.
#
# RETURNS
#     0        if all goes well
#     1        if an error occurs
#
proc mag_to_circles { mag o_circle_list } {
  global debug
  upvar $o_circle_list circle_list

  if { $debug > 0 } {
    puts "entering mag_to_circles"
  }

  if { [get_draw_star_info min_mag cycle_size max_filled_radius \
                         ring_increment ] != 0 } {
    puts stderr "mag_to_circles: get_draw_star_info fails"
    return 1
  }

  if { $mag > $min_mag } {
    set circle_list [list 0]
    if { $debug > 0 } {
      puts "   mag $mag fainter than limit $min_mag, draw zero"
    }
  } else {

    # figure out how many rings we'll need
    set num_rings [expr int(($min_mag-$mag)/$cycle_size)]
    set left_over [expr ($min_mag - $mag) - ($num_rings*$cycle_size)]
    if { $debug > 0 } {
      puts "   num_rings $num_rings   left_over $left_over "
    }
   
    # how big should the filled circle be?  It depends on the 
    #   left-over amount of magnitude
    set radius [expr ($left_over/$cycle_size)*$max_filled_radius]
    set circle_list [list $radius]
    if { $debug > 0 } {
      puts "   filled circle radius $radius"
    }

    # and now, we add as many rings as needed
    for { set i_ring 0 } { $i_ring < $num_rings } { incr i_ring } {
      set radius [expr $max_filled_radius + ($i_ring+1)*$ring_increment]
      if { $debug > 0 } {
        puts "   add ring radius $radius"
      }
      lappend circle_list [list $radius]
    }

  }

  if { $debug > 0 } {
    puts "mag_to_circles: final list is ..$circle_list.."
  }

  return 0
}


#############################################################################
# PROCEDURE: draw_star_symbol
#
# DESCRIPTION: Given an (x,y) position on the focal plane, and given a list
#              which consists of circle radii, we 
#
#                 a) draw a filled circle at the position with radius
#                        given by the first element of list
#
#                 b) draw an empty circle at the position with radius
#                        given by the second element of list
#
#              and repeat step b) for all remaining elements of the list.
#
# RETURNS
#     0        if all goes well
#     1        if an error occurs
#
proc draw_star_symbol { x y circle_list } {
  global debug

  if { $debug > 0 } {
    puts "entering draw_star_symbol"
  }

  set llen [llength $circle_list]

  # if the list is empty, draw nothing
  if { $llen == 0 } {
    return 0
  }

  # if the list has a single element of size 0, draw nothing 
  if { ($llen == 1) && ([lindex $circle_list 0] == 0) } {
    return 0
  }
 
  # draw a filled circle with radius given by the first element
  set radius [lindex $circle_list 0]
  set fill_color black
  if { $debug > 0 } {
    puts "draw_star_symbol: draw_circle $radius $x $y $fill_color"
  }
  if { [draw_circle $radius $x $y $fill_color] != 0 } {  
    puts stderr "draw_star_symbol: draw_circle fails for $x $y $radius $fill_color"
    return 1
  }

  # and draw a ring for each of the remaining elements
  for { set i_elem 1 } { $i_elem < $llen } { incr i_elem } {
    set radius [lindex $circle_list $i_elem]
    if { $debug > 0 } {
      puts "draw_star_symbol: draw_circle $radius $x $y "
    }
    if { [draw_circle $radius $x $y ] != 0 } {  
      puts stderr "draw_star_symbol: draw_circle fails for $x $y $radius"
      return 1
    }
  }

  return 0
}



#############################################################################
# PROCEDURE: focplane_to_canvas
#
# DESCRIPTION: Given coordinates (x, y) in mm in the focal plane,
#              convert to the corresponding canvas coordinates.
#              Place those values (in pixels) into the two output args.
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc focplane_to_canvas { x y canvas_x canvas_y } {
  global debug
  upvar $canvas_x cx $canvas_y cy

  if { $debug > 0 } {
    puts "entering focplane_to_canvas"
  }

  # get information about the size of the canvas
  if { [get_canvas_info c_name c_size c_magnify] != 0 } {
    puts stderr "focplane_to_canvas: get_canvas_info fails"
    return 1
  }

  # scale the position (in mm) to position (in pixels)
  set x [expr $x*$c_magnify]
  set y [expr $y*$c_magnify]

  # first, we invert the y coordinate, since the canvas has increasing
  #   y values going DOWNWARD on the screen (or page)
  set y [expr 0.0-$y]

  # next, we move the origin to the center of the canvas
  set half [expr $c_size/2]
  set cx [expr $x+$half]
  set cy [expr $y+$half]


  # all done
  return 0
}




#############################################################################
# PROCEDURE: get_filters_for_chip
#
# DESCRIPTION: Given a chip ID number, return a list of keyed lists
#              which give the quadrant and filter index for all filters
#              on that chip.
#
#              For example, 
#                  { 1 2 }  { 2 4 }  { 3 3 }  { 4 1 }
#              means that this is a chip with 4 filters on it.  In quadrant
#              1 is filter with index 2, in quadrant 2 is filter with 
#              index 4, etc.
#
#              On the other hand,
#                  { 0 7 }
#              means that there is only one filter, index 7, over the entire
#              chip.
#
#
#              Place the list into the output arg "o_filter_list"
#
# RETURNS:
#     0         if all goes well
#     1         if an error occurs
#
proc get_filters_for_chip { chip_id o_filter_list } {
  global debug
  upvar $o_filter_list filter_list

  if { $debug > 0 } {
    puts "entering get_filters_for_chip"
  }
  
  set filter_list {}

  set filter_assn [get_filter_assn]
  set num_filt [get_param $filter_assn "num_filt"]
  set filt_id 0
  while { $filt_id < $num_filt } {
    set name [format "filt_%03d" $filt_id]
    set ll [get_param $filter_assn $name]
    if { $ll == {} } {
      puts stderr "get_filters_for_chip: can't read info for filter $name"
      exit 1
    }
    set flist [lindex $ll 0]
    if { $debug > 5 } {
      puts "    name $name flist $flist"
    }

    set cid [lindex $flist 2]
    if { $cid == $chip_id } {
      set quadrant [lindex $flist 3]
      set filter_index [lindex $flist 4]
      if { $filter_list == {} } { 
        set filter_list [list [list $quadrant $filter_index]]
      } else {
        lappend filter_list [list $quadrant $filter_index]
      }
    }

    incr filt_id
  }

  return 0
}



############################################################################
# PROCEDURE: get_filter_color
#
# DESCRIPTION: Given a filter ID, return a color for drawing it
#
# RETURNS:
#   a Tk color index
#
proc get_filter_color { filter_index } {
  global debug

  if { $debug > 0 } {
    puts "entering get_filter_color"
  }

  set namelist [list violet blue lightblue green yellow orange coral red brown]

  if { ($filter_index < 0) || ($filter_index > [llength $namelist]) } {
    puts stderr "get_filter_color: invalid index $filter_index"
    exit 1
  }
  
  return [lindex $namelist $filter_index]
}


############################################################################
# PROCEDURE: get_corners
#
# DESCRIPTION: Given a chip's entry from the chip database,
#              and given a list which looks like this:
#                       { quadrant filter_index}
#              figure out the corners of the quadrant for the chip.
#              Quadrants are, in the chip's own coord system,
#                     0 = entire chip
#                     1 = upper left
#                     2 = upper right
#                     3 = lower left
#                     4 = lower right
#
# RETURNS:
#   a list of 8 values, the corners of the quadrant
#
proc get_corners { clist qlist } {
  global debug

  if { $debug > 0 } {
    puts "entering get_corners"
  }

  if { $debug > 5 } {
    puts "get_corners: $clist .. $qlist"
  }
  set quadrant [lindex $qlist 0]

  # we assume that the rotation is an integral multiple of 90 degrees
  set angle [lindex $clist 4]
  if { ($angle != 0) && ($angle != 90) && ($angle != 180) && ($angle != 270) } {
    puts stderr "get_corners: can't handle angle $angle in $clist"
    exit 1
  }

  # find the min, max X,Y values of the corners
  set minx [lindex $clist 7]
  if { [lindex $clist 9] < $minx } {
    set minx [lindex $clist 9]
  }
  if { [lindex $clist 11] < $minx } {
    set minx [lindex $clist 11]
  }
  set maxx [lindex $clist 7]
  if { [lindex $clist 9] > $maxx } {
    set maxx [lindex $clist 9]
  }
  if { [lindex $clist 11] > $maxx } {
    set maxx [lindex $clist 11]
  }

  set miny [lindex $clist 8]
  if { [lindex $clist 10] < $miny } {
    set miny [lindex $clist 10]
  }
  if { [lindex $clist 12] < $miny } {
    set miny [lindex $clist 12]
  }
  set maxy [lindex $clist 8]
  if { [lindex $clist 10] > $maxy } {
    set maxy [lindex $clist 10]
  }
  if { [lindex $clist 12] > $maxy } {
    set maxy [lindex $clist 12]
  }


  # find the half-way points
  set halfx [expr ($maxx+$minx)/2.0]
  set halfy [expr ($maxy+$miny)/2.0]

  # these are the four possibilities for quadrant corner lists
  set qa [list $minx $halfy $halfx $halfy $halfx $maxy $minx $maxy]
  set qb [list $halfx $halfy $maxx $halfy $maxx $maxy $halfx $maxy]
  set qc [list $minx $miny $halfx $miny $halfx $halfy $minx $halfy]
  set qd [list $halfx $miny $maxx $miny $maxx $halfy $halfx $halfy]

  # and this is the corner list for the entire chip
  set qq [list [lindex $clist 7] [lindex $clist 8] \
               [lindex $clist 9] [lindex $clist 10] \
               [lindex $clist 11] [lindex $clist 12] \
               [lindex $clist 13] [lindex $clist 14]]

  if { $angle == 0 } {
    if { $quadrant == 0 } {
      set corner_list $qq
    }
    if { $quadrant == 1 } {
      set corner_list $qa
    }
    if { $quadrant == 2 } {
      set corner_list $qb
    } 
    if { $quadrant == 3 } {
      set corner_list $qc
    }
    if { $quadrant == 4 } {
      set corner_list $qd
    }
  }

  if { $angle == 90 } {
    if { $quadrant == 0 } {
      set corner_list $qq
    }
    if { $quadrant == 1 } {
      set corner_list $qc
    }
    if { $quadrant == 2 } {
      set corner_list $qa
    } 
    if { $quadrant == 3 } {
      set corner_list $qd
    }
    if { $quadrant == 4 } {
      set corner_list $qb
    }
  }
  
  if { $angle == 180 } {
    if { $quadrant == 0 } {
      set corner_list $qq
    }
    if { $quadrant == 1 } {
      set corner_list $qd
    }
    if { $quadrant == 2 } {
      set corner_list $qc
    } 
    if { $quadrant == 3 } {
      set corner_list $qb
    }
    if { $quadrant == 4 } {
      set corner_list $qa
    }
  }

  if { $angle == 270 } {
    if { $quadrant == 0 } {
      set corner_list $qq
    }
    if { $quadrant == 1 } {
      set corner_list $qb
    }
    if { $quadrant == 2 } {
      set corner_list $qd
    } 
    if { $quadrant == 3 } {
      set corner_list $qa
    }
    if { $quadrant == 4 } {
      set corner_list $qc
    }
  }
  
  return $corner_list
}




namespace eval Draw_Focal {
  namespace export init_draw_focal \
                         set_canvas_info get_canvas_info get_canvas_magnify \
                         get_canvas_export_name get_draw_star_info \
                         is_graphics_enabled 
  variable Canvas_Name {}
  variable Canvas_Size {}
  variable Canvas_Magnify {}
  variable Canvas_Export_Name {}
  variable Draw_Star_Min_Mag {}
  variable Draw_Star_Cycle_Size {}
  variable Draw_Star_Max_Filled {}
  variable Draw_Star_Ring_Increment {}


#######################################################################
# Initialize some drawing quantities that we'll read once and never again
#
# The "param_file" used here is the "focal.param" file; and yes,
#   there is a separate "init_focal" routine which sets other
#   private quantities elsewhere
#
proc init_draw_focal { param_file } {
  global debug
  variable Canvas_Export_Name
  variable Draw_Star_Min_Mag {}
  variable Draw_Star_Cycle_Size {}
  variable Draw_Star_Max_Filled {}
  variable Draw_Star_Ring_Increment {}

  if { $debug > 0 } {
    puts "entering init_draw_focal"
  }

  set output_dir [get_param $param_file "output_dir"]
  set name       [get_param $param_file "output_file"]
  set Canvas_Export_Name [format "%s/%s" $output_dir $name]

  # these parameters all control the drawing of star symbols
  set Draw_Star_Min_Mag  [get_param $param_file "foc_draw_min_mag"]
  set Draw_Star_Cycle_Size  [get_param $param_file "foc_draw_cycle_size"]
  set Draw_Star_Max_Filled  [get_param $param_file "foc_draw_max_filled"]
  set Draw_Star_Ring_Increment  [get_param $param_file "foc_draw_ring_increment"]

  return 0
}


#######################################################################
# Set the name of the canvas, and its size
#
proc set_canvas_info { name size magnification } {
  global debug
  variable Canvas_Name
  variable Canvas_Size
  variable Canvas_Magnify {}

  if { $debug > 0 } {
    puts "entering set_canvas_info "
  }
  set Canvas_Name $name
  set Canvas_Size $size
  set Canvas_Magnify $magnification

  return 0
}

######################################################################
# Return the name and size of the canvas in the given output args
#
proc get_canvas_info { canvas_name canvas_size canvas_magnify } {
  global debug
  upvar $canvas_name name $canvas_size size $canvas_magnify magnify
  variable Canvas_Name
  variable Canvas_Size
  variable Canvas_Magnify

  if { $debug > 0 } {
    puts "entering get_canvas_info"
  }
  set name $Canvas_Name
  set size $Canvas_Size
  set magnify $Canvas_Magnify

  return 0
}

######################################################################
# Place the name of the postscript file into which we'll export
#   the canvas into the output arg.
#
# RETURNS
#   0        if all goes well
#   1        if an error occurs (name not set)
#
proc get_canvas_export_name { o_export_name } {
  global debug
  upvar $o_export_name export_name
  variable Canvas_Export_Name

  if { $debug > 0 } {
    puts "entering get_canvas_export_name"
  }
  if { $Canvas_Export_Name == {} } {
    puts stderr "get_canvas_export_name: name not set yet ?!"
    return 1
  }

  set export_name $Canvas_Export_Name
  return 0
}

######################################################################
# Place a number of parameters which control the drawing of stars
#   into the given output args
#
proc get_draw_star_info { o_min_mag o_cycle_size o_max_filled o_ring_increment } {
  global debug
  upvar $o_min_mag min_mag $o_cycle_size cycle_size
  upvar $o_max_filled max_filled $o_ring_increment ring_increment
  variable Draw_Star_Min_Mag
  variable Draw_Star_Cycle_Size
  variable Draw_Star_Max_Filled
  variable Draw_Star_Ring_Increment

  if { $debug > 0 } {
    puts "entering get_draw_star_info"
  }

  set min_mag $Draw_Star_Min_Mag
  set cycle_size $Draw_Star_Cycle_Size
  set max_filled $Draw_Star_Max_Filled
  set ring_increment $Draw_Star_Ring_Increment

  return 0
}


######################################################################
# Figure out if we can draw graphics (i.e. if we are running under "wish").
#   If so, return 1.  If we can't draw graphics, return 0.
#
# We try to call the "tk" procedure.  If we can, graphics are okay.
#   If we can't, then graphics aren't possible.
#
proc is_graphics_enabled { } {
  global debug

  if { $debug > 0 } {
    puts "entering is_graphics_enabled"
  }
  
  if { [catch { tk scaling }] == 1 } {
    if { $debug > 0 } {
      puts "is_graphics_enabled: graphics are not possible"
    }
    return 0
  } else {
    return 1
  }

}




  # end of namespace Draw_Focal
}


