#
# This TCL script works with the filters on the SNAP focal plane.
#
# MWR 8/15/2003  
#    
# Now the code will convolve input filter transmission curves with
#   data on reflectivity of mirrors and QE of detectors, so that
#   the transmission files in the output_dir include these effects.
#   They become overall passbands, not just filter transmissions.
#   MWR 5/13/2004
#   


if { [catch { set debug } ] == 1 } {
  set debug 2
}

source param.tcl


###########################################################################
#
# RETURNS
#    0           if all goes well
#    1           if there's a problem
#    
proc filter { param_file } {
  global debug

  if { $debug > 0 } {
    puts "entering filter"
  }

  # get information from the param file
  if { [file exists $param_file] != 1 } {
    puts stderr "filter: can't open param file $param_file"
    return 1
  }
  set input_dir [get_param $param_file "input_dir"]
  set output_dir [get_param $param_file "output_dir"]
  set num_fiducial [get_param $param_file "num_fiducial"]
  set fiducial_base [get_param $param_file "fiducial_base"]

  # this stuff for modifying transmission curves
  set reflectivity_file [get_param $param_file "reflectivity_file"]
  set num_reflections [get_param $param_file "num_reflections"]
  set optical_qe_file [get_param $param_file "optical_qe_file"]
  set ir_qe_file [get_param $param_file "ir_qe_file"]
  set convolve_exe [get_param $param_file "convolve_exe"]


  set filter_assn [get_filter_assn]
  if { $filter_assn == {} } {
    puts stderr "filter: can't get filter_assn "
    return 1
  }

  # figure out which filter(s) go on top of each chip
  if { [assign_filters $filter_assn] != 0 } {
    puts stderr "filter: assign_filters returns with error"
    return 1
  }

  # copy the appropriate fiducial filter transmission curves to 
  #   a separate data file, one per actual filter
  if { [copy_filters_with_shift $input_dir $output_dir \
                                $filter_assn $reflectivity_file \
                                $num_reflections \
                                $optical_qe_file $ir_qe_file \
                                $convolve_exe \
                                $output_dir] != 0 } {
    puts stderr "filter: copy_filters_with_shift returns with error"
    return 1
  }

  # this is where we WOULD modify some or all of the filters
  #   from their fiducial values



  return 0
}



#############################################################################
# PROCEDURE: assign_filters
#
# DESCRIPTION: We decide which filters will be placed over which 
#              chips (or sub-sections of chips).  The assignments
#              will be written into a file in the output directory,
#              for future reference.
#
#              Args:
#                      filter_assn            name of the data file into
#                                                which we'll write the
#                                                assignment info
#
#              The format of the assignment file is like so:
#
#                  id  name  chip  quadrant  fid  virgin  transfile
#
#              where
#                     id             is a unique numerical ID for each
#                                         actual filter in the focal plane
#
#                     name           is a combination of chip number and
#                                         quadrant code, like this:
#                                             008A, 023B, 055
#                                         for full-chip filters, there is
#                                         no quadrant code
#
#                    chip            is ID number of chip
#
#                    quadrant        is  0   for a full-chip filter
#                                        1   for top-left quadrant (A)
#                                        2   for top-left quadrant (B)
#                                        3   for top-left quadrant (C)
#                                        4   for top-left quadrant (D)
#                     
#                    fid             is ID of fiducial filter type
#                                        should range 0-8
#
#                    virgin          is 1 if the actual filter is identical
#                                            to the fiducial
#                                       0 if not
#
#                    transfile       is name of file with actual transmission
#                                         curve for this filter 
#                                         (will be in output directory)
#
#
# RETURNS
#     0           if all goes well
#     1           if an error occurs
#              
proc assign_filters { assnfile } {
  global debug
  
  if { $debug > 0 } {
    puts "entering assign_filters"
  }

  if { $debug > 4 } {
    puts "assign_filters: assnfile is ..$assnfile.." 
  }

  # create the file into which we'll write the information
  if { [catch { set fout [open $assnfile "w"] } ] != 0 } {
    puts stderr "assign_filters: can't open file $assnfile for writing"
    return 1
  }

  # we start the filter ID at 0, and increment it each time we create
  #   a new filter
  set filt_id 0

  # assign filters in East block, visible section
  if { [assign_a_block $fout 0 "vis" filt_id] != 0 } {
    puts stderr "assign_filters: assign_a_block fails for base 0"
  }
  # assign filters in East block, IR section
  if { [assign_a_block $fout 9 "ir" filt_id] != 0 } {
    puts stderr "assign_filters: assign_a_block fails for base 9"
  }
  
  # if we are putting detectors only in the East Quadrant (to save time),
  #   then don't bother assigning filters to chips in the other quadrants.
  if { [get_foc_east_only] == 0 } {
    # assign filters in North block, visible section
    if { [assign_a_block $fout 18 "vis" filt_id] != 0 } {
      puts stderr "assign_filters: assign_a_block fails for base 18"
    }
    # assign filters in North block, IR section
    if { [assign_a_block $fout 27 "ir" filt_id] != 0 } {
      puts stderr "assign_filters: assign_a_block fails for base 27"
    }
    # assign filters in West block, visible section
    if { [assign_a_block $fout 36 "vis" filt_id] != 0 } {
      puts stderr "assign_filters: assign_a_block fails for base 36"
    }
    # assign filters in West block, IR section
    if { [assign_a_block $fout 45 "ir" filt_id] != 0 } {
      puts stderr "assign_filters: assign_a_block fails for base 45"
    }
    # assign filters in South block, visible section
    if { [assign_a_block $fout 54 "vis" filt_id] != 0 } {
      puts stderr "assign_filters: assign_a_block fails for base 54"
    }
    # assign filters in South block, IR section
    if { [assign_a_block $fout 63 "ir" filt_id] != 0 } {
      puts stderr "assign_filters: assign_a_block fails for base 63"
    }
  }

  # and now, at the end, we know how many different filters there are
  puts $fout [format "num_filt %d" $filt_id]

  close $fout

  return 0
}


##############################################################################
# PROCEDURE: assign_a_block
#
# DESCRIPTION: assign a block's worth of filters.
#
#              Args:
#                    fout                        file ID into which we write
#                    first_ccd_id 
#                    type                        "vis" or "IR"
#                    next_ID                
#
#              The "next_ID" variable is passed by reference.  We
#              modify its value so that on exit, it will be ready
#              for the next block
#
# RETURNS:
#     0        if all goes well
#     1        if there's a problem
#
proc assign_a_block { fout first_ccd_id type o_next_id } {
  global debug
  upvar $o_next_id next_id

  if { $debug > 0 } {
    puts "entering assign_a_block"
  }

  if { $debug > 5 } {
    puts "assign_a_block: first_ccd_id $first_ccd_id type $type next_id $next_id"
  }

  set chip_base 0
  set block_id 0
  set chipfile [get_chipfile_name]
  if { $debug > 4 } {
    puts "assign_a_block: chipfile is ..$chipfile.. "
  }
  # sanity checks
  if { [file exists $chipfile] != 1 } {
    puts stderr "assign_a_block: chipfile $chipfile doesn't exist?"
    return 1
  }

  set quadrant_code_list [list X A B C D]

  # these tell us how to arrange the visible filter on the CCD chips ...
  #   (note 4 filters per chip)
  #   these work for chip rotation angle 0 and 180 (South and North)
  set quadrant_values(0) [list 4 5 5 3]
  set quadrant_values(1) [list 2 0 3 4]
  set quadrant_values(2) [list 0 1 1 2]
  set quadrant_values(3) [list 3 1 4 2]
  set quadrant_values(4) [list 1 5 5 0]
  set quadrant_values(5) [list 2 3 0 4]
  set quadrant_values(6) [list 2 0 0 1]
  set quadrant_values(7) [list 3 4 1 2]
  set quadrant_values(8) [list 4 5 5 3]
  #   and these work for chip rotation angle 90 and 270 (West and East)
  set quadrant_values_r(0) [list 5 3 4 5]
  set quadrant_values_r(1) [list 0 4 2 3]
  set quadrant_values_r(2) [list 1 2 0 1]
  set quadrant_values_r(3) [list 1 2 3 4]
  set quadrant_values_r(4) [list 5 0 1 5]
  set quadrant_values_r(5) [list 3 4 2 0]
  set quadrant_values_r(6) [list 0 1 2 0]
  set quadrant_values_r(7) [list 4 2 3 1]
  set quadrant_values_r(8) [list 5 3 4 5]

  # and these tell us how to arrange the IR filters on the HgCdTe chips ...
  #   (only 1 filter per chip)
  set fullchip_values(0) 8
  set fullchip_values(1) 7
  set fullchip_values(2) 6
  set fullchip_values(3) 6
  set fullchip_values(4) 8
  set fullchip_values(5) 7
  set fullchip_values(6) 7
  set fullchip_values(7) 6
  set fullchip_values(8) 8

  while { $block_id < 9 } {

    set chip_id [expr $first_ccd_id+$block_id]
    set chip_name [format "chip_%03d" $chip_id]
    set ll [get_param $chipfile $chip_name]
    set clist [lindex $ll 0]
    
    if { $debug > 5 } {
      puts stdout [format "chip_id %d clist $clist " $chip_id]
    }

    # sanity checks
    if { [lindex $clist 0] != $chip_id } {
      puts stderr [format "assign_a_block: chip %d has ID %d " \
                          $chip_id [lindex $clist 0] ]
      return 1
    }

    set chip_angle [lindex $clist 4]
    if { [string compare $type "vis"] == 0 } { 
      foreach quadrant { 1 2 3 4 } {
        set qcode [lindex $quadrant_code_list $quadrant]
        if { ($chip_angle == 0) || ($chip_angle == 180) } {
          set filt [lindex $quadrant_values($block_id) [expr $quadrant-1]]
        } else {
          set filt [lindex $quadrant_values_r($block_id) [expr $quadrant-1]]
        }
        set name [format "%03d%s" $chip_id $qcode]
        set virgin 1
        set transfile [format "filt_%s.dat" $name]

        puts $fout [format "filt_%03d { %3d %5s %3d %3d %3d %3d %s } " \
                 $next_id $next_id $name $chip_id $quadrant \
                 $filt $virgin $transfile]
        incr next_id
      }
    } else {
      set quadrant 0
      set filt $fullchip_values($block_id) 
      set name [format "%03d" $chip_id ]
      set virgin 1
      set transfile [format "filt_%s.dat" $name]

      puts $fout [format "filt_%03d { %3d %5s %3d %3d %3d %3d %s } " \
               $next_id $next_id $name $chip_id $quadrant \
               $filt $virgin $transfile]
      incr next_id
    }

    incr block_id
  }

  return 0
}



#############################################################################
# PROCEDURE: copy_filters
#
# DESCRIPTION: For each actual filter, create a transmission curve file 
#              in the output_dir.
#              We just copy the transmission curves of the fiducial filters;
#              a later procedure may modify them.
#
#              Args are:
#                  assn_file             name of param file with data
#                                             for each filter
#
#                  output_dir            directory into which we'll
#                                             place data for each filter
#
# RETURNS
#     0           if all goes well
#     1           if an error occurs
#              
proc copy_filters { assnfile output_dir } {
  global debug
  
  if { $debug > 0 } {
    puts "entering copy_filters"
  }

  if { [file exists $assnfile] == 0 } {
    puts stderr "copy_filters: can't access file $assnfile"
    return 1
  }
  set num_filt   [get_param $assnfile "num_filt"]

  set i_filt 0
  while { $i_filt < $num_filt } {

    set name [format "filt_%03d" $i_filt]
    set ll [get_param $assnfile $name]
    if { $ll == {} } {
      puts stderr "copy_filters: can't get info for filter $name"
    }
    set flist [lindex $ll 0]
    
    # this is the fiducial filter index number
    set fiducial_index [lindex $flist 4]
    
    # this is the name of the fiducial transmission curve file
    set fiducial_name [get_filter_fiducial_name $fiducial_index]

    # and this is the file into which we make the copy
    set copy_name [lindex $flist 6]
    set copy_name [format "%s/%s" $output_dir $copy_name]

    # make the copy
    set cmd "cp $fiducial_name $copy_name"
    if { $debug > 0 } {
      puts "copy_filters: about to exec ..$cmd.. "
    }
    if { [catch { set result [eval exec $cmd] } ] == 1 } {
      puts stderr "copy_filters: cmd ..$cmd.. fails ?!"
      return 1
    }

    incr i_filt
  }

  return 0
}


#############################################################################
# PROCEDURE: copy_filters_with_shift
#
# DESCRIPTION: For each actual filter, create a transmission curve file 
#              in the output_dir.
#              We modify the transmission curve of each filter,
#              shifting it by a fraction "fractional_shift" to larger
#              wavelengths.  Thus, a positive value shifts filter
#              to the red, a negative value shifts filter to the blue.
#
#              Args are:
#                  input_dir             directory where we can find input
#                                             files 
#
#                  output_dir            directory where temp and output
#                                             files go
#
#                  assn_file             name of param file with data
#                                             for each filter
#      
#                  reflectivity_file     name of a file with reflectivity
#                                             vs. lambda for mirrors
# 
#                  num_reflections       number of times light encounters 
#                                             a mirror in the optical path
#
#                  optical_qe_file       name of a file with QE vs. lambda
#                                             of optical detectors
#
#                  ir_qe_file            name of a file with QE vs. lambda
#                                             of IR detectors
#
#                  output_dir            directory into which we'll
#                                             place data for each filter
#
#
# RETURNS
#     0           if all goes well
#     1           if an error occurs
#              
proc copy_filters_with_shift { input_dir output_dir \
                                 assnfile reflectivity_file num_reflections \
                                 optical_qe_file ir_qe_file \
                                 convolve_exe output_dir } {
  global debug
  
  if { $debug > 0 } {
    puts "entering copy_filters_with_shift"
  }

  # this controls the size of the shift in wavelength 
  #    (it is the fractional shift)
  set fractional_shift  0.00

  if { [file exists $assnfile] == 0 } {
    puts stderr "copy_filters_with_shift: can't access file $assnfile"
    return 1
  }
  set num_filt   [get_param $assnfile "num_filt"]

  set i_filt 0
  while { $i_filt < $num_filt } {

    set name [format "filt_%03d" $i_filt]
    set ll [get_param $assnfile $name]
    if { $ll == {} } {
      puts stderr "copy_filters_with_shift: can't get info for filter $name"
    }
    set flist [lindex $ll 0]
    
    # this is the fiducial filter index number
    set fiducial_index [lindex $flist 4]
    
    # this is the name of the fiducial transmission curve file
    set fiducial_name [get_filter_fiducial_name $fiducial_index]

    # and this is the file into which we make the copy
    set copy_name [lindex $flist 6]
    set copy_name [format "%s/%s" $output_dir $copy_name]

    # make the copy
    set orig_fid [open $fiducial_name "r"]
    set copy_fid [open $copy_name "w"]
    if { ($orig_fid < 0) || ($copy_fid < 0) } {
      puts stderr "copy_filters_with_shift: fids are $orig_fid $copy_fid ?!"
      return 1
    }
    while { [gets $orig_fid line] != -1 } {
      set len [llength $line]
      if { $len == 2 } {
        set orig_lambda [lindex $line 0]
        set trans [lindex $line 1]
        set new_lambda [expr $orig_lambda*(1.0 + $fractional_shift)]
        puts $copy_fid "$new_lambda $trans"
      }
    }
    close $copy_fid
    close $orig_fid

    # and this is a good place to convolve the filter transmission 
    #   with additional spectral effects: reflectivity of the mirror(s)
    #   and QE of the appropriate detector.
    #
    # first, we figure out if this is an optical (CCD) or near-IR (other)
    #   detector, so that we know which data file holds the QE
    set detector_type [get_detector_type [lindex $flist 1]]
    # now, we modify the transmission file by including effects of
    #   reflectivity and detector QE
    if { [create_full_passband $input_dir $output_dir \
                               $copy_name $detector_type \
                               $reflectivity_file $num_reflections \
                               $optical_qe_file $ir_qe_file \
                               $convolve_exe ] != 0 } {
      puts stderr "copy_filters_with_shift: create_full_passband fails ?!"
      return 1
    }


    incr i_filt
  }

  return 0
}


########################################################################
# PROCEDURE: create_full_passband 
#
# DESCRIPTION: Given arguments
#
#                 input_dir            directory for input files 
#
#                 output_dir           directory where output goes (need for
#                                             a temporary file)
#
#                 filter_name          (on input) filter(only) transmission 
#                                                       file name
#                                      its contents will be overwritten with
#                                             a convolution of all factors
#
#                 detector_type        either "optical" or "ir"
#
#                 reflectivity_file    file with reflectivity as func
#                                             of wavelength
#
#                 num_reflections      number of bounces off mirror
#
#                 optical_qe_file      QE of optical detectors vs. wavelength
#
#                 ir_qe_file           QE of IR detectors vs. wavelength
#
#                 convolve_exe         name of executable program which 
#                                             computes convolution of 
#                                             several data files
#
#              we build a command line which will call the convolving program
#              on all the appropriate factors, yielding an overall passband
#              as a function of wavelength.
#
#              We will replace the contents of file "filter_name" with 
#              the convolved output.
#                   
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc create_full_passband { input_dir output_dir filter_name detector_type \
                            reflectivity_file num_reflections \
                            optical_qe_file ir_qe_file \
                            convolve_exe } {
  global debug
 
  if { $debug > 0 } {
    puts "entering create_full_passband "
  }


  # we are going to build a (shell) command line, then execute it
  #   to call an external program to do some work for us.
  #   We'll need a temporary file, which we'll place in the output directory
  set temp_name "QQcreate_full_passband.dat"
  set temp_file [format "%s/%s" $output_dir $temp_name]
  
  # we start with the executable program
  set cmd_line [format "%s " $convolve_exe]

  # then comes the filter, which has already been placed into the output_dir
  set cmd_line [format "%s %s " $cmd_line $filter_name]

  # next, the detector; again, the file is in the input_dir
  if { [string compare $detector_type "optical"] == 0 } {
    set cmd_line [format "%s %s/%s " $cmd_line $input_dir $optical_qe_file ]
  }
  if { [string compare $detector_type "ir"] == 0 } {
    set cmd_line [format "%s %s/%s " $cmd_line $input_dir $ir_qe_file ]
  }

  # we add one copy of the mirror reflectivity for each bounce
  for { set i 0 } { $i < $num_reflections } { incr i } {
    set cmd_line [format "%s %s/%s " $cmd_line $input_dir $reflectivity_file]
  }

  # finally, we specify that the output should go into our temp file
  set cmd_line [format "%s -o %s " $cmd_line $temp_file]


  if { $debug > 2 } {
    puts "create_full_passband: cmd_line is ..$cmd_line.. "
  }

  # run the command to create the full passband in the temp file 
  if { [catch { set result [eval exec $cmd_line ] } ] != 0 } {
    puts stderr "create_full_passband: eval fails for ..$cmd_line.."
    return 1
  }

  # now, we must overwrite the given filter file in the output_dir
  #   with this new, convolved data
  set cmd_line [format "mv $temp_file $filter_name"]
  if { $debug > 2 } {
    puts "create_full_passband: cmd_line is ..$cmd_line.. "
  }
  # run the command to move the temp file into the filter file
  if { [catch { set result [eval exec $cmd_line ] } ] != 0 } {
    puts stderr "create_full_passband: eval fails for ..$cmd_line.."
    return 1
  }

  return 0
}



########################################################################
# PROCEDURE: get_detector_type 
#
# DESCRIPTION: Given the name of a detector, like "001A" or "064",
#              figure out if it is an optical CCD or a near-IR device.
#              We use the file name: 
#
#                   3-char name, digits only:    064        means    IR
#                   4-char name, letter at end:  063B       means  optical
#
#              Return a string describing the device.  If not a recognized
#              type, halt with error message.
#
# RETURNS:
#      "ccd"       if optical CCD
#      "ir"        if near-IR detector
#
proc get_detector_type { filter_name } {
  global debug
 
  if { $debug > 0 } {
    puts "entering get_detector_type "
  }
  if { $debug > 2 } {
    puts "get_detector_type: filter_name is $filter_name "
  }

  # check first to see if this is a set of 3 digits
  if { [string length $filter_name] == 3 } {
    # this looks like an IR device ....
    set type "ir"
    if { [string is digit [string range $filter_name 0 0]] != 1 } {
      set type "error"
    }
    if { [string is digit [string range $filter_name 1 1]] != 1 } {
      set type "error"
    }
    if { [string is digit [string range $filter_name 2 2]] != 1 } {
      set type "error"
    }

    if { $type == "error" } {
      puts stderr "get_detector_type: invalid ir filter name ..$filter_name.."
      exit 1
    }
    # okay, it really is a valid near IR name 
    if { $debug > 2 } {
      puts "get_detector_type: type is $type "
    }
    return $type
  } 

  # check next to see if this is a set of 3 digits plus a letter
  if { [string length $filter_name] == 4 } {
    # this looks like an optical device
    set type "optical"
    if { [string is digit [string range $filter_name 0 0]] != 1 } {
      set type "error"
    }
    if { [string is digit [string range $filter_name 1 1]] != 1 } {
      set type "error"
    }
    if { [string is digit [string range $filter_name 2 2]] != 1 } {
      set type "error"
    }
    if { [string is alpha [string range $filter_name 3 3]] != 1 } {
      set type "error"
    }

    if { $type == "error" } {
      puts stderr "get_detector_type: invalid optical filter name ..$filter_name.."
      exit 1
    }
    # okay, it really is a valid near optical name 
    if { $debug > 2 } {
      puts "get_detector_type: type is $type "
    }
    return $type
  } 
  
  # we should never get here!  It must be an invalid name
  puts stderr "get_detector_type: invalid filter name ..$filter_name.."
  exit 1

}




########################################################################
# PROCEDURE: chip_to_filter 
#
# DESCRIPTION: Given a coordinate (chip, row, col), figure out which
#              filter covers it.  Place the filter index and the
#              name of the filter file (which will be in the output_dir)
#              into the two output args.
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc chip_to_filter { chip row col o_filter_index o_filter_file } {
  global debug
  upvar $o_filter_index filter_index $o_filter_file filter_file
 
  if { $debug > 0 } {
    puts "entering chip_to_filter "
  }

  set filter_index -1
  set filter_file {}

  # get a list of properties for all filters on the chip
  #   i.e. a list of filter-property lists
  if { [get_filters_on_chip $chip filter_list] != 0 } {
    puts stderr "chip_to_filter: get_filters_on_chip fails"
    return 1
  }

  # figure out which of the filters is the one at the given (row, col)
  #   There are two possibilities: 1 filter on the entire chip
  #                                4 filters, one per quadrant
  #   We can distinguish them by the number of elements in the 
  #   "filter_list" returned by get_filters_on_chip"
  set len [llength $filter_list]
  if { $len == 1 } {
    # only 1 filter on the entire chip
    set fl [lindex $filter_list 0]

  } elseif { $len == 4 } {
    # 4 filters, one per quadrant.  So figure out which quadrant we're in
    set chipfile_name [get_chipfile_name]
    set cname [format "chip_%03d" $chip]
    set chip_info [get_param $chipfile_name $cname]
    set c_list [lindex $chip_info 0]
    set npix [lindex $c_list 2]
    set half [expr $npix/2]
    if { ($row < $half) && ($col < $half) } {
      # bottom-left quadrant, aka quadrant 3, aka quadrant C
      set fl [lindex $filter_list 2]

    } elseif { ($row < $half) && ($col > $half) } {
      # bottom-right quadrant, aka quadrant 4, aka quadrant D
      set fl [lindex $filter_list 3]
 
    } elseif { ($row > $half) && ($col < $half) } {
      # upper-left quadrant, aka quadrant 1, aka quadrant A
      set fl [lindex $filter_list 0]

    } else {
      # upper-right quadrant, aka quadrant 2, aka quadrant B
      set fl [lindex $filter_list 1]
    }

  } else {
    # invalid number of filters on this chip?!
    puts stderr "chip_to_filter: invalid filter_list length $len ?!"
    puts stderr "   filter_list is ..$filter_list.. "
    return 1
  }
    
  set filter_index [lindex $fl 0]
  set filter_file [lindex $fl 6]

  return 0
}




#############################################################################
# PROCEDURE: get_filters_on_chip
#
# DESCRIPTION: Given a chip index number, find all the filters which
#              are connected with the chip.  For each filter, grab
#              the list of properties from the filter assignment file.
#              Place a list of all those lists into the output arg.
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc get_filters_on_chip { chip_index o_filter_list } {
  global debug
  upvar $o_filter_list filter_list

  if { $debug > 0 } {
    puts "entering get_filters_on_chip"
  }
 
  set filter_list {}

  set filter_assn [get_filter_assn]  
  set num_filt [get_param $filter_assn "num_filt"]
  for { set i_filt 0 } { $i_filt < $num_filt } { incr i_filt } {
    set name [format "filt_%03d" $i_filt]
    set ll [get_param $filter_assn $name]
    if { $ll == {} } {
      puts stderr "get_filters_on_chip: get_param $filter_assn $name fails"
      return 1
    }
    set chip_info [lindex $ll 0]

    set this_chip [lindex $chip_info 2]
    if { $this_chip == $chip_index } {
      if { $filter_list == {} } {
        set filter_list [list $chip_info]
      } else {
        lappend filter_list $chip_info
      }
    }

  }

  return 0
}


########################################################################
# PROCEDURE: focxy_to_filter
#
# DESCRIPTION: Given a coordinate (x, y) on the focal plane, figure out which
#              filter covers it.  Place the filter index and the
#              name of the filter file (which will be in the output_dir)
#              into the two output args.
#
#              If the given position doesn't lie within one of the detectors,
#              then set filter_index = -1 and filter_file = {}
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc focxy_to_filter { x y o_filter_index o_filter_file } {
  global debug
  upvar $o_filter_index filter_index $o_filter_file filter_file
 
  if { $debug > 0 } {
    puts "entering focxy_to_filter "
  }

  # prepare for situation that position is outside any chip
  set filter_index -1
  set filter_file {}

  # convert from (x,y) to (chip,row,col)
  if { [focxy_to_chip $x $y chip row col] != 0 } {
    puts stderr "focxy_to_filter: focxy_to_chip fails $x $y "
    return 1
  }
  if { $debug > 5 } {
    puts "focxy_to_chip returns ..$chip.. ..$row.. ..$col.. "
  }
  if { $chip == -1 } {
    # doesn't fall within a chip -- so immediately return
    #   as we have set the output args already to proper values
    return 0
  }

  # now get the filter from the chip coords
  if { [chip_to_filter $chip $row $col filter_index filter_file] != 0 } {
    puts stderr "focxy_to_filter: chip_to_filter fails for $chip $row $col"
    return 1
  }

  return 0
}


########################################################################
# PROCEDURE: radec_to_filter
#
# DESCRIPTION: Given a coordinate (RA, Dec), figure out which
#              filter covers it.  Place the filter index and the
#              name of the filter file (which will be in the output_dir)
#              into the two output args.
#
#              If the given position doesn't lie within one of the detectors,
#              then set filter_index = -1 and filter_file = {}
#
# RETURNS:
#    0         if all goes well
#    1         if an error occurs
#
proc radec_to_filter { ra dec o_filter_index o_filter_file } {
  global debug
  upvar $o_filter_index filter_index $o_filter_file filter_file
 
  if { $debug > 0 } {
    puts "entering radec_to_filter "
  }

  # prepare for situation that position is outside any chip
  set filter_index -1
  set filter_file {}

  # convert from (RA, Dec) to (chip,row,col)
  if { [radec_to_chip $ra $dec chip row col] != 0 } {
    puts stderr "radec_to_filter: radec_to_chip fails $ra $dec "
    return 1
  }
  if { $debug > 5 } {
    puts "radec_to_chip returns ..$chip.. ..$row.. ..$col.. "
  }
  if { $chip == -1 } {
    # doesn't fall within a chip -- so immediately return
    #   as we have set the output args already to proper values
    return 0
  }

  # now get the filter from the chip coords
  if { [chip_to_filter $chip $row $col filter_index filter_file] != 0 } {
    puts stderr "radec_to_filter: chip_to_filter fails for $chip $row $col"
    return 1
  }

  return 0
}

#######################################################################
# PROCEDURE: get_filter_info
#
# DESCRIPTION: Given a filter index number, place into output arg the list 
#              of information about that filter from the 
#              filter assignment file.
#
#              Example: given filter_index 41, we might set the output arg to 
#
#                     41   014  14   0   7   1 filt_014.dat
#
# RETURNS
#   0         if all goes well
#   1         if an error occurs
#
proc get_filter_info { filter_index o_filter_info } {
  global debug 
  upvar $o_filter_info filter_info

  if { $debug > 0 } {
    puts "entering get_filter_info"
  }

  set filter_assn_file [get_filter_assn]
  set name [format "filt_%03d" $filter_index]
  set ll [get_param $filter_assn_file $name]
  if { $ll == {} } {
    puts stderr "get_filter_info: get_param fails for name ..$name.."
    return 1
  }
  set filter_info [lindex $ll 0]

  return 0
}







namespace eval Filter {
  namespace export init_filter get_filter_assn get_filter_fiducial_name \
                               get_filter_num_fiducial \
                               get_filter_index_to_fiducial_index
  variable Filter_Assn_File {}
  variable Filter_Fiducial_Base {}
  variable Filter_Num_Fiducial {}
  variable Filter_Index_To_Fiducial_Index {}


#######################################################################
# Check to make sure that the files for the fiducial filters
#   do exist in the expected place.
#
# The arg should be the name of the filter parameter file
#
# RETURNS
#    0         if all goes well
#    1         if an error occurs
#
proc init_filter { param_file } {
  global debug
  variable Filter_Assn_File
  variable Filter_Fiducial_Base {}
  variable Filter_Num_Fiducial {}
  variable Filter_Index_To_Fiducial_Index {}

  if { $debug > 0 } {
    puts "entering init_filter "
  }
  
  if { [file exists $param_file] != 1 } {
    puts stderr "init_filter: can't open param file $param_file"
    return 1
  }
  set input_dir       [get_param $param_file "input_dir"]
  set output_dir      [get_param $param_file "output_dir"]
  set num_fiducial    [get_param $param_file "num_fiducial"]
  set fiducial_base   [get_param $param_file "fiducial_base"]
  set filter_assn     [get_param $param_file "filter_assn"]

  set retcode 0

  set filtnum 0
  while { $filtnum < $num_fiducial } {
    set filename [format "%s/%s%d.filter" $input_dir $fiducial_base $filtnum]
    if { [file exists $filename] != 1 } {
      puts stderr "init_filter: file $filename doesn't exist?!"
      set retcode 1
    }
    incr filtnum
  }

  # set a private variable to the number of fiducial filters
  set Filter_Num_Fiducial $num_fiducial

  # now, we set the name of the file which contains actual filter
  #   assignments; we will refer to this file later on in several
  #   places as we simulate observations
  set Filter_Assn_File [format "%s/%s" $output_dir $filter_assn]

  # we'll need this later, too
  set Filter_Fiducial_Base [format "%s/%s" $input_dir $fiducial_base]

  # set this to be a big list, with AT LEAST as many elements as there
  #   are filters in the focal plane.  We don't know exactly how many
  #   there will be yet, so we create a ridiculously large number of
  #   elements in the list here, to be on the safe side.  
  #   The list will translate a filter index into the corresponding
  #   fiducial filter index.
  #   We initialize all elements to "-1", which is an invalid fiducial
  #   index; this means "we haven't set the element yet."  The
  #   function "get_filter_index_to_fiducial_index" will do that
  #   initialization for us later.
  set lots_of_filters 500
  set Filter_Index_To_Fiducial_Index [list -1]
  for { set i_filter 0 } { $i_filter < $lots_of_filters } { incr i_filter } {
    lappend Filter_Index_To_Fiducial_Index [list -1]
  }
  if { $debug > 4 } {
    puts "Filter_Index_To_Fiducial_Index is ..$Filter_Index_To_Fiducial_Index.."
  }

  return $retcode
}

#######################################################################
# Return the number of fiducial filters
#
# RETURNS
#    the number of fiducials
#
proc get_filter_num_fiducial { } {
  global debug
  variable Filter_Num_Fiducial

  if { $debug > 0 } {
    puts "entering get_filter_num_fiducial "
  }
  
  return $Filter_Num_Fiducial
} 


#######################################################################
# Return the full name of the file which contains filter assignments
#   to the detectors.
#
# RETURNS
#    name of file
#
proc get_filter_assn { } {
  global debug
  variable Filter_Assn_File

  if { $debug > 0 } {
    puts "entering get_filter_assn "
  }
  
  return $Filter_Assn_File
} 

#######################################################################
# Return the full name of the file which contains filter assignments
#   to the detectors.
#
# RETURNS
#    name of file
#
proc get_filter_fiducial_name { fiducial_index } {
  global debug
  variable Filter_Fiducial_Base

  if { $debug > 0 } {
    puts "entering get_filter_fiducial_name "
  }
  
  set name [format "%s%d.filter" $Filter_Fiducial_Base $fiducial_index]
  return $name
} 


#######################################################################
# Given a filter index number, check to see if we have already looked
#   up the fiducial index corresponding to it.  If we have, just 
#   return the fiducial index.  If not, then go into the 
#   filter assignment file, look up the fiducial index, save it
#   for future reference, and _then_ return the fiducial index.
#
# RETURNS
#   fiducial index corresponding to this filter index
#   -1            if an error occurs
#
proc get_filter_index_to_fiducial_index { filter_index } {
  global debug
  variable Filter_Index_To_Fiducial_Index

  if { $debug > 0 } {
    puts "entering get_filter_index_to_fiducial_index "
  }
  
  set fid_index -1

  set fid_index [lindex $Filter_Index_To_Fiducial_Index $filter_index]
  if { $fid_index == -1 } {
    # we haven't set this element yet.  So look it up ..
    if { $debug > 5 } {
      puts "must look up fiducial index for filter $filter_index"
    }
    if { [get_filter_info $filter_index filter_info] != 0 } {
      puts stderr "get_filter_index_to_fiducial_index: get_filter_index fails for $filter_index"
      return -1
    }
    set fid_index [lindex $filter_info 4]
    set Filter_Index_To_Fiducial_Index [lreplace \
                     $Filter_Index_To_Fiducial_Index \
                     $filter_index $filter_index $fid_index ]
  } else {
    # okay, we know what this is already
    if { $debug > 5 } {
      puts "we know fiducial for filter $filter_index already: $fid_index "
    }
    
  }

  return $fid_index
  
} 



  # end of namespace Filter
}
