
	/*
	 * Calculate the "synthetic magnitude" of a star, given
	 *    - file with "flam" -- flux per unit wavelength interval
	 *    - file with passband transmission at tabulated wavelengths
	 *
	 * The sore point is that the tabulated wavelengths are NOT THE SAME
	 * in the two files.  Sigh.
	 *
	 * Use simple linear interpolation and a fixed bin size.
	 *
	 * Note that the program "lam_ab.c" does the same thing, but 
	 * with files which contain "fnu" -- flux per unit frequency interval.
	 *
	 * MWR 11/20/2002
	 *
	 * Remove incorrect conversion of stellar flux values when reading
	 *   from input file.
	 *
	 * MWR 11/21/2002
	 *
	 * Add a third and fourth argument: 
	 *              fiducial_angle     angle of incidence (rad) at which
	 *                                      filter is designed
	 *              incident_angle     angle (rad) at which light strikes filter
	 *
	 *   We assume that the filter bandpasses shift with angle of 
	 *   incidence as given in Alex Kim's document "Effective Filters",
	 *   mentioned in Calibration mailing list on Nov 2, 2002.
	 *
	 * MWR 8/27/2003
	 *
	 * If the convolved flux is zero (or negative), set the "magnitude"
	 * output value to BADMAG = 99.0, instead of "inf".
	 *
	 * MWR 9/16/2003
	 *
	 * Changed assumption that input spectrum is joules/cm^2-sec-Angstrom
	 * to assumption of ergs/cm^2-sec-Angstrom.  
	 * Also, decrease BIN_WIDTH from 5 to 1 Angstrom.
	 *
	 * MWR 4/19/2004
	 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#undef DEBUG
#define BUFLEN  1024
#define COMMENT_CHAR  '#'

/* we use physical constants in the MKS system */
#define PLANCK_H    6.62607E-34
#define LIGHTSPEED  299792458.0

/* width of bins used for numerical integration (Angstroms) */
#define BIN_WIDTH 1.0

/* arbitrary zero point of "magnitude" */
#define MAGZERO   25.0

/* this means "no data" */
#define BADMAG    99.0

/* max allowed data points in a single file */
#define MAXPTS  100000


int    g_stellar_ndata;
double g_stellar_data[MAXPTS], g_stellar_lam[MAXPTS];
int    g_passband_ndata;
double g_passband_data[MAXPTS], g_passband_lam[MAXPTS];


/* subroutines */
static int read_stellar_data(char *stellar_file);
static int read_passband_data(char *passband_file);
static int shift_passband_data(double fiducial_angle, double incident_angle);
static double get_stellar_flux(double lambda);
static double get_passband_transmission(double lambda);
static int    do_integration(void);
static double effective_lambda(double fiducial_angle, 
                               double incident_angle, double lambda);




	/* 
	 * Usage: flam_mag  stellar_file  passband_file  fiducial_angle incident_angle 
	 */


int 
main
	(
	int argc,
	char *argv[]
	)
{
	char stellar_file[BUFLEN];
	char passband_file[BUFLEN];
	double fiducial_angle, incident_angle;
#ifdef DEBUG
	double lambda;
	double flux;
#endif


	if (argc != 5) {
		fprintf(stderr, "usage: flam_mag stellar_file passband_file fiducial_angle incident_angle\n");
		exit(1);
	}
	sscanf(argv[1], "%s", stellar_file);
	sscanf(argv[2], "%s", passband_file);
	sscanf(argv[3], "%lf", &fiducial_angle);
	sscanf(argv[4], "%lf", &incident_angle);
#ifdef DEBUG
	printf(" stellar_file %s passband_file %s fiducial_angle %lf incident_angle %lf\n",
			stellar_file, passband_file, fiducial_angle, incident_angle);
#endif

	/* read in the data for each file */
	if (read_stellar_data(stellar_file) != 0) {
		fprintf(stderr, "error reading data from stellar file %s \n", 
				stellar_file);
		exit(1);
	}
	if (read_passband_data(passband_file) != 0) {
		fprintf(stderr, "error reading data from passband file %s \n", 
				passband_file);
		exit(1);
	}

	/* shift the wavelength data according to the angle of incidence */
	if (shift_passband_data(fiducial_angle, incident_angle) != 0) {
		fprintf(stderr, "error shifting passband data for angle %lf \n", 
				incident_angle);
		exit(1);
	}

#ifdef DEBUG2
	printf(" gimme lambda: ");
	scanf("%lf", &lambda);
	flux = get_stellar_flux(lambda);
	printf("  stellar flux  %9.4e \n", flux);
	flux = get_passband_transmission(lambda);
	printf("  transmission  %9.4f \n", flux);
#endif


	/* okay, we're ready for the integration */
	if (do_integration() != 0) {
		fprintf(stderr, "flam_mag: do_integration returns with error\n");
		exit(1);
	}

	return(0);
}


/***************************************************************************
 * PROCEDURE: read_stellar_data
 *
 * DESCRIPTION: reads the data about the stellar AB flux from the given file.
 *              The file is assumed to contain two columns of data:
 *
 *                  lam   flam
 *
 *              where
 *                    lambda       wavelength (in Angstroms)
 *
 *                    flam         is flux per unit wavelength
 *                                           (ergs/cm^2-s-Angstrom)
 *
 *              We leave the flux in its native units of
 *              (ergs per cm^2 per second per Angstrom);
 *              these are placed into the array g_stellar_data[].  
 *              The wavelength is placed into 
 *              into g_stellar_lam[].  The number
 *              of data values is placed into g_stellar_ndata variable.
 * RETURNS:
 *    0        if all goes well
 *    1        if an error occurs
 */

static int
read_stellar_data
	(
	char *stellar_file            /* I: name of file with stellar data */
	)
{
	int i, ndata;
	char line[BUFLEN+1];
	double abval, flux;
	FILE *fp;

	if ((fp = fopen(stellar_file, "r")) == NULL) {
		fprintf(stderr, "read_stellar_data: can't open file %s\n", 
				stellar_file);
		return(1);
	}

	/* 
	 * go line-by-line through the data file.
	 * Skip lines starting with "#", or blank lines.
	 */
	ndata = 0;
	while (fgets(line, BUFLEN, fp) != NULL) {

		/* check to see if we've run out of space. If so, quit with error */
		if (ndata >= MAXPTS) {
			fprintf(stderr, "read_stellar_data: too many data points\n");
			return(1);
		}

		/* skip lines starting with comment char */
		if (line[0] == COMMENT_CHAR) {
			continue;
		}

		/* skip blank lines, too */
		for (i = 0; isblank(line[i]); i++) {
			;
		}
		if (line[i] == '\n') {
			continue;
		}

#ifdef DEBUG
		printf("read_stellar_data: next good line is ..\n%s", line);
#endif

		if (sscanf(line, "%lf %lf", &g_stellar_lam[ndata],
				&abval) != 2) {
			fprintf(stderr, "read_stellar_data: bad data line follows; quitting\n");
			fprintf(stderr, "%s\n", line);
			return(1);
		}

		/* 
		 * we assume that the value we read is already
		 *      ergs / cm^2-s-Angstrom
		 */
		flux = abval;
		g_stellar_data[ndata] = flux;

#ifdef DEBUG
		printf(" lam %8.2f  data %9.4e \n", 
				    g_stellar_lam[ndata], g_stellar_data[ndata]);
#endif

		ndata++;
	}

	fclose(fp);
	g_stellar_ndata = ndata;

	return(0);
}



/***************************************************************************
 * PROCEDURE: read_passband_data
 *
 * DESCRIPTION: reads the data about a passband from the given file.
 *              The file is assumed to contain two columns of data:
 *
 *                  lambda    trans
 *
 *              where
 *                    lambda       wavelength (in Angstroms)
 *
 *                    trans        transmission (normalized)
 *
 *              The values are placed into the global array
 *              g_passband_data[] and g_passband_lam[].  The number
 *              of data values is placed into g_passband_ndata variable.
 * RETURNS:
 *    0        if all goes well
 *    1        if an error occurs
 */

static int
read_passband_data
	(
	char *passband_file         /* I: name of file with passband data */
	)
{
	int i, ndata;
	char line[BUFLEN+1];
	FILE *fp;

	if ((fp = fopen(passband_file, "r")) == NULL) {
		fprintf(stderr, "read_passband_data: can't open file %s\n", 
				passband_file);
		return(1);
	}

	/* 
	 * go line-by-line through the data file.
	 * Skip lines starting with "#", or blank lines.
	 */
	ndata = 0;
	while (fgets(line, BUFLEN, fp) != NULL) {

		/* check to see if we've run out of space. If so, quit with error */
		if (ndata >= MAXPTS) {
			fprintf(stderr, "read_passband_data: too many data points\n");
			return(1);
		}

		/* skip lines starting with comment char */
		if (line[0] == COMMENT_CHAR) {
			continue;
		}

		/* skip blank lines, too */
		for (i = 0; isblank(line[i]); i++) {
			;
		}
		if (line[i] == '\n') {
			continue;
		}

#ifdef DEBUG
		printf("read_passband_data: next good line is ..\n%s", line);
#endif

		if (sscanf(line, "%lf %lf", &g_passband_lam[ndata],
				&g_passband_data[ndata]) != 2) {
			fprintf(stderr, "read_passband_data: bad data line follows; quitting\n");
			fprintf(stderr, "%s\n", line);
			return(1);
		}

#ifdef DEBUG
		printf(" line %4d  lam %8.2f  data %8.4f \n", ndata, 
				g_passband_lam[ndata], g_passband_data[ndata]);
#endif

		ndata++;
	}

	fclose(fp);
	g_passband_ndata = ndata;

	return(0);
}



/***************************************************************************
 * PROCEDURE: shift_passband_data
 *
 * DESCRIPTION: Shift the nominal passband wavelengths to their
 *              effective values, based on the angle of incidence.
 *
 *              We simply walk through the values in the global array
 *              g_passband_lam[], modifying each one as we go.  
 *
 * RETURNS:
 *    0        if all goes well
 *    1        if an error occurs
 */

static int
shift_passband_data
	(
	double fiducial_angle,      /* I: angle of incidence (radians) at which */
	                            /*       filter was designed; the tabulated */
	                            /*       values apply at this angle only */
	double incident_angle       /* I: actual angle of incidence (radians) */
	)
{
	int i;
	double old_lambda, new_lambda;

	for (i = 0; i < g_passband_ndata; i++) {
		old_lambda = g_passband_lam[i];
		new_lambda = effective_lambda(fiducial_angle, incident_angle, old_lambda);
		g_passband_lam[i] = new_lambda;

#ifdef DEBUG
		printf(" line %4d  old lam %8.2f  new lam  %8.2f \n", i,
				old_lambda, new_lambda);
#endif

	}
	return(0);
}




/*****************************************************************************
 * PROCEDURE: get_stellar_flux
 *
 * DESCRIPTION: Given a wavelength, return the stellar flux, 
 *              units Watts per cm^2 per sec per Angstrom.
 *              Use simple linear interpolation between the two nearest
 *              wavelengths.
 *
 *              If the given wavelength is outside the range of wavelengths
 *              with known irradiance values, return zero.
 *
 * RETURNS:
 *    flux (or zero)
 *
 */

static double
get_stellar_flux
	(
	double lambda             /* I: wavelength of interest */
	)
{
	int i;
	double low_lam, high_lam;
	double low_flux, high_flux;
	double frac;
	double flux;

	/* check for out of bounds */
	if (lambda <= g_stellar_lam[0]) {
		return(0);
	}
	if (lambda >= g_stellar_lam[g_stellar_ndata-1]) {
		return(0);
	}

	for (i = 1; i < g_stellar_ndata; i++) {
		if (g_stellar_lam[i] > lambda) {

			/* interpolate linearly between "i" and "i-1" */
			low_lam = g_stellar_lam[i - 1];
			high_lam = g_stellar_lam[i];
			/* sanity check */
			if (high_lam <= low_lam) {
				fprintf(stderr, "get_stellar_flux: bad stellar lambdas\n");
				fprintf(stderr, "  %5d %8.2f  %5d %8.2f \n", i-1, low_lam, i, high_lam);
				exit(1);
			}
						
			low_flux = g_stellar_data[i - 1];
			high_flux = g_stellar_data[i];

			frac = (lambda - low_lam)/(high_lam - low_lam);
			flux = low_flux + frac*(high_flux - low_flux);

			return(flux);

		}
	}

	/* should never get here! */
	fprintf(stderr, "get_stellar_flux: dropped off end?!\n");
	exit(1);

}


/*****************************************************************************
 * PROCEDURE: get_passband_transmission
 *
 * DESCRIPTION: Given a wavelength, return the passband transmission
 *              (where "0" means no light passes through, and "1" means
 *              all light passes through).
 *              Use simple linear interpolation between the two nearest
 *              wavelengths.
 *
 *              If the given wavelength is outside the range of wavelengths
 *              with known transmission values, return zero.
 *
 * RETURNS:
 *    transmission (or zero)
 *
 */

static double
get_passband_transmission
	(
	double lambda             /* I: wavelength of interest */
	)
{
	int i;
	double low_lam, high_lam;
	double low_trans, high_trans;
	double frac;
	double trans;

	/* check for out of bounds */
	if (lambda <= g_passband_lam[0]) {
		return(0);
	}
	if (lambda >= g_passband_lam[g_passband_ndata-1]) {
		return(0);
	}

	for (i = 1; i < g_passband_ndata; i++) {
		if (g_passband_lam[i] > lambda) {

			/* interpolate linearly between "i" and "i-1" */
			low_lam = g_passband_lam[i - 1];
			high_lam = g_passband_lam[i];
			/* sanity check */
			if (high_lam <= low_lam) {
				fprintf(stderr, "get_passband_transmission: bad passband lambdas\n");
				fprintf(stderr, "  %5d %8.2f  %5d %8.2f \n", i-1, low_lam, i, high_lam);
				exit(1);
			}
						
			low_trans = g_passband_data[i - 1];
			high_trans = g_passband_data[i];

			frac = (lambda - low_lam)/(high_lam - low_lam);
			trans = low_trans + frac*(high_trans - low_trans);

			return(trans);

		}
	}

	/* should never get here! */
	fprintf(stderr, "get_passband_transmission: dropped off end?!\n");
	exit(1);

}


/***************************************************************************
 * PROCEDURE: do_integration
 *
 * DESCRIPTION: performs the convolution/integration over the passband.
 *              Calculates the number of photons per cm^2 per second.
 *              Prints out the number of photons,
 *              plus a "magnitude" based upon it with an arbitrary
 *              zero point.
 *
 * RETURNS:
 *   0        if all goes well
 *   1        if an error occurs
 */

static int
do_integration
	(
	void
	)
{
	double start_lam, end_lam, lam, cent_lam, delta_lam;
	double flux, trans;
	double ergs, sum_ergs;
	double joules_per_photon, ergs_per_photon;
	double photons, sum_photons;
	double mag;

	/* figure out the limits of integration (based on passband) */
	start_lam = g_passband_lam[0];
	end_lam = g_passband_lam[g_passband_ndata-1];
#ifdef DEBUG
	printf("do_integration: integration limits %9.2f %9.2f \n", 
			start_lam, end_lam);
#endif

	/* loop over wavelengths */
	sum_photons = 0;
	sum_ergs = 0;
	for (lam = start_lam; lam < end_lam; lam += BIN_WIDTH) {

		/* use the central wavelength within this bin */
		cent_lam = lam + (BIN_WIDTH/2.0);
		flux = get_stellar_flux(cent_lam);
		trans = get_passband_transmission(cent_lam);
#ifdef DEBUG
		printf("lam %9.2f   flux %9.4e   trans %9.4f \n",
				cent_lam, flux, trans);
#endif

		/* 
		 * calculate the width of this bin in Angstroms
		 */
		delta_lam = BIN_WIDTH;
#ifdef DEBUG
		printf("    start_lam   %9.4e  end_lam  %9.4e   delta_lam  %9.4e\n",
				lam, lam + BIN_WIDTH, BIN_WIDTH);
#endif

		/* 
		 * first, calculate the summed flux (assumed to be ergs per sec per cm^2)
		 * through the passband within this little bin of wavelengths 
		 */
		ergs = flux*trans*delta_lam;
		sum_ergs += ergs;
#ifdef DEBUG
		printf("   ergs  %9.5e   sum_ergs  %9.5e \n", ergs , sum_ergs );
#endif

		/* 
		 * now, calculate the number of photons per sec per cm^2 
		 * in this bin
		 */
		joules_per_photon = (PLANCK_H*LIGHTSPEED)/(cent_lam*1E-10);
		ergs_per_photon = joules_per_photon*1.0e7;
		photons = ergs/ergs_per_photon;
		sum_photons += photons;
#ifdef DEBUG
		printf("   ergs   %9.5e  photons %9.5e  sum_photons %9.5e \n",
				ergs_per_photon, photons, sum_photons);
#endif

	}

	/* convert number of photons to a "mag" */
	if (sum_photons <= 0) {
		mag = BADMAG;
	}
	else {
	 	mag = MAGZERO - 2.5*log10(sum_photons);
	}

	/* all done -- print out the results */
	printf("   ergs   %9.4e  photons %9.4e   'mag' %9.3f \n",
			sum_ergs, sum_photons, mag);

	return(0);
}



/***************************************************************************
 * PROCEDURE: effective_lambda
 *
 * DESCRIPTION: Given the angle of incidence (in radians), and a
 *              wavelength, calculate the effective wavelength.
 *              The filter bandpasses shift a bit depending on the
 *              angle at which light strikes them, as described
 *              in Alex Kim's document "Effective Filters".
 *
 * RETURNS:
 *     the shifted wavelength
 *
 */
static double 
effective_lambda
	(
	double fiducial_angle,          /* I: design angle of incidence (rad) */
	double incident_angle,          /* I: actual angle of incidence (rad) */
	double lambda                   /* I: filter's ordinary wavelength */
	)
{
	double xx, new_lambda;

	/* calculate the shifted wavelength */
	xx = cos(incident_angle) - cos(fiducial_angle);
	new_lambda = lambda * ( 1.0 + ((1.0/3.0)*xx) );

	return(new_lambda);
}
