#
# This TCL script reads a high-level description of the SNAP focal
#   plane and generates a lower-level description describing the
#   location and orientation and pixel size of each chip.
#   The output will be placed into a .param file itself,
#   and then used by modules downstream.
# 
# MWR  8/12/2003
#    
# Added routine "create_sensitivity_params" to generate a model of 
#   the variation in sensitivity across each chip, based on the value
#   of the "foc_chip_variation" parameter.
# MWR  8/18/2004


if { [catch { set debug } ] == 1 } {
  set debug 2
}

source param.tcl


###########################################################################
# Read information from the high-level focal plane description,
#   and generate a low-level focal plane description.
#
# RETURNS
#    0           if all goes well
#    1           if there's a problem
#    
proc focal { param_file } {
  global debug

  if { $debug > 0 } {
    puts "entering focal"
  }

  # get information from the param file
  if { [file exists $param_file] != 1 } {
    puts stderr "focal: can't open param file $param_file"
    return 1
  }
  set param_dir [get_param $param_file "param_dir"]
  set foc_output_file [get_param $param_file "foc_output_file"]
  set foc_annulus_inner [get_param $param_file "foc_annulus_inner"]
  set foc_annulus_outer [get_param $param_file "foc_annulus_outer"]
  set foc_gridsize [get_param $param_file "foc_gridsize"]
  set foc_vis_npix [get_param $param_file "foc_vis_npix"]
  set foc_vis_pixsize [get_param $param_file "foc_vis_pixsize"]
  set foc_ir_npix [get_param $param_file "foc_ir_npix"]
  set foc_ir_pixsize [get_param $param_file "foc_ir_pixsize"]
  set foc_east_only [get_param $param_file "foc_east_only"]
  set foc_evis [get_param $param_file "foc_evis"]
  set foc_eir  [get_param $param_file "foc_eir"]
  if { $foc_east_only == 0 } {
    set foc_wvis [get_param $param_file "foc_wvis"]
    set foc_wir  [get_param $param_file "foc_wir"]
    set foc_nvis [get_param $param_file "foc_nvis"]
    set foc_nir  [get_param $param_file "foc_nir"]
    set foc_svis [get_param $param_file "foc_svis"]
    set foc_sir  [get_param $param_file "foc_sir"]
  }
  set foc_vis_prop [get_param $param_file "foc_vis_prop"]
  set foc_ir_prop  [get_param $param_file "foc_ir_prop"]
  set foc_chip_variation [get_param $param_file "foc_chip_variation"]

  
  # create the output param file, and open it for writing
  set output_filename [format "%s/%s" $param_dir $foc_output_file]
  if { [catch { set fid [open $output_filename "w"] } ] != 0 } {
    puts stderr "focal: can't open $output_filename for writing"
    return 1
  }

  # generate a header for the output file
  if { [write_chip_header $fid] != 0 } {
    puts stderr "focal: write_chip_header returns with error"
    return 1
  }

  #
  # Now, we just walk through the list of all the blocks of detectors,
  #    generating param information for the chips in each block
  if { [init_chipnum] != 0 } {
    puts stderr "focal: init_chipnum returns with error"
    return 1
  }
  if { [generate_block $fid $foc_gridsize $foc_vis_npix $foc_vis_pixsize \
                    $foc_vis_prop $foc_evis $foc_chip_variation] != 0 } {
    puts stderr "focal: generate_block ..$foc_evis.. returns with error"
    return 1
  }
  if { [generate_block $fid $foc_gridsize $foc_ir_npix $foc_ir_pixsize \
                    $foc_ir_prop $foc_eir $foc_chip_variation] != 0 } {
    puts stderr "focal: generate_block ..$foc_eir.. returns with error"
    return 1
  }
  if { $foc_east_only == 0 } {
    if { [generate_block $fid $foc_gridsize $foc_vis_npix $foc_vis_pixsize \
                      $foc_vis_prop $foc_nvis $foc_chip_variation] != 0 } {
      puts stderr "focal: generate_block ..$foc_nvis.. returns with error"
      return 1
    }
    if { [generate_block $fid $foc_gridsize $foc_ir_npix $foc_ir_pixsize \
                      $foc_ir_prop $foc_nir $foc_chip_variation] != 0 } {
      puts stderr "focal: generate_block ..$foc_nir.. returns with error"
      return 1
    }
    if { [generate_block $fid $foc_gridsize $foc_vis_npix $foc_vis_pixsize \
                      $foc_vis_prop $foc_wvis $foc_chip_variation] != 0 } {
      puts stderr "focal: generate_block ..$foc_wvis.. returns with error"
      return 1
    }
    if { [generate_block $fid $foc_gridsize $foc_ir_npix $foc_ir_pixsize \
                      $foc_ir_prop $foc_wir $foc_chip_variation] != 0 } {
      puts stderr "focal: generate_block ..$foc_wir.. returns with error"
      return 1
    }
    if { [generate_block $fid $foc_gridsize $foc_vis_npix $foc_vis_pixsize \
                      $foc_vis_prop $foc_svis $foc_chip_variation] != 0 } {
      puts stderr "focal: generate_block ..$foc_svis.. returns with error"
      return 1
    }
    if { [generate_block $fid $foc_gridsize $foc_ir_npix $foc_ir_pixsize \
                      $foc_ir_prop $foc_sir $foc_chip_variation] != 0 } {
      puts stderr "focal: generate_block ..$foc_sir.. returns with error"
      return 1
    }
  }


  # write the total number of chips in the low-level param file
  set num_chips [next_chipnum]
  puts $fid [format "num_chips     %5d " $num_chips]



  # close the output file
  close $fid
    
  
  return 0
}



############################################################################
# PROCEDURE: write_chip_header
# 
# DESCRIPTION: Write a bit of information at the top of the low-level chip
#              param file.  Args:
#
#                  fid               file descriptor for output param file
#
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc write_chip_header { fid } {
  global debug 

  if { $debug > 0 } {
    puts "entering write_chip_header "
  }

  set datestr [exec date]
  puts $fid "# generated on $datestr"


  return 0
}



############################################################################
# PROCEDURE: generate_block
# 
# DESCRIPTION: Generate the parameters for one block (3x3 array) of 
#              detectors, based on the arguments.  Write the parameters
#              into the give file descriptor.
#
#              Args:
#
#                  fid               file descriptor for output param file
#                  gridsize          size of grid in focal plane on which
#                                         chips are positioned (mm)
#                  npix              number of pixels along each dir of chip
#                  pixsize           size of pixel (mm)
#                  detector_prop     list of detector properties, which has
#                        dark             dark current
#                        read             readout noise
#                        gain             gain (e-/DN)
#                        sat              saturation/non-lin level (e-)
#                        offset           loss in sensitivity (mag)
#                  block_list        which contains
#                        igrid            x-center of block in grid coords
#                        jgrid            y-center of block in grid coords
#                        type             "vis" or "ir"
#                        angle            rotation angle of chips (degrees)
#                  chip_variation    describes the way we will insert
#                                         sensitivity variations across each
#                                         chip
# 
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc generate_block { fid gridsize npix pixsize detector_prop block_list \
                                      chip_variation } {
  global debug 

  if { $debug > 0 } {
    puts "entering generate_block "
  }

  set DEG2RAD [expr 3.14159/180.0]

  # sanity checks
  set pl [lindex $detector_prop 0]
  if { [llength $pl] != 5 } {
    puts stderr [format "generate_block: detector_prop has %d elem, not 5 ?!" \
                                [llength $pl] ]
    return 1
  }
  set bl [lindex $block_list 0]
  if { [llength $bl] != 4 } {
    puts stderr [format "generate_block: block_list has %d elem, not 4 ?!" \
                                [llength $bl] ]
    return 1
  }

  # is this a visible-light CCD or an near-IR detector?
  #   "chip_type" will be set to "vis" or "ir"
  set chip_type [lindex $block_list 2]

  # we loop over the 3x3 array of chips within this block
  foreach iblock { 0 1 2 } {
   foreach jblock { 0 1 2 } {

     # get the next available chip number
     set chipnum [next_chipnum]

     # get the data from the list
     set igrid [lindex $bl 0]
     set jgrid [lindex $bl 1]
     set type  [lindex $bl 2]
     set angle [lindex $bl 3]
   
     if { $debug > 1 } {
       puts "generate_block: $iblock $jblock  igrid $igrid jgrid $jgrid type $type angle $angle"
     }
  
     # the center of this chip in the focal plane (mm)
     set chip_center_x [expr ($igrid + ($iblock-1))*$gridsize]
     set chip_center_y [expr ($jgrid + ($jblock-1))*$gridsize]

     # prepare a line for the new param file, describing everything 
     #   about this chip
     set line [format "chip_%03d { %3d %4s %4d %6.4f %6.2f " \
                 $chipnum $chipnum $type $npix $pixsize $angle ]
     set line [format "%s %8.3f %8.3f " $line $chip_center_x $chip_center_y]

     # now, the locations of the four corners in the focal plane (mm)
     #   we always list the corners as follows:
     #      (row=0, col=0)    
     #      (row=0, col=N)    
     #      (row=N, col=N)    
     #      (row=N, col=0)    
     set dist [expr ($npix/2)*$pixsize*sqrt(2.0)]
     foreach c_angle { 0 90 180 270 } {
       set theta [expr 225+$angle+$c_angle]
       set theta_rad [expr $theta*$DEG2RAD]
       set chip_corner_x [expr $chip_center_x + $dist*cos($theta_rad)]
       set chip_corner_y [expr $chip_center_y + $dist*sin($theta_rad)]

       set line [format "%s %8.3f %8.3f " $line $chip_corner_x $chip_corner_y]
     }

     # get the chip properties
     set dark [get_image_value $pl "dark"]
     set read [get_image_value $pl "read"]
     set gain [get_image_value $pl "gain"]
     set sat  [get_image_value $pl "sat"]
     set offset [get_image_value $pl "offset"]
     if { $debug > 2 } {
       puts "generate_block: dark $dark read $read gain $gain sat $sat offset $offset"
     }
     set line [format "%s %.4f %4.1f %4.2f %5.0f %5.3f " \
                      $line $dark $read $gain $sat $offset ]

     # here, we insert the parameters which describe the variations
     #   in sensitivity across this chip
     if { [create_sensitivity_params $chip_variation $npix \
                                         sensitivity_list ] != 0 } {
       puts stderr "generate_block: create_sensitivity_params fails"
       return 1
     }
     set line [format "%s { %s } " $line $sensitivity_list]

     set line [format "%s } " $line ]

     # okay, now we're ready to write out the entire line, which contains
     #   a list like this:
     #
     #     chip_NNN { id type ... dark read gain sat offset { sens_list }  }
     #
     puts $fid $line


   }
  }

  return 0

}



#############################################################################
# PROCEDURE: radec_to_focxy
#
# DESCRIPTION: Given an (RA, Dec) position (J2000), 
#              calculate the (x, y) in mm on the focal plane.
#
#              We currently read the central (RA, Dec) position of the 
#              telescope, and the roll angle, from the focal.param file.
#
# RETURNS:
#   0          if all goes well
#   1          if an error occurs
#
proc radec_to_focxy { ra dec o_x o_y } {
  global debug
  upvar $o_x x $o_y y 
  
  if { $debug > 0 } {
    puts "entering radec_to_focxy"
  }

  set DEGTORAD [expr 3.14159/180.0]

  set flist [get_foc_info]
  set scale_factor [lindex $flist 0]
  set roll_angle [lindex $flist 1]
  set central_ra [lindex $flist 2]
  set central_dec [lindex $flist 3]
  if { $debug > 5 } {
    puts "  scale_factor $scale_factor"
    puts "  roll_angle   $roll_angle "
    puts "  central_ra   $central_ra"
    puts "  central_dec  $central_dec"
  }

  # convert roll angle to radians
  set roll_angle [expr $DEGTORAD*$roll_angle]

  # to convert from (RA, Dec) to project plate coords (xi, eta), 
  #    use the complex formulae shown in Olkin et al., 
  #    PASP 108, 202 (1996).
  #    (this is better done in C for speed, I'm sure)
  set cosdec0 [expr cos($DEGTORAD*$central_dec)]
  set sindec0 [expr sin($DEGTORAD*$central_dec)]
  set cosdec  [expr cos($DEGTORAD*$dec)]
  set sindec  [expr sin($DEGTORAD*$dec)]
  set sin_delta_ra [expr sin($DEGTORAD*($ra - $central_ra))]
  set cos_delta_ra [expr cos($DEGTORAD*($ra - $central_ra))]

  # calculate xi, the pseudo-RA coordinate 
  set xx [expr $cosdec*$sin_delta_ra]
  set yy [expr $sindec0*$sindec + $cosdec0*$cosdec*$cos_delta_ra]
  set xi [expr $xx/$yy]
  if { $debug > 5 } {
    puts "  xx $xx   yy $yy   xi $xi "
  }

  # calculate eta, the pseudo-Dec coordinate 
  set xx [expr $cosdec0*$sindec - $sindec0*$cosdec*$cos_delta_ra]
  set eta [expr $xx/$yy]
  if { $debug > 5 } {
    puts "  xx $xx   yy $yy   eta $eta "
  }

  # calculate the polar equivalent of (xi, eta)
  set r [expr sqrt($xi*$xi + $eta*$eta)]
  if { $r == 0 } {
    set theta 0
  } else {
    set theta [expr atan2($eta, $xi)]
  }
  if { $debug > 5 } {
    puts "  r $r  theta $theta "
  }
  
  # and add to this angle the spacecraft's roll angle
  set theta [expr $theta+$roll_angle]
  if { $debug > 5 } {
    puts "   now  theta $theta "
  }

  # now multiply the polar radius by the scale factor to convert to mm 
  #   on the focal plane
  set r_mm [expr $r*$scale_factor]
  if { $debug > 5 } {
    puts "   r_mm $r_mm "
  }

  # and finally, go back to cartesian coords from polar coords
  set x [expr $r_mm*cos($theta)]
  set y [expr $r_mm*sin($theta)]
  if { $debug > 5 } {
    puts "   x $x   y $y "
  }

  # invert the "x" coordinate so that we have East to the left,
  #   and East to negative values of "x"
  set x [expr 0.0-$x]
  if { $debug > 5 } {
    puts "   x $x   y $y "
  }

  return 0
}


#############################################################################
# PROCEDURE: focxy_to_radec
#
# DESCRIPTION: Given an (x, y) position in mm on the focal plane,
#              calculate the (RA, Dec) in J2000.
#
#              We currently read the central (RA, Dec) position of the 
#              telescope, the roll angle and other information
#              from the focal.param file.
#
# RETURNS:
#   0          if all goes well
#   1          if an error occurs
#
proc focxy_to_radec { x y o_ra o_dec } {
  global debug
  upvar $o_ra ra $o_dec dec 
  
  if { $debug > 0 } {
    puts "entering focxy_to_radec"
  }

  set DEGTORAD [expr 3.14159/180.0]

  set flist [get_foc_info]
  set scale_factor [lindex $flist 0]
  set roll_angle [lindex $flist 1]
  set central_ra [lindex $flist 2]
  set central_dec [lindex $flist 3]
  if { $debug > 5 } {
    puts "  scale_factor $scale_factor"
    puts "  roll_angle   $roll_angle "
    puts "  central_ra   $central_ra"
    puts "  central_dec  $central_dec"
  }

  # convert roll angle to radians
  set roll_angle [expr $DEGTORAD*$roll_angle]

  # invert the "x" coordinate; it has been inverted so that we have 
  #   East to the left, and we now return it to a right-handed coord system
  set x [expr 0.0-$x]
  if { $debug > 5 } {
    puts "   x $x   y $y "
  }

  # go from cartesian coords to polar coords on focal plane
  set r_mm [expr sqrt($x*$x + $y*$y)]
  if { $r_mm == 0.0 } {
    set alpha 0.0
  } else {
    set alpha [expr atan2($y, $x)]
  }
  if { $debug > 5 } {
    puts "   r_mm $r_mm   alpha $alpha "
  }

  # now divide the polar radius by the scale factor to convert from mm 
  #   to plate coords (radians) on focal plane
  set r [expr $r_mm/$scale_factor]
  if { $debug > 5 } {
    puts "   r $r "
  }

  # subtract the spacecraft's roll angle from the angle alpha, to get
  #   the angle relative to (RA, Dec) coord system
  set theta [expr $alpha-$roll_angle]
  if { $debug > 5 } {
    puts "   now  theta $theta "
  }

  # calculate the cartesian equivalent of (r, theta)
  #   these will be plate coords (xi, eta)
  set xi [expr $r*cos($theta)]
  set eta [expr $r*sin($theta)]
  if { $debug > 5 } {
    puts "  xi $xi  eta $eta "
  }
  

  # to convert from plate coords (xi, eta) to (RA, Dec)  
  #    use the complex formulae shown in Olkin et al., 
  #    PASP 108, 202 (1996).
  #    (this is better done in C for speed, I'm sure)
  set cosdec0 [expr cos($DEGTORAD*$central_dec)]
  set sindec0 [expr sin($DEGTORAD*$central_dec)]

  # calculate RA
  set xx $xi
  set yy [expr ($cosdec0 - $eta*$sindec0)]
  set delta_ra [expr atan2($xx, $yy)]
  set ra [expr $central_ra + ($delta_ra/$DEGTORAD)]

  # calculate Dec 
  set xx [expr ($sindec0 + $eta*$cosdec0)*(cos($delta_ra))]
  set zz [expr ($xx/$yy)]
  set dec [expr atan2($xx, $yy)/$DEGTORAD]

  if { $debug > 5 } {
    puts "  RA $ra   Dec $dec  "
  }


  return 0
}


#############################################################################
# PROCEDURE: radec_to_focxy_and_back
#
# DESCRIPTION: Test the routines which take (RA, Dec) -> (x, y)
#              and then back again.  
#
#              This routine used only in testing, obviously.
#
# RETURNS
#     0        if all tests pass
#     1        if any test fails
#
proc radec_to_focxy_and_back { } {
  global debug

  if { $debug > 0 } {
    puts "entering radec_to_focxy_and_back"
  }

  set retcode 0

  # we allow the initial and final RA and Dec to differ by at most
  #   this much (decimal degrees).  
  set epsilon 0.00001

  # get the central RA and Dec, which we'll need to run the tests
  set flist [get_foc_info]
  set scale_factor [lindex $flist 0]
  set roll_angle [lindex $flist 1]
  set central_ra [lindex $flist 2]
  set central_dec [lindex $flist 3]
  if { $debug > 5 } {
    puts "  scale_factor $scale_factor"
    puts "  roll_angle   $roll_angle "
    puts "  central_ra   $central_ra"
    puts "  central_dec  $central_dec"
  }

  # We move by this far (degrees) away from the center in
  #   both RA and Dec to test each of the quadrants
  set shift 0.5

  # We run a number of tests, using the center and each of the 
  #   eight octants of the focal plane.  
  #
  # The center
  set ra $central_ra
  set dec $central_dec
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for center"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 1: right-right-up
  set ra [expr $central_ra-(2.0*$shift)]
  set dec [expr $central_dec+(1.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 1"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 2: right-up-up
  set ra [expr $central_ra-(1.0*$shift)]
  set dec [expr $central_dec+(2.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 2"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 3: left-up-up
  set ra [expr $central_ra+(1.0*$shift)]
  set dec [expr $central_dec+(2.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 3"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 4: left-left-up
  set ra [expr $central_ra+(2.0*$shift)]
  set dec [expr $central_dec+(1.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 4"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 5: left-left-down
  set ra [expr $central_ra+(2.0*$shift)]
  set dec [expr $central_dec-(1.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 5"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 6: left-down-down
  set ra [expr $central_ra+(1.0*$shift)]
  set dec [expr $central_dec-(2.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 6"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 7: right-down-down
  set ra [expr $central_ra-(1.0*$shift)]
  set dec [expr $central_dec-(2.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 7"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }

  # Move to octant 8: right-right-down
  set ra [expr $central_ra-(2.0*$shift)]
  set dec [expr $central_dec-(1.0*$shift)]
  if { [test_one_radec_conversion $ra $dec $epsilon] != 0 } {
    puts stderr "radec_to_focxy_and_back: test fails for octant 8"
    incr retcode
  }
  if { $debug > 10 } {
    puts "hit return to continue "; set xxx [gets stdin]
  }



  return $retcode
}

#############################################################################
# PROCEDURE: test_one_radec_conversion
#
# DESCRIPTION: Given a single (RA, Dec) position, convert it to focal plane
#              (x, y) coordinates, and then convert those back into
#              RA and Dec again.  Compare the initial and final RA, Dec
#              values.  If they differ by more than the "epsilon" arg,
#              print an error message and return with positive error code:
#              the number of errors which occur in the conversion.
#
# RETURNS
#     0        if the initial and final positions are close enough
#     > 0      if the positions differ
#
proc test_one_radec_conversion { ra dec epsilon } {
  global debug

  if { $debug > 0 } {
    puts "test_one_radec_conversion "
  }

  set retcode 0

  if { [radec_to_focxy $ra $dec x y] != 0 } {
    puts stderr "test_one_radec_conversion: radec_to_focxy fails on $ra $dec"
    incr retcode 
  }  
  if { [focxy_to_radec $x $y final_ra final_dec] != 0 } {
    puts stderr "test_one_radec_conversion: focxy_to_radec fails on $x $y "
    incr retcode 
  }  
  set dra [expr $final_ra-$ra]
  set ddec [expr $final_dec-$dec]
  if { [expr abs($dra)-$epsilon] > 0 } {
    puts stderr "test_one_radec_conversion: dra $dra > eps $epsilon "
    incr retcode
  }
  if { [expr abs($ddec)-$epsilon] > 0 } {
    puts stderr "test_one_radec_conversion: ddec $ddec > eps $epsilon "
    incr retcode
  }
 
  return $retcode
}





  




#############################################################################
# PROCEDURE: radec_to_chip
#
# DESCRIPTION: Given an (RA, Dec) position (J2000), 
#              calculate the (chip, row, col) coordinates.
#
#              This is just a combination of 
#                       radec_to_focxy
#                       focxy_to_chip
#
# RETURNS:
#   0          if all goes well
#   1          if an error occurs
#
proc radec_to_chip { ra dec o_chip o_row o_col } {
  global debug
  upvar $o_chip chip $o_row row $o_col col 
  
  if { $debug > 0 } {
    puts "entering radec_to_chip"
  }

  if { [radec_to_focxy $ra $dec x y] != 0 } {
    puts stderr "radec_to_chip: radec_to_focxy fails "
    return 1
  }
  if { [focxy_to_chip $x $y chip row col] != 0 } {
    puts stderr "radec_to_chip: focxy_to_chip fails "
    return 1
  }

  return 0
}




#############################################################################
# PROCEDURE: chip_to_focxy
#
# DESCRIPTION: Given a chip ID and (row, col) pixel position, 
#              calculate the (x, y) in mm on the focal plane.
#              Place the x and y values into the final two output args.
#
# RETURNS:
#   0          if all goes well
#   1          if an error occurs
#
proc chip_to_focxy { chip_id row col o_x o_y } {
  global debug
  upvar $o_x x $o_y y 
  
  if { $debug > 0 } {
    puts "entering chip_to_focxy"
  }

  set deg2rad [expr 3.14159/180.0]

  # get the name of the param file with chip information
  set chipfile_name [get_chipfile_name]
  if { $chipfile_name == {} } { 
    puts stderr "chip_to_focxy: get_chipfile_name returns NULL"
    return 1
  }
 
  # now, we read information from the chip param file
  set name [format "chip_%03d" $chip_id]
  set ll [get_param $chipfile_name $name]
  if { $ll == {} } {
    puts stderr "chip_to_focxy: can't read info for $name ?!"
    return 1
  }
  if { $debug > 5 } {
    puts "ll is ..$ll.. "
  }
  set chip_info [lindex $ll 0]

  # this is the central (row, col) on the chip
  set npix [lindex $chip_info 2]
  set row_center [expr $npix/2]
  set col_center [expr $npix/2]
  if { $debug > 5 } {
    puts "row_center $row_center  col_center $col_center"
  }

  # calculate the distance in pixels between given (row, col) and chip center
  set delta_x [expr $col-$col_center]
  set delta_y [expr $row-$row_center]
  set r [expr sqrt($delta_x*$delta_x + $delta_y*$delta_y)]

  # convert the distance "r" from pixel to mm 
  set pixscale [lindex $chip_info 3]
  set r_mm [expr $r*$pixscale]
  if { $debug > 5 } {
    puts "r_mm $r_mm"
  }

  # this is the location of the chip center in the focal plane (in mm)
  set x_center [lindex $chip_info 5]
  set y_center [lindex $chip_info 6]
  if { $debug > 5 } {
    puts "x_center $x_center  y_center $y_center"
  }

  # calculate the angle relative to chip center
  if { $r == 0.0 } {
    set gamma 0
  } else {
    set gamma [expr atan2($delta_y, $delta_x)]
  }
  if { $debug > 5 } { 
    puts "gamma $gamma"
  }

  # now add the angle of chip orientation 
  #   (after converting to radians)
  set chip_angle [lindex $chip_info 4]
  set chip_angle_rad [expr $chip_angle*$deg2rad]
  set theta [expr $gamma+$chip_angle_rad]
  if { $debug > 5 } {
    puts "theta $theta"
  }

  # finally, we can calculate row and col positions
  set x [expr $x_center + ($r_mm*cos($theta))]
  set y [expr $y_center + ($r_mm*sin($theta))]
  if { $debug > 5 } {
    puts "x $x  y $y "
  }

  # and we're done
  return 0
}


#############################################################################
# PROCEDURE: focxy_to_chip
#
# DESCRIPTION: Given (x, y) in mm on the focal plane, figure out if the
#              position lies within a detector chip.  If yes, place the
#              chip ID and (row, col) values into the output args.
#              If not, place -1 in all output args.
#
# RETURNS:
#   0          if all goes well
#   1          if an error occurs
#
proc focxy_to_chip { x y o_chipid o_row o_col } {
  global debug
  upvar $o_chipid chipid $o_row row $o_col col
  
  if { $debug > 0 } {
    puts "entering focxy_to_chip"
  }

  set deg2rad [expr 3.14159/180.0]

  # get the name of the param file with chip information
  set chipfile_name [get_chipfile_name]
  if { $chipfile_name == {} } { 
    puts stderr "focxy_to_chip: get_chipfile_name returns NULL"
    return 1
  }
 
  # now, we read information from the chip param file
  set num_chips [get_param $chipfile_name "num_chips"]
  if { $num_chips < 1 } {
    puts stderr "focxy_to_chip: invalid number of chips ?!"
    return 1
  }
  for { set i_chip 0 } { $i_chip < $num_chips } { incr i_chip } {
    set name [format "chip_%03d" $i_chip]
    set ll [get_param $chipfile_name $name]
    set chip_info [lindex $ll 0]

    # does the given position fall within this chip?
    if { [is_inside_chip $x $y $chip_info] == 1 } {

       set chipid $i_chip
 
       # yes, it is inside this chip.  So we need to perform a coordinate
       #   transformation

       # this is the location of the chip center in the focal plane
       set x_center [lindex $chip_info 5]
       set y_center [lindex $chip_info 6]
       if { $debug > 5 } {
         puts "x_center $x_center  y_center $y_center"
       }

       set delta_x [expr $x-$x_center]
       set delta_y [expr $y-$y_center]
       set r [expr sqrt($delta_x*$delta_x + $delta_y*$delta_y)]
       if { $debug > 5 } {
         puts "delta_x $delta_x  delta_y $delta_y  r $r"
       }
    
       # calculate the angle relative to chip center
       if { $r == 0.0 } {
         set gamma 0
       } else {
         set gamma [expr atan2($delta_y, $delta_x)]
       }
       if { $debug > 5 } { 
         puts "gamma $gamma"
       }

       # now subtract the angle of chip orientation 
       #   (after converting to radians)
       set chip_angle [lindex $chip_info 4]
       set chip_angle_rad [expr $chip_angle*$deg2rad]
       set theta [expr $gamma-$chip_angle_rad]
       if { $debug > 5 } {
         puts "theta $theta"
       }

       # convert the distance "r" from mm to pixel units
       set pixscale [lindex $chip_info 3]
       set r_pix [expr $r/$pixscale]
       if { $debug > 5 } {
         puts "r_pix $r_pix"
       }

       # what is the (row, col) position of the center of this chip?
       set npix [lindex $chip_info 2]
       set row_center [expr $npix/2]
       set col_center [expr $npix/2]
       if { $debug > 5 } {
         puts "row_center $row_center  col_center $col_center"
       }
  
       # finally, we can calculate row and col positions
       set row [expr $row_center + ($r_pix*sin($theta))]
       set col [expr $col_center + ($r_pix*cos($theta))]
       if { $debug > 5 } {
         puts "row $row  col $col "
       }

       # and we're done
       return 0
    }
  }

  # nope, the given position doesn't fall inside any chip
  set chipid -1
  set row -1
  set col -1

  return 0
}

#############################################################################
# PROCEDURE: chip_to_focxy_and_back
#
# DESCRIPTION: Test the routines which take (chip, row, col) -> (x, y)
#              and then back again.  
#
#              This routine used only in testing, obviously.
#
# RETURNS
#     0        if all tests pass
#     1        if any test fails
#
proc chip_to_focxy_and_back { } {
  global debug

  if { $debug > 0 } {
    puts "entering chip_to_focxy_and_back"
  }

  # we allow the initial and final chip (row, col) and Dec to differ by at most
  #   this much (pixels)
  set epsilon 0.01

  set retcode 0

  # We run a number of tests, one for each chip in the chip database.
  #   On each chip, we try one position.
  set chipfile_name [get_chipfile_name]
  set num_chips [get_param $chipfile_name "num_chips"]
  set i_chip 0
  while { $i_chip < $num_chips } {
  
    foreach row { 10 2000 } {
      foreach col { 10 2000 } {

        if { $debug > 0 } {
          puts [format "test: chip %3d  row %4.0f col %4.0f " \
                  $i_chip $row $col ]
        }
        if { [test_one_chip_conversion $i_chip $row $col $epsilon] != 0 } {
          puts stderr "chip_to_focxy_and_back: test fails for $i_chip $row $col "
          incr retcode
        } else {
          if { $debug > 0 } {
            puts "   passes ok "
          }
        }

      }
    }

    if { $debug > 10 } {
      puts "hit return to continue "; set aaa [gets stdin]
    }

    incr i_chip
  }

  return $retcode

}


#############################################################################
# PROCEDURE: test_one_chip_conversion
#
# DESCRIPTION: Given a single (chip, row, col) position, 
#              convert it to focal plane (x, y) coordinates, 
#              and then convert those back into (chip, row, col) again.
#              Compare the initial and final chip ID.  Compare the
#              initial and final row and col positions.  If they
#              differ by more than "epsilon",
#              print an error message and return with positive error code:
#              the number of errors which occur in the conversion.
#
# RETURNS
#     0        if the initial and final positions are close enough
#     > 0      if the positions differ
#
proc test_one_chip_conversion { chip row col epsilon } {
  global debug

  if { $debug > 0 } {
    puts "test_one_chip_conversion "
  }

  set retcode 0

  if { [chip_to_focxy $chip $row $col x y] != 0 } {
    puts stderr "test_one_chip_conversion: chip_to_focxy fails on $chip $row $col"
    incr retcode 
  }  
  if { [focxy_to_chip $x $y final_chip final_row final_col ] != 0 } {
    puts stderr "test_one_chip_conversion: focxy_to_chip fails on $x $y "
    incr retcode 
  }  
  if { $chip != $final_chip } {
    puts stderr "test_one_chip_conversion: chip $chip != final $final_chip "
    incr retcode
  }
  set drow [expr $final_row-$row]
  set dcol [expr $final_col-$col]
  if { [expr abs($drow)-$epsilon] > 0 } {
    puts stderr "test_one_chip_conversion: drow $drow > eps $epsilon "
    incr retcode
  }
  if { [expr abs($dcol)-$epsilon] > 0 } {
    puts stderr "test_one_chip_conversion: ddec $dcol > eps $epsilon "
    incr retcode
  }
 
  return $retcode
}


###########################################################################
# PROCEDURE: chip_to_radec
#
# DESCRIPTION: Given a (chip, row, col) position, figure out the (RA, Dec)
#              equivalent.  Place the RA and Dec into the final two
#              arguments.
#
# RETURNS
#    0         if all goes well
#    1         if an error occurs
#
proc chip_to_radec { chip row col o_ra o_dec } {
  global debug
  upvar $o_ra ra $o_dec dec

  if { $debug > 0 } {
    puts "entering chip_to_radec"
  }

  if { [chip_to_focxy $chip $row $col x y] != 0 } {
    puts "chip_to_radec: chip_to_focxy fails for $chip $row $col "
    return 1
  }
  if { [focxy_to_radec $x $y ra dec] != 0 } {
    puts "chip_to_radec: focxy_to_radec fails for $x $y"
    return 1
  }

  return 0
}





###########################################################################
# DESCRIPTION: is_inside_chip
#
# PROCEDURE: Given a (x, y) in mm position on the focal plane, and a list
#            with information about one particular chip, figure out if
#            the position lies inside the chip's boundaries.  
#
#            The current version can handle only chips which have 
#            orientation angles of 0, 90, 180, or 270 -- which make
#            the inside/outside question easy to answer.  If I ever
#            care about chips which are tilted at other angles, I'll
#            have to add more sophisticated code to deal with them.
#
# RETURNS:
#    0       if position is NOT inside the chip
#    1       if position is     inside the chip
#
proc is_inside_chip { x y chip_info } {
  global debug
  
  if { $debug > 2 } {
    puts "entering is_inside_chip"
  }

  set chip_id [lindex $chip_info 0]
  set chip_angle [lindex $chip_info 4]
  if { ($chip_angle != 0) && ($chip_angle != 90) && ($chip_angle != 180) && \
       ($chip_angle != 270) } { 
    puts stderr "is_inside_chip: can't handle orientation angle $chip_angle"
    exit 1
  }

  # find the min and max X and Y values of the chip's corners
  set minx [lindex $chip_info 7]
  set maxx $minx
  set miny [lindex $chip_info 8]
  set maxy $miny
  foreach corner { 1 2 3 } {
    set this_x [lindex $chip_info [expr 7+(2*$corner)] ]
    set this_y [lindex $chip_info [expr 8+(2*$corner)] ]

    if { $this_x < $minx } {
      set minx $this_x 
    }
    if { $this_x > $maxx } {
      set maxx $this_x 
    }
    if { $this_y < $miny } {
      set miny $this_y 
    }
    if { $this_y > $maxy } {
      set maxy $this_y 
    }
  }

  if { $debug > 3 } {
    puts [format "   chip %3d has x %8.3f %8.3f   y %8.3f %8.3f " \
                $chip_id $minx $maxx $miny $maxy ]
  }

  # okay, now we can simply compare the (x, y) position with these extremes
  if { ($x >= $minx) && ($x <= $maxx) && ($y >= $miny) && ($y <= $maxy) } {
    return 1
  }

  return 0
}


###########################################################################
# PROCEDURE: focxy_to_incidence
#
# DESCRIPTION: Given an (x, y) position on the focal plane (in mm),
#              calculate the angle away from normal at which light 
#              will strike the detectors.  Express the angle in radians.
#              Place this angle into the output arg.
#
# RETURNS
#    0         if all goes well
#    1         if an error occurs
#
proc focxy_to_incidence { x y o_angle } {
  global debug
  upvar $o_angle angle

  if { $debug > 0 } {
    puts "entering focxy_to_incidence"
  }

  set r [expr sqrt($x*$x + $y*$y)]

  # this formula is based upon a simple linear fit to the 
  #   graph shown in Alex Kim's document "Effective Filters", 
  #   available on the SNAP Calibration hypertext archive, 
  #   dated Nov 1, 2002
  set ll [get_incidence_info]
  set theta_inner [lindex $ll 2]
  set theta_slope [lindex $ll 3]
  set r_inner [lindex $ll 0]
  set angle [expr $theta_inner + ($theta_slope*($r - $r_inner))]

  return 0
}



###########################################################################
# PROCEDURE: create_sensitivity_params
#
# DESCRIPTION: Generate a set of numbers which describe the variations
#              in sensititivity across a single detector chip.  
#              The general model is Reciprocal Quantum Efficiency given by
#
#                   RQE  =   a  + b*x  + c*y  + d*x*x  + e*x*y  + f*y*y
#
#              where 
#                           x  =  (row - row0)
#                           y  =  (col - col0)
#
#              When we want to calculate the sensitivity of a particular
#              pixel location, we use 
#
#                     sensitivity = 1/RQE
#
#              So, this routine will come up with 8 values:
#
#                         row0, col0, a, b, c, d, e, f
#
#              We return these in a list as an output argument.
#
#              The input argument "chip_variation" tells us how to 
#              generate these numbers.  There are several possibilities:
#
#                     none     means    no variation across any chip
#
#                     center   means    identical radially symmetric decrease
#                                           in sensitivity away from center
#                                           in all chips
#
#                     linear   means    identical 1-D linear decrease in 
#                                           sensitivity in all chips
#
#                     random   means    random values for all 8 parameters
#                                           so each chip is different
#
# RETURNS
#    0         if all goes well
#    1         if an error occurs
#
proc create_sensitivity_params { chip_variation npix o_param_list } {
  global debug
  upvar $o_param_list param_list

  if { $debug > 0 } {
    puts "entering create_sensitivity_params"
  }

  # initialize to the empty list as a backup signal if failure occurs
  set param_list {}
  # we set the fixed coefficient to 1.0 in all options
  set a 1.0

  # these values for the coefficients will lead to variations of a few
  #   percent from one side of the chip to another
  set amplitude_percent 20.0
  set linear_coeff [expr ($amplitude_percent*0.01)/($npix)]
  set quadratic_coeff [expr ($amplitude_percent*0.01)/($npix*$npix)]

  if { [string compare $chip_variation "none"] == 0 } {
    set center_row [expr $npix/2]
    set center_col [expr $npix/2]
    set b 0.0
    set c 0.0
    set d 0.0
    set e 0.0
    set f 0.0
  } elseif { [string compare $chip_variation "center"] == 0 } {
    set center_row [expr $npix/2]
    set center_col [expr $npix/2]
    set b 0.0
    set c 0.0
    set d $quadratic_coeff
    set e 0.0
    set f $quadratic_coeff
  } elseif { [string compare $chip_variation "linear"] == 0 } {
    set center_row [expr $npix/2]
    set center_col [expr 0.0]
    set b 0.0
    set c $linear_coeff
    set d 0.0
    set e 0.0
    set f 0.0
  } elseif { [string compare $chip_variation "random"] == 0 } {
    set center_row [expr ($npix/4) + rand()*($npix/2)]
    set center_col [expr ($npix/4) + rand()*($npix/2)]
    set b [expr rand()*$linear_coeff]
    set c [expr rand()*$linear_coeff]
    set d [expr rand()*$quadratic_coeff]
    set e [expr rand()*$quadratic_coeff]
    set f [expr rand()*$quadratic_coeff]
  } else {
    # this value wasn't recognized -- oops
    puts stderr "create_sensitivity_params: bad keyword ..$chip_variation.."
    return 1
  }

  # make this look nice ...
  set center_row [format "%6.1f" $center_row] 
  set center_col [format "%6.1f" $center_col] 
  set a [format "%10.4e" $a] 
  set b [format "%10.4e" $b] 
  set c [format "%10.4e" $c] 
  set d [format "%10.4e" $d] 
  set e [format "%10.4e" $e] 
  set f [format "%10.4e" $f] 
  set param_list [list $center_row $center_col $a $b $c $d $e $f]

  if { $debug > 0 } {
    puts "create_sensitivity_params: list is $param_list "
  }

  return 0

}



namespace eval Focal {
  namespace export init_focal get_chipfile_name init_chipnum next_chipnum \
                          get_foc_info get_incidence_info set_foc_radec \
                          get_foc_east_only
  variable Chipfile_Name {}
  variable Chip_Number {}
  variable Scale_Factor {}
  variable Roll_Angle {}
  variable Center_RA {}
  variable Center_Dec {}
  variable Annulus_Inner {}
  variable Incidence_Inner {}
  variable Incidence_Slope {}
  variable East_Quadrant_Only {}

#######################################################################
# Set the name of the new chip param file we'll create from a high-level
#   description. 
#   Also set several factors which we read from the "focal.param" file
#   and will want to use later on.
#
proc init_focal { param_file } {
  global debug
  variable Chipfile_Name 
  variable Scale_Factor
  variable Roll_Angle
  variable Center_RA
  variable Center_Dec
  variable Annulus_Inner 
  variable Incidence_Fiducial 
  variable Incidence_Inner 
  variable Incidence_Slope
  variable East_Quadrant_Only 

  if { $debug > 0 } {
    puts "entering init_focal "
  }

  set param_dir [get_param $param_file "param_dir"]
  set foc_output_file [get_param $param_file "foc_output_file"]
  set Chipfile_Name [format "%s/%s" $param_dir $foc_output_file]

  set Scale_Factor [get_param $param_file "foc_scale_factor"]
  set Roll_Angle   [get_param $param_file "foc_roll_angle"]
  set Center_RA    [get_param $param_file "foc_center_ra"]
  set Center_Dec   [get_param $param_file "foc_center_dec"]

  # we need these to calculate angle at which light strikes focal plane
  #    in the "focxy_to_incidence" procedure
  set Annulus_Inner [get_param $param_file "foc_annulus_inner"]
  set Incidence_Fiducial [get_param $param_file "foc_incidence_fiducial"]
  set Incidence_Inner [get_param $param_file "foc_incidence_inner"]
  set Incidence_Slope [get_param $param_file "foc_incidence_slope"]

  # if the "foc_east_only" parameter is set, we place detectors only
  #   on the East Quadrant of the focal plane;  if not, we place detectors
  #   all over the focal plane, in East, North, West, South quadrants
  set East_Quadrant_Only [get_param $param_file "foc_east_only"]

  
  return 0
}


#######################################################################
# Set the telescope's optical axis to point to the given (RA, Dec) position
#
# RETURNS:
#    0        if all goes well
#
proc set_foc_radec { ra dec } {
  global debug
  variable Center_RA
  variable Center_Dec

  if { $debug > 0 } {
    puts "entering set_foc_radec "
  }

  set Center_RA $ra
  set Center_Dec $dec

  return 0
}


#######################################################################
# Return a list which contains several pieces of information about
#   the focal plane, like so:
#
#      { scale_factor  roll_angle  center_ra  center_dec }
#
proc get_foc_info { } {
  global debug
  variable Scale_Factor
  variable Roll_Angle
  variable Center_RA
  variable Center_Dec

  if { $debug > 0 } {
    puts "entering get_foc_info "
  }

  set ll [list $Scale_Factor $Roll_Angle $Center_RA $Center_Dec]
  return $ll
}

#######################################################################
# Return a list which contains several pieces of information which
#   we can use to calculate angle of incidence.  The pieces are
#
#      foc_annulus_inner  
#      foc_incidence_fiducial 
#      foc_incidence_inner  
#      foc_incidence_slope 
#
proc get_incidence_info { } {
  global debug
  variable Annulus_Inner 
  variable Incidence_Fiducial 
  variable Incidence_Inner 
  variable Incidence_Slope 

  if { $debug > 0 } {
    puts "entering get_incidence_info "
  }

  set ll [list $Annulus_Inner $Incidence_Fiducial \
                            $Incidence_Inner $Incidence_Slope]
  return $ll
}


 
######################################################################
# Return the name of the (generatd) chip param file 
#
proc get_chipfile_name { } {
  global debug
  variable Chipfile_Name

  if { $debug > 0 } {
    puts "entering get_chipfile_name"
  }
  return $Chipfile_Name
}


#######################################################################
# Initialize the number of chips we've placed into the chip param file
#   to zero
#
proc init_chipnum { } {
  global debug
  variable Chip_Number 

  if { $debug > 0 } {
    puts "entering init_chipnum "
  }
  set Chip_Number 0

  return 0
}

######################################################################
# Return the ID number of the next available chip in focal plane
#
proc next_chipnum { } {
  global debug
  variable Chip_Number

  if { $debug > 0 } {
    puts "entering next_chipnum"
  }
  
  set retval $Chip_Number
  incr Chip_Number

  return $retval
}


######################################################################
# Return the value of the "foc_east_only" parameter, which should be
#      0   if we place detectors all over the focal plane (all 4 quadrants)
#      1   if we place detectors only on the East quadrant
#
proc get_foc_east_only { } {
  global debug
  variable East_Quadrant_Only

  if { $debug > 0 } {
    puts "entering get_foc_east_only"
  }
  return $East_Quadrant_Only
}


  # end of namespace Focal
}


