# These routines act on the "file list", an ASCII file with one line
#   per image taken by the Mark IV.  Each line looks something
#   like this:
#   
#          name    filter  JD  exposure_time   type   ra  dec  
#
# MWR 5/20/2000
#
# Routines herein:
#   
#   init_list_name             sets a "static" variable within this
#                                   namespace to the name of a file
#                                   containing information about all
#                                   the images in this run.  Must be
#                                   called before any other proc in 
#                                   this file
# 
#   get_list_info              return a list of values for one parameter
#                                   such as all the types of filter, 
#                                   or all the different exposure times
#                                   
#   select_input_list          return a list of image names, for images
#                                   with parameter(s) which match 
#                                   given values:
#                                      "all V-band images
#                                      "all V-band and 'dark' images"
#
#   single_image_info          return a list of pairs: { keyword value }
#                                   with information about a single image,
#                                   given the image's name
#
#   get_image_value            given a list of pairs: { keyword value }
#                                   and a keyword, return the value
#                                   corresponding to the keyword
#
#                                   
# This file contains a namespace "Proc_File_List", 
#   to hold some private variables.  It allows us to call an initialization
#   routine which reads information from the param file,
#   then save those values so we can call other routines 
#   without having to pass the param file as a parameters again
#
#
# Removed "airmass" from set of values listed in file.
#     --- MWR 1/20/2001   


if { [catch { set debug } ] == 1 } {
  set debug 2
}


namespace eval Proc_File_List {
  namespace export init_list_name get_list_info \
                   select_input_list single_image_info get_image_value
  variable List_File_Name {}


#############################################################################
# Set the "list_file" name.  The "list_file" is a file containing a list
#   of raw image file names, together with properties such as
#   filter, exposure time, etc.
#   
# This proc sets a local "static" variable, List_File_Name,
#   which is used by several other procs within this namespace
# 
proc init_list_name { list_file } {
  global debug
  variable List_File_Name

  if { $debug > 0 } {
    puts "entering init_list_name"
  }
  set List_File_Name $list_file

  return 0
}


#############################################################################
# Parse the "image list" file (created by make_list)
#   to find all the images of a particular sort.  
#   Note that "init_list_name" must have been called before this proc.
#   
# usage:  get_list_info    info_list info_type
#
#               info_list          is a variable we will set to the 
#                                       list of name, or filters, or etc.
#                                       encountered among the images
#                                       in list_file
#
#               info_type          is   "name"      to gather list of names
#                                       "filter"    to gather list of filters
#                                       "jd"        to gather list of Julian
#                                                       date values
#                                       "exposure"  to gather list of 
#                                                       exposure times
#                                       "type"      to gather list of types
#                                       "ra"        to gather list of RA vals
#                                       "dec"       to gather list of Dec vals
#                                                       
#                                       
# Returns
#    0            if all goes well
#    1            if there's a problem
#
proc get_list_info { info_list info_type } {
  upvar $info_list i_list
  global debug
  variable List_File_Name

  if { $debug > 0 } {
    puts "entering get_list_info"
  }

  set list_file $List_File_Name
  if { [catch { set fid [open $list_file "r"] } ] != 0 } {
    puts stderr "get_list_info: can't open file $list_file"
    return 1
  }


  # we walk through the list file, expecting lines of the form
  #
  #          name    filter  JD  exposure_time   type   ra  dec  
  #          
  #   skip any lines which start with a pound sign
  #   skip (and print warning) for any lines which don't have the 
  #      expected format
  #      
  set i_list ""
  while { [gets $fid line] != -1 } {

    # skip comments
    if { [regexp \#.* $line] == 1 } {
      continue
    }

    # split line into pieces, expect exactly 7 words
    if { [llength $line] != 7 } {
      puts stderr "get_list_info: bad line in file $list_file, will skip"
      puts stderr $line
      continue
    }

    # okay, this looks like a good entry
    #   parse it and pick out the desired parameter value
    if { [string compare $info_type "name"] == 0 } {
      set this_item [lindex $line 0]
      if { $debug > 1 } {
        puts "    this name is $this_item"
      }
    } elseif { [string compare $info_type "filter"] == 0 } {
      set this_item [lindex $line 1]
      if { $debug > 1 } {
        puts "    this filter is $this_item"
      }
    } elseif { [string compare $info_type "jd"] == 0 } {
      set this_item [lindex $line 2]
      if { $debug > 1 } {
        puts "    this jd is $this_item"
      }
    } elseif { [string compare $info_type "exposure"] == 0 } {
      set this_item [lindex $line 3]
      if { $debug > 1 } {
        puts "    this exposure is $this_item"
      }
    } elseif { [string compare $info_type "type"] == 0 } {
      set this_item [lindex $line 4]
      if { $debug > 1 } {
        puts "    this type is $this_item"
      }
    } elseif { [string compare $info_type "ra"] == 0 } {
      set this_item [lindex $line 5]
      if { $debug > 1 } {
        puts "    this RA is $this_item"
      }
    } elseif { [string compare $info_type "dec"] == 0 } {
      set this_item [lindex $line 6]
      if { $debug > 1 } {
        puts "    this Dec is $this_item"
      }
    } 

    if { [lsearch $i_list $this_item] == -1 } {
      lappend i_list $this_item
      if { $debug > 1 } {
        puts "    adding to list, now it is $i_list"
      }
    }

  }

  # all done
  close $fid

  if { $debug > 1 } {
    puts "get_list_info: returning list $i_list "
  }

  return 0
}



#############################################################################
# Parse the "image list" file, created by make_list.
#   Given specific value(s) for each parameter(s) of interest,
#   pick out the images in the list which have the given specific
#   value(s)
#   Note that "init_list_name" must have been called before this proc.
#   
# usage:  select_input_list    param_pair_list 
#
#                                       
#               param_pair_list    is a list of pairs of things.  Each pair
#                                       consists of a parameter name 
#                                       (such as "filter" or "type")
#                                       and an associated value.  
#                                       Examples:
#                                       
#                                           { { filter V } }
#                                       or     
#                                           { { filter V } { exposure 60 } }
#                                       
#                                       
# Returns
#    a list of the appropriate image files
#    {}                   if given bad input
#
proc select_input_list { param_pair_list } {
  global debug
  variable List_File_Name

  if { $debug > 0 } {
    puts "entering select_input_list"
  }

  set list_file $List_File_Name
  if { [catch { set fid [open $list_file "r"] } ] != 0 } {
    puts stderr "select_input_list: can't open file $list_file"
    return {}
  }

  # we walk through the list file, expecting lines of the form
  #
  #          name    filter  JD   exposure_time   type   ra  dec  
  #          
  #   skip any lines which start with a pound sign
  #   skip (and print warning) for any lines which don't have the 
  #      expected format
  #      
  set the_input_list ""
  while { [gets $fid line] != -1 } {

    # skip comments
    if { [regexp \#.* $line] == 1 } {
      continue
    }

    # split line into pieces, expect exactly 7 words
    if { [llength $line] != 7 } {
      puts stderr "select_input_list: bad line in file $list_file, will skip"
      puts stderr $line
      continue
    }

    if { $debug > 1 } {
      puts "   next is $line"
    }

    set this_name     [lindex $line 0]
    set this_filter   [lindex $line 1]
    set this_jd       [lindex $line 2]
    set this_exposure [lindex $line 3]
    set this_type     [lindex $line 4]
    set this_ra       [lindex $line 5]
    set this_dec      [lindex $line 6]

    # by default, we will add this entry to the list (so set 'addit' to 1),
    #   but we are going to check each provided parameter, and if any
    #   don't match the provided value, we disqualify this entry
    #   (by setting 'addit' to 0)
    #    
    set addit 1
    foreach pair $param_pair_list {

      if { [llength $pair] != 2 } {
        puts stderr "select_input_list: bad param pair $pair"
        return {}
      }
      set this_param [lindex $pair 0]
      set this_value [lindex $pair 1]

      # now check to see if this line meets the condition(s)
      #   if it does NOT, we set "addit" to 0
      if { [string compare $this_param "name"] == 0 } {
        if { [string compare $this_name $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because name $this_name not $this_value"
          }
        } 
      }
      if { [string compare $this_param "filter"] == 0 } {
        if { [string compare $this_filter $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because filter $this_filter not $this_value"
          }
        } 
      }
      if { [string compare $this_param "jd"] == 0 } {
        if { [string compare $this_jd $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because jd $this_jd not $this_value"
          }
        } 
      }
      if { [string compare $this_param "exposure"] == 0 } {
        if { [string compare $this_exposure $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because exposure $this_exposure not $this_value"
          }
        } 
      }
      if { [string compare $this_param "type"] == 0 } {
        if { [string compare $this_type $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because type $this_type not $this_value"
          }
        } 
      }
      if { [string compare $this_param "ra"] == 0 } {
        if { [string compare $this_ra $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because ra    $this_ra not $this_value"
          }
        } 
      }
      if { [string compare $this_param "dec"] == 0 } {
        if { [string compare $this_dec $this_value] != 0 } {
          set addit 0
          if { $debug > 1 } {
            puts "     disqual because dec  $this_dec not $this_value"
          }
        } 
      }
    }

    # if this still looks like a good entry, add it to the output list
    if { $addit == 1 } {
      lappend the_input_list [lindex $line 0]
      if { $debug > 1 } {
        puts "    adding to list, now it is $the_input_list"
      }
    }

  }

  # all done
  close $fid

  if { $debug > 1 } {
    puts "select_input_list: returning list $the_input_list "
  }

  return $the_input_list
}



#############################################################################
# Parse the "image list" file, created by make_list, 
#   Given one specific image name,
#   look for that name in the image list and (if found) return 
#   information about the image: name, filter, etc.
#   
# usage:  single_image_info    name  value_pair_list
#
#                                       
#               name               is the name of the image of interest
#
#               value_list         will be set on output to a
#                                  list of pairs of keywords and values:
#                                       { name "foobar.fts" } { filter "V" }
#                                  
#                                       
# Returns
#    0                    if all goes well
#    1                    if there's a problem
#
proc single_image_info { name value_list } {
  upvar $value_list v_list 
  global debug
  variable List_File_Name

  if { $debug > 0 } {
    puts "entering single_image_info"
  }

  set list_file $List_File_Name
  if { [catch { set fid [open $list_file "r"] } ] != 0 } {
    puts stderr "single_image_info: can't open file $list_file"
    return 1
  }

  # we walk through the list file, expecting lines of the form
  #
  #          name    filter  JD  exposure_time   type   ra  dec  
  #          
  #   skip any lines which start with a pound sign
  #   skip (and print warning) for any lines which don't have the 
  #      expected format
  #      
  set v_list ""
  while { [gets $fid line] != -1 } {

    # skip comments
    if { [regexp \#.* $line] == 1 } {
      continue
    }

    # split line into pieces, expect exactly 7 words
    if { [llength $line] != 7 } {
      puts stderr "single_image_info: bad line in file $list_file, will skip"
      puts stderr $line
      continue
    }

    if { $debug > 1 } {
      puts "   next is $line"
    }

    set this_name     [lindex $line 0]
    set this_filter   [lindex $line 1]
    set this_jd       [lindex $line 2]
    set this_exposure [lindex $line 3]
    set this_type     [lindex $line 4]
    set this_ra       [lindex $line 5]
    set this_dec      [lindex $line 6]

    if { [string compare $this_name $name] == 0 } {
      
      # found it!  Build the list
      lappend v_list [list name $this_name ]
      lappend v_list [list filter $this_filter ]
      lappend v_list [list jd $this_jd ]
      lappend v_list [list exposure $this_exposure ]
      lappend v_list [list type $this_type ]
      lappend v_list [list ra $this_ra ]
      lappend v_list [list dec $this_dec ]
     
      # no need to check other images in the list
      break
    }

  }


  # all done
  close $fid

  if { $debug > 1 } {
    puts "single_image_info: returning list $v_list "
  }

  return 0
}


#############################################################################
# Given a list of paired items describing an image, look for a particular
#   keyword and return the corresponding value.
#   
# usage:    get_image_value   image_info_list  keyword
# 
# where
#               image_info_list       is a list of pairs of items, like this:
#                                        { name "foobar" } { filter "V" }
#                                        
#               keyword               is the keyword of desired value
#               
# Returns
#   the desired value            if the keyword exists
#   terminates                   if keyword not found
#  
proc get_image_value { image_info_list keyword } {
  global debug

  if { $debug > 0 } {
    puts "entering get_image_value"
  }

  foreach pair $image_info_list {
    
    if { [llength $pair] != 2 } {
      puts stderr "get_image_value: pair $pair has length != 2"
      puts stderr " ABORTING"
      exit 1
    }
    set this_keyword [lindex $pair 0]
    set this_value [lindex $pair 1]
    if { [string compare $keyword $this_keyword] == 0 } {
      return $this_value
    }

  }

  # if we get here, was a bad keyword, not in list
  puts stderr "get_image_value: no keyword $keyword in following list"
  puts stderr "   $image_info_list"
  puts stderr " ABORTING"
  exit 1
}


 # end of the namespace Proc_File_List
}
