#
# This TCL script converts a "real" catalog of stellar positions
#   and colors into an "artificial" catalog which contains
#   
#           a) RA and Dec, identical to the real stars'
#           b) spectral type, based on real colors
#           c) magnitudes and auxiliary information for each
#                 star in each of the SNAP filters
#
# It writes a file to disk in the output directory which contains
#   self-consistent spectral types and magnitudes in the natural
#   SNAP system.
#
# MWR 12/03/2002
#    
# Modified "create_color_table so it expects to see (B-R) color
#   at the end of every input line, after the UBVRI magnitudes.
#   Helps with debugging.
#
# MWR 9/5/2003
#
# Modified "get_stellar_info" to make special check for SN, using an
#   otherwise ignored field in the line for a catalog entry
# 
# MWR 9/11/2003


if { [catch { set debug } ] == 1 } {
  set debug 2
}

source param.tcl


###########################################################################
# Read information from a big catalog.  Select only stars which are
#   with the boundaries of images taken on a particular night.
#   Spit out a subset of the the big catalog into a small
#   "reference catalog" we'll use for astrometry and photometry.
#
#
# RETURNS
#    0           if all goes well
#    1           if there's a problem
#    
proc refcat { param_file } {
  global debug

  if { $debug > 0 } {
    puts "entering refcat"
  }

  # get information from the param file
  if { [file exists $param_file] != 1 } {
    puts stderr "refcat: can't open param file $param_file"
    return 1
  }
  set input_dir [get_param $param_file "input_dir"]
  set refcat_name [get_param $param_file "refcat_name"]
  set color_table [get_param $param_file "color_table"]
  set canonical_filter [get_param $param_file "canonical_filter"]
  set convolve_program [get_param $param_file "convolve_program"]
  set output_dir [get_param $param_file "output_dir"]
  set output_cat [get_param $param_file "output_cat"]

  
  # open the input catalog
  set file [get_refcat_name]
  if { [file exists $file] != 1 } {
    puts stderr "refcat: input file $file doesn't exist -- quitting"
    return 1
  }
  set ref_fileid [open $file "r"]
  if { [catch { set ref_fileid } ] == 1 } {
    puts stderr "refcat: error opening input file $file -- quitting"
    return 1
  }

  # open the table with magnitudes for stars of all spectral types,
  #   and read it into a big list-of-lists
  set color_file $input_dir/$color_table
  if { [file exists $color_file] != 1 } {
    puts stderr "refcat: can't open color table $color_file -- quitting"
    return 1
  }
  if { [create_color_table $color_file color_table] != 0 } {
    puts stderr "refcat: create_color_table fails -- quitting"
    return 1
  }


  # get the list of filters we'll need, and verify that we can find
  #   the file for each one.  
  set num_fiducial [get_filter_num_fiducial]
  set useless [ catch { unset filter_fileids } ]
  for { set fiducial_index 0 } { $fiducial_index < $num_fiducial } \
               { incr fiducial_index } {
    set filter_file [get_filter_fiducial_name $fiducial_index]
    if { [file exists $filter_file] != 1 } {
      puts stderr "refcat: filter file $filter_file doesn't exist -- quitting"
      return 1
    }
  }

  # we need to set the canonical filter, which we use to calculate 
  #    the offset which takes synthetic mag to "real" mag.
  #    We use the fiducial filter whose index 
  #    is set in the "canonical_filter" parameter.
  #    This is a sanity check
  if { ($canonical_filter < 0) || ($canonical_filter >= $num_fiducial) } {
    puts stderr "refcat: invalid canonical index $canonical_filter ?!"
    return 1
  }
  

  # open the output file
  set output_name [get_outcat_name]
  set output_fid [open $output_name "w"]
  if { $output_fid < 0 } {
    puts stderr "refcat: can't open output file $output_name"
  }

 
  # Now, we loop through the catalog of stars.  For each star, we
  #    a) read in information from catalog
  #    b) decide its spectral type
  #    c) find the spectrum corresponding to it
  #    d) start building output line
  #    
  #    For the canonical filter (probably V-like)
  #          e1) convolve spectrum with filter to calc raw mag
  #          e2) calc offset to turn raw mag into catalog mag
  #                   catalog_mag =  raw_mag + offset
  #          e3) add offset and doublet (filter, catalog_mag) to output line
  #
  #    For all other filters, 
  #          f1) convolve spectrum with filter to calc raw mag
  #          f2) add doublet (filter, catalog_mag) to output line
  #
  #    g) write output line to output file
  # 

  while { [gets $ref_fileid line] > 0 } {

    if { $debug > 3 } {
      puts "next line is ..$line.."
    }

    # skip lines starting with the "#" character
    if { [regexp ^# $line] == 1 } {
      continue
    }

    # Steps a+b. read information from the line: RA, Dec, "mag", spec_type
    #   Ignore any line which doesn't look right ...
    if { [read_stellar_info $line $color_table ra dec \
                                          canonical_mag sptype] != 0 } {
      puts stderr "refcat: read_stellar_info fails on line ..$line.. "
      continue
    }

    # Step c.  Find the spectrum corresponding to this star
    if { [get_spectrum_file $input_dir $sptype spectrum_file] != 0 } {
      puts stderr "refcat: get_spectrum_file fails for type $sptype "
      return 1
    }

    # Step d.  Start building output line for this star
    set output_line [format "%10.5f %10.5f  %6s " $ra $dec $sptype]

    # Steps e1+e2.  Figure out the magnitude of this star observed through
    #   the canonical filter, and also figure out the offset which
    #   takes synthetic magnitudes to "real" ones
    if { [calc_magnitude $input_dir $spectrum_file $canonical_filter \
                                  $convolve_program synth_mag] != 0 } {
      puts stderr "refcat: calc_magnitude fails for $canonical_filter "
      return 1
    }
    set offset [expr $canonical_mag-$synth_mag]
    # Step e3: Add offset and doublet for this passband to the output line
    set output_line [format "%s %6.3f   %s %6.3f  " $output_line \
                                $offset $canonical_filter $canonical_mag ]

    # Steps f1+f2:  For all other filters (besides the canonical one),
    #   calculate the synthetic magnitude, offset it, and add to the
    #   output line
    for { set filter_index 0 } { $filter_index < $num_fiducial } \
                               { incr filter_index } {

       if { $filter_index == $canonical_filter } {
         continue
       }

       if { [calc_magnitude $input_dir $spectrum_file $filter_index \
                                  $convolve_program synth_mag] != 0 } {
         puts stderr "refcat: calc_magnitude fails for $filter "
         return 1
       }
       if { [string compare $synth_mag "inf"] == 0 } {
         puts stderr "refcat: calc_magnitude yields 'inf' for $spectrum_file $filter_index"
         return 1
       }
       set mag [expr $synth_mag+$offset]
       set output_line [format "%s  %d %6.3f " $output_line \
                                $filter_index $mag ]
  
       if { $debug > 4 } {
          puts "output_line is now ..$output_line.. "
       }
    }
       
    # write this line to the output file
    puts $output_fid $output_line
  }

  # close all files we've opened
  close $output_fid
  close $ref_fileid


  return 0
}


############################################################################
# PROCEDURE: calc_magnitude
# 
# DESCRIPTION: Convolve a spectrum with a particular passband to get 
#              a synthetic magnitude.
#
#                  input_dir         directory with spectral and filter data
#                  spectrum_file     full pathname of file with spectrum
#                  filter_index      index in set of fiducial filters
#                  convolve_program  full name of executable which convolves
#                                         spectrum with passband
#                  synth_mag         the synthetic mag for star in this filter
#                                         (set on output)
#
#              We assume that the light is striking the filter at the
#              fiducial angle, so that there are no shifts due to 
#              incident angle.
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc calc_magnitude { input_dir spectrum_file filter_index convolve_program \
                                                          o_synth_mag } {
  global debug 
  upvar $o_synth_mag synth_mag 

  if { $debug > 0 } {
    puts "entering calc_magnitude "
  }

  # verify that the spectrum file exists (a sanity check)
  if { [file exists $spectrum_file] != 1 } {
    puts stderr "calc_magnitude: file $spectrum_file doesn't exist?"
    return 1
  }

  # build the name of the filter file, and verify that it exists
  set filter_file [get_filter_fiducial_name $filter_index]
  if { [file exists $filter_file] != 1 } {
    puts stderr "calc_magnitude: file $filter_file doesn't exist?"
    return 1
  }

  # use the "lam_mag" program to calculate the synthetic magnitude
  #   we assume that light is coming at the fiducial angle, 
  #   and so use the same value for "fiducial" and "actual" angles.
  set ll [get_incidence_info]
  set fiducial_angle [lindex $ll 1]
  set incident_angle $fiducial_angle
  set cmd [format "%s %s %s %6.3f %6.3f" $convolve_program $spectrum_file \
                                 $filter_file $fiducial_angle $incident_angle]
  if { $debug > 3 } {
    puts "calc_magnitude: cmd is ..$cmd.."
  }
  set retcode [catch { set result_line [eval exec $cmd] } ]
  if { $retcode != 0 } {
    puts stderr "calc_magnitude: error exec'ing ..$cmd.."
    return 1
  } 
  set synth_mag [lindex $result_line 5]


  return 0
}




##############################################################################
# PROCEDURE: get_spectrum_file
# 
# DESCRIPTION: Given a spectral type, get the full pathname of the file
#              containing this spectrum.  
#
#              Place the full name into the "o_fullname" argument.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc get_spectrum_file { input_dir sptype o_spectrum_file } {
  global debug 
  upvar $o_spectrum_file spectrum_file

  if { $debug > 0 } {
    puts "entering get_spectrum_file "
  }

  # build the spectrum file name, including the path
  set spectrum_file [format "%s/uk%s.dat" $input_dir $sptype]
  
  # verify that the file exists (a sanity check)
  if { [file exists $spectrum_file] != 1 } {
    puts stderr "get_spectrum_file: file $spectrum_file doesn't exist?"
    return 1
  }

  return 0
}



##############################################################################
# PROCEDURE: create_color_table
# 
# DESCRIPTION: Read information from a file which has lines like this:
#
#                   spec_type  U  B  V  R  I   bminusr
#
#              for each spectral type, calculate the (B-R) color, and 
#              generate a little list like this:
#
#                           {  a0v  0.00  }
#
#              Put all these little lists together into a single big
#              list.  Then sort that list by the color values.
#
#              Place the sorted, big list-of-lists into the second argument.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc create_color_table { color_file o_color_table } {
  global debug 
  upvar $o_color_table color_table

  set fid [open $color_file "r"]
  if { $fid < 0 } {
    puts stderr "create_color_table: can't open file $color_file"
    return 1
  }

  set color_table {}

  while { [gets $fid line] > 0 } {

    # we expect a line like this:
    #
    #       sptype  u b v r i
    #
    if { [llength $line] < 6 } {
      puts "create_color_table: not at least 6 words in line ..$line.."
      return 1
    }
    set sptype [lindex $line 0]
    set bmag [lindex $line 2]
    set rmag [lindex $line 4]
    set color [expr $bmag-$rmag]

    set tiny_list [list $sptype $color]
    lappend color_table $tiny_list

    if { $debug > 2 } {
      puts "color_table is now $color_table"
    }
  }
  close $fid

  # sort the table in ascending order of color
  set color_table [lsort -real -index 1 $color_table]
  if { $debug > 2 } {
    puts "color_table is now $color_table"
  }



  return 0
}



##############################################################################
# PROCEDURE: read_stellar_info
# 
# DESCRIPTION: Read information from a line in the input stellar catalog.
#              We will parse it and pick out
#
#                  RA              Right Ascension in decimal degrees (J2000)
#                  Dec             Declination in decimal degrees (J2000)
#                  mag             magnitude in a canonical passband
#                                          (something like V)
#                  sptype          guess at a spectral type. the format
#                                          is like "a0v" or "f6iii".
#
#              These four values are placed into the output args.
#
#         Special cases:
#
#              Supernovae:
#                 In order to handle SNe in addition to the regular stars,
#                 we have special convention: if the second item in the
#                 line (which is "distance-from-center" in USNO A2.0 query)
#                 starts with the letters "sn", then we parse the line in
#                 a special manner, expecting to see
#
#                  SN_type         in item 1
#                  RA              in item 2 (as for regular stars)
#                  Dec             in item 3 (as for regular stars)
#                  bmag            in item 7 (as for regular stars)
#                  rmag            in item 8 (as for regular stars)
#
#                 We still use the "bmag" and "rmag" values to define
#                 an average magnitude for the zeropoint of the star.
#
#              Background sky:
#                 We create pseudo-stars which represent the amount
#                 of light falling within a single square arcsecond.
#                 If the second item in the line is "sky", then
#                 we parse the line expecting:
#
#                  "sky"           in item 1
#                  RA              in item 2 (as for regular stars)
#                  Dec             in item 3 (as for regular stars)
#                  vmag            in item 7 (is "bmag" for regular stars)
#                  vmag            in item 8 (is "rmag" for regular stars)
#
#                 where the "vmag" represents magnitudes per square
#                 arcsecond in the V-band, similar to the SNAP filter 1.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc read_stellar_info { line color_table o_ra o_dec o_mag o_sptype } {
  global debug 
  upvar $o_ra ra $o_dec dec $o_mag mag $o_sptype sptype

  # we expect a line like this:
  #
  #       id dist ra dec name ra dec bmag rmag epoch
  #
  if { [llength $line] != 10 } {
    puts "read_stellar_info: not 10 words in line ..$line.."
    return 1
  }
  set ra [lindex $line 2]
  set dec [lindex $line 3]
  set bmag [lindex $line 7]
  set rmag [lindex $line 8]

  # now, we'll use the average of B and R mags as "canonical mag",
  #   since it will probably be close to V
  set mag [expr 0.5*($bmag + $rmag)]

  # make a special check to see if this a SN; 
  #   if it is, we get the SN spectral type immediately and return
  set sntype [lindex $line 1]
  if { [regexp ^sn $sntype] == 1 } {
    set sptype $sntype
    if { $debug > 0 } {
      puts " special code $sntype yields SN sptype $sptype"
    }
    return 0
  }

  # special check to see if this is a "sky" star
  #   if it is, we get the "sky" spectral type and return
  if { [string compare "sky" $sntype] == 0 } {
    set sptype "sky"
    if { $debug > 0 } {
      puts " special code $sntype yields sky spectrum "
    }
    return 0
  }
    

  # add some "fuzz" to the color so that we sample different portions
  #   of the color table
  # set color_fuzz 0.30
  set color_fuzz 0.0
  set color [expr $bmag - $rmag]
  set color [expr $color + $color_fuzz*(1 - rand()*2)]
  if { $debug > 4 } {
    puts [format "read_stellar_info:  bmag %6.3f  rmag %6.3f  color %6.3f " $bmag $rmag $color]
  }

  # subtract a teeny bit to avoid roundoff errors 
  #   (when we create a list of artificial stars whose colors SHOULD
  #    be exactly the same as the tabulated spectral types, but which
  #    are somehow rounded to be slightly REDDER than the tabulated 
  #    colors ....)
  set color [expr $color-0.001]
  

  # okay, how to turn (B-R) color into spectral type?  We use a
  #   simple lookup table which is sorted by color.  Walk through
  #   the table until we reach the color of this star.
  # if the star has a color which is outside the range of the
  #   table, we assign a random type.
  set sptype {}
  set table_size [llength $color_table]
  set this_index 0
  while { $this_index < $table_size } {
    set this_list [lindex $color_table $this_index]
    set this_color [lindex $this_list 1]
    if { $color <= $this_color } {
      # pick either this star (which is bluer) or the next one (redder)
      if { $this_index == 0 } {
        # this star is bluer than our bluest type.  Accept the bluest type
        set sptype [lindex $this_list 0]
        break
      } else {
        # check to see which is closer in color -- "this" star or the next one
        set this_type [lindex $this_list 0]
        set next_list [lindex $color_table [expr $this_index+1]]
        set next_type [lindex $next_list 0]
        set next_color [lindex $next_list 1]
        set this_diff [expr abs($color-$this_color)]
        set next_diff [expr abs($next_color-$color)]
        if { $this_diff < $next_diff } {
          set sptype $this_type
        } else { 
          set sptype $next_type
        }
      }
       
      # we're done, so break out of loop
      break
    }


    incr this_index
  }
  # if we didn't assign a type, the star's colors are outside the range
  #   of all entries in the table.  In this case, pick a random type
  if { $sptype == {} } {
    set random_index [expr round(rand()*($table_size-1))]
    set this_list [lindex $color_table $random_index]
    set sptype [lindex $this_list 0]
  }

  if { $debug > 0 } {
    puts " color $color yields type $sptype "
  }
  


  return 0
}




##############################################################################
# Given a fileid for the list of observed fields, the columns within
#   that file of RA and Dec values, and the size of the
#   field of view in degrees, figure out the limits of RA and Dec
#   for each observed field.  Create a set of lists:
#
#        min_ra1             (low) min RA of each field  (decimal degrees)
#        max_ra1             (low) max RA of each field  (decimal degrees)
#        min_ra2             (high) min RA of each field  (decimal degrees)
#        max_ra2             (high) max RA of each field  (decimal degrees)
#        min_dec             min Dec of each field  (decimal degrees)
#        max_dec             max Dec of each field  (decimal degrees)
#
# If a field does not wrap around the 0-360 degree boundary in RA,
#   then we set only the "min_ra1" and "max_ra1" values to the real
#   limits; we set the "min_ra2" and "max_ra2" values to bogus values
#   in that case.  But if an observed field does wrap around the
#   0-360 boundary, we set both sets of RA limits.
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc create_limit_lists { fileid ra_col dec_col fov \
                                   min_ra1 max_ra1 min_ra2 max_ra2 \
                                   min_dec max_dec } {
  global debug 
  upvar $min_ra1 minr1 $max_ra1 maxr1 $min_ra2 minr2 $max_ra2 maxr2
  upvar $min_dec mind  $max_dec maxd

  set DEG2RAD [expr 3.14159/180.0]

  if { $debug > 0 } {
    puts "entering create_limits_lists"
  }

  set minr1 {}
  set maxr1 {}
  set minr2 {}
  set maxr2 {}
  set mind  {}
  set maxd  {}

  # get the list of all object fields, for which we'll need reference stars
  set param_list [list [list "type" "object"]]
  set image_list [select_input_list $param_list]
  if { [llength $image_list] == 0 } {
    puts stderr "create_limits_list: no object fields?"
	 return 1
  }

  # we walk through the list of fields, adding the limits of each one 
  #   to the growing TCL lists "minr1", "maxr1", etc.
  foreach img $image_list {
    if { $debug > 0 } {
	   puts "  next image is $img"
	 }
	 if { [single_image_info $img info_list] != 0 } {
	   puts stderr "create_limits_list: single_image_info fails for $img"
		return 1
	 }

    # get central RA and Dec
    set ra_cent [get_image_value $info_list "ra"]
	 set dec_cent [get_image_value $info_list "dec"]
    if { $debug > 0 } {
	   puts "create_limit_lists: ra_cent $ra_cent  dec_cent $dec_cent"
	 }

    # we calculate the number of degrees of RA covered by this field
    set cosdec [expr cos($dec_cent*$DEG2RAD)]
    set ra_half [expr ($fov/2.0)/$cosdec]
    

    # now, we calculate the limits in RA -- must watch for wraparound
    #   for convenience, place the RA limits in variables
    #            inr1, axr1              min, max RA on low end
    #            inr2, axr2              min, max RA on high end (if needed)
    #
    set minr [expr $ra_cent - $ra_half]
    set maxr [expr $ra_cent + $ra_half]
    if { $minr < 0 } {
      set inr1 0
      set axr1 $maxr
      set inr2 [expr 360.0+$minr]
      set axr2 360
    } elseif { $maxr > 360 } {
      set inr1 0
      set axr1 [expr $maxr-360.0]
      set inr2 $minr
      set axr2 360
    } else {
      set inr1 $minr
      set axr1 $maxr
      # there is no need for a second set of RA values, so we just
      #   pick some values which will never include a real star's RA
      set inr2 400
      set axr2 400
    }

    # now, we calculate the limits in Dec.
    #   for convenience, place the Dec
    #            ind, axd                min, max Dec
    set dec_half [expr ($fov/2.0)]
    set ind [expr $dec_cent - $dec_half]
    set axd [expr $dec_cent + $dec_half]

    # now, we append these limits to the lists
    lappend minr1 $inr1
    lappend maxr1 $axr1
    lappend minr2 $inr2
    lappend maxr2 $axr2
    lappend mind  $ind
    lappend maxd  $axd

  }

  return 0
}



##############################################################################
# Given fileids for input reference catalog and output file,
#   and given a set of TCL lists containing field limits:
#
#        min_ra1             (low) min RA of each field  (decimal degrees)
#        max_ra1             (low) max RA of each field  (decimal degrees)
#        min_ra2             (high) min RA of each field  (decimal degrees)
#        max_ra2             (high) max RA of each field  (decimal degrees)
#        min_dec             min Dec of each field  (decimal degrees)
#        max_dec             max Dec of each field  (decimal degrees)
#
# we walk through the reference catalog.  For each star, we check to see
#   if it falls within at least one field.  
#
#         If yes, we copy the line containing that star to the output
#         If no,  we skip star
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc make_subset { ref_fileid output_fileid  ra_col dec_col \
                     min_ra1 max_ra1 min_ra2 max_ra2 min_dec max_dec } {
  global debug 

  if { $debug > 0 } {
    puts "entering make_subset"
  }

  set nfields [llength $min_ra1]

  if { $debug > 2 } {
    set field 0
    while { $field < $nfields } {
      puts [format "make_subset: RA %6.1f %6.1f (%6.1f %6.1f) Dec %6.1f %6.1f " \
                             [lindex $min_ra1 $field] \
                             [lindex $max_ra1 $field] \
                             [lindex $min_ra2 $field] \
                             [lindex $max_ra2 $field] \
                             [lindex $min_dec $field] \
                             [lindex $max_dec $field]  ]
      incr field
    }
  }


  # we walk through the reference catalog, checking one star at a time
  set count 0
  set subcount 0
  while { [gets $ref_fileid line] != -1 } {

    incr count
    if { $debug > 1 } { 
      if { [expr $count%1000] == 0 } {
        puts "count $count"
      }
    }

    # skip lines starting with the comment character '#'
    if { [regexp "^#" $line] == 1 } {
      continue
    }

    # sanity check
    if { ([llength $line] < $ra_col) || ([llength $line] < $dec_col) } {
      puts stderr "make_subset: skipping bad line ..$line.."
      continue
    }

    set ra [lindex $line $ra_col]
    set dec [lindex $line $dec_col]

    set field 0
    while { $field < $nfields } {

      if { (($ra >= [lindex $min_ra1 $field]) && \
            ($ra <= [lindex $max_ra1 $field]))      ||    \
           (($ra >= [lindex $min_ra2 $field]) && \
	    ($ra <= [lindex $max_ra2 $field])) } {
        if { ($dec >= [lindex $min_dec $field]) &&  \
	     ($dec <= [lindex $max_dec $field]) } {
          puts $output_fileid $line
          incr subcount
	  break
        }
      }

      incr field
    }

  }

  if { $debug > 1 } {
    puts "make_subset: select $subcount of $count stars"
  }

  return 0
}


namespace eval Refcat {
  namespace export init_refcat get_refcat_name get_outcat_name \
                              get_snap_filters get_convolve_program
  variable Refcat_Name {}
  variable Outcat_Name {}
  variable Snap_Filters {}
  variable Convolve_Program {}

#######################################################################
# Set the name of the mini-catalog we'll create from a big catalog.
#   The mini-catalog contains only those stars which appear in 
#   the area covered by a night's observations.
# Set also the list of filters which we'll use in creating the catalog.
#
proc init_refcat { param_file } {
  global debug
  variable Refcat_Name 
  variable Outcat_Name 
  variable Snap_Filters
  variable Convolve_Program

  if { $debug > 0 } {
    puts "entering init_refcat "
  }
  
  set input_dir    [get_param $param_file "input_dir"]
  set output_dir   [get_param $param_file "output_dir"]
  set refcat  [get_param $param_file "refcat_name"]
  set Refcat_Name [format "%s/%s" $input_dir $refcat]
  set outcat  [get_param $param_file "output_cat"]
  set Outcat_Name [format "%s/%s" $output_dir $outcat]
  set snap_filters [get_param $param_file "snap_filters"]
  set Snap_Filters $snap_filters
  set Convolve_Program [get_param $param_file "convolve_program"]


  return 0
}

######################################################################
# Return the name of the input catalog we use as a fundamental
#   reference for astrometry and photometry.  We will use this
#   to create own own "outcat" table with expected instrumental magnitudes.

proc get_refcat_name { } {
  global debug
  variable Refcat_Name

  if { $debug > 0 } {
    puts "entering get_refcat_name"
  }
  return $Refcat_Name
}

######################################################################
# Return the full path/name of the mini-catalog we use as a photometric
#   and astrometric reference.  This is the version we create
#   in the "refcat" procedure, based on some actual catalog from
#   the USNO or elsewhere.
#
proc get_outcat_name { } {
  global debug
  variable Outcat_Name

  if { $debug > 0 } {
    puts "entering get_outcat_name"
  }
  return $Outcat_Name
}


######################################################################
# Return the list of filter abbreviations for the magnitudes
#   which appear in the self-consistent catalog
#
proc get_snap_filters { } {
  global debug
  variable Snap_Filters

  if { $debug > 0 } {
    puts "entering get_snap_filters"
  }
  return $Snap_Filters
}


######################################################################
# Return the name of the program which we use to convolve spectra
#   with filter transmission curves
#
proc get_convolve_program { } {
  global debug
  variable Convolve_Program

  if { $debug > 0 } {
    puts "entering get_convolve_program"
  }
  return $Convolve_Program
}

  # end of namespace Refcat
}
