# TCL script which runs routines to perform simulations of SNAP observations,
#   and analysis thereof.
#
# Returns
#   0               if all goes well
#   1               if there's a problem
#   
# MWR 8/12/2003
#

global debug
set debug 2

set date_str [exec date]; puts "started $date_str"

source param.tcl
source proc_file_list.tcl
source setup.tcl
source focal.tcl
source draw_focal.tcl
source filter.tcl
source refcat.tcl
source telescope.tcl
source snapshot.tcl
source plot.tcl
source snap_foo.tcl


set param_file "snap_driver.param"

set myname [get_param $param_file myname]
set do_foo         [get_param $param_file do_foo]
set do_setup       [get_param $param_file do_setup]
set do_make_list   [get_param $param_file do_make_list]
set do_filter      [get_param $param_file do_filter]
set do_refcat      [get_param $param_file do_refcat]
set do_focal       [get_param $param_file do_focal]
set do_draw_focal  [get_param $param_file do_draw_focal]
set do_telescope   [get_param $param_file do_telescope]
set do_snapshot    [get_param $param_file do_snapshot]
set do_plot        [get_param $param_file do_plot]


# read in names of the param files controlling subroutines
set setup_param    [get_param $param_file setup_param]
set list_param     [get_param $param_file list_param]
set filter_param   [get_param $param_file filter_param]
set refcat_param   [get_param $param_file refcat_param]
set focal_param    [get_param $param_file focal_param]
set telescope_param [get_param $param_file telescope_param]
set snapshot_param [get_param $param_file snapshot_param]
set plot_param     [get_param $param_file plot_param]


#
# Replace default values in param files with those listed in the
#   "setup.param" file (which user has presumably edited)
# 
if { $do_setup == 1 } {
  if { [setup $setup_param] != 0 } {
    puts stderr "$myname: setup returns with error"
    return 1
  }
}




# Squirrel away values of several directories, which we'll
#   use later to call external programs.
#   The namespace "Directory", defined in "setup.tcl", contains
#   the routines to set/get directory values.
#
namespace import -force Directory::*
if { [init_directory $setup_param] != 0 } {
  puts stderr "$myname: init_directory returns with error"
  return 1
}


# The namespace Proc_File_List, defined in "proc_file_list.tcl",
#   contains procedures which pick out pieces of a general TCL list
#   in a convenient fashion.
namespace import -force Proc_File_List::*



# The namespace Filter, defined in "filter.tcl",
#   contains procedures which initialize, set and retrieve information
#   related to filters.
namespace import -force Filter::*
if { [init_filter $filter_param] != 0 } {
  puts stderr "$myname: init_filter fails"
  return 1
}

# The namespace Refcat, defined in "refcat.tcl",
#   contains procedures to set the name of the mini_catalog
#   file and filter abbreviations, and to retrieve that information.
#   This code is used in "refcat.tcl", "astrom.tcl", "photom.tcl"
#
namespace import -force Refcat::*
if { [init_refcat $refcat_param ] != 0 } {
  puts stderr "$myname: init_refcat returns with error"
  return 1
}


# The namespace Focal, defined in "focal.tcl",
#   contains procedures to set the name of the low-level chip param file,
#   and to retrieve that information.
#   This code is used in "focal.tcl",  ...
#
namespace import -force Focal::*
if { [init_focal $focal_param ] != 0 } {
  puts stderr "$myname: init_focal returns with error"
  return 1
}


# The namespace Draw_Focal, defined in "draw_focal.tcl",
#   contains procedures to draw pictures of the focal plane.
#   This code is used in "draw_focal.tcl",  ...
#   And yes, it does read the "focal.param" file, same as the
#   "init_focal" proc
#
namespace import -force Draw_Focal::*
if { [init_draw_focal $focal_param ] != 0 } {
  puts stderr "$myname: init_draw_focal returns with error"
  return 1
}


# The namespace Telescope, defined in "telescope.tcl",
#   contains procedures which describe effects of the telescope
#   optics on light.
#   This code is used in "telescope.tcl",  ...
#
namespace import -force Telescope::*
if { [init_telescope $telescope_param ] != 0 } {
  puts stderr "$myname: init_telescope returns with error"
  return 1
}


# The namespace Snapshot, defined in "snapshot.tcl",
#   contains procedures which describe the sequence of
#   images we'll take with the telescope:
#       starting location, number of steps, step size, etc.
#   This code is used in "snapshot.tcl",  ...
#
namespace import -force Snapshot::*
if { [init_snapshot $snapshot_param ] != 0 } {
  puts stderr "$myname: init_snapshot returns with error"
  return 1
}





  # 
  # run some routine which is not part of the regular pipepline
  if { $do_foo == 1 } {
    if { [do_foo] != 0 } {
      puts "$myname: do_foo returns with error"
    }
    exit 0
  }
  

  # create a small reference catalog of stars, based on a much larger
  #   catalog.  We'll use the small catalog for both astrometry and
  #   photometry.
  # 
  if { $do_refcat == 1 } {
    if { [refcat $refcat_param] != 0 } {
      puts stderr "$myname: refcat returns with error"
      return 1
    }
  }


  # generate a detailed, low-level param file describing the layout
  #   of detectors in the focal plane
  # 
  if { $do_focal == 1 } {
    if { [focal $focal_param] != 0 } {
      puts stderr "$myname: focal returns with error"
      return 1
    }
  }


  # set up the filter assignments
  # 
  if { $do_filter == 1 } {
    if { [filter $filter_param] != 0 } {
      puts stderr "$myname: filter returns with error"
      return 1
    }
  }




  # draw a picture of the focal plane
  # 
  if { $do_draw_focal == 1 } {
    if { [draw_focal $focal_param] != 0 } {
      puts stderr "$myname: draw_focal returns with error"
      return 1
    }
  }

  # take a snapshot
  # 
  if { $do_snapshot == 1 } {
    if { [snapshot $snapshot_param] != 0 } {
      puts stderr "$myname: snapshot returns with error"
      return 1
    }
  }



set date_str [exec date]; puts "ended $date_str"

# end of pipeline
return 0


