#
# This TCL script performs all the steps involved in taking
#   a simulated image with the SNAP camera.
#
# It writes output file(s) to the output_dir.
#
# MWR 8/20/2003 
#    
#
# Removed a redundant call to find incident_angle from step j2 of the
#   "snapshot" routine.
# MWR 3/11/2004   
#
# Now check to see if graphics are possible at start of this stage;
#   if not, then we re-set the "snapshot_draw_stars" variable to zero
#   but keep running.
# MWR 7/28/2004
#
# Print magnitudes and uncerts to 4 places after decimal point, not 3.
# MWR 8/7/2004


if { [catch { set debug } ] == 1 } {
  set debug 2
}

source param.tcl


###########################################################################
# PROCEDURE: snapshot
# 
# DESCRIPTION: Take a simulated image(s) with the SNAP camera.
#
#              Walk through a reference catalog, which has positions,
#              spectral types, zeropoints (expected magnitudes) for all stars.
#              For each star, figure out where on the focal plane it falls,
#              convolve with appropriate filter, add in effects for the
#              optics and filter and chip, etc., ending up with an
#              instrumental magnitude.
#
#              Write the result into output file(s).
#
#              We now (9/2/2003) use the parameters in snapshot.param
#              to control the number and locations of a sequence of 
#              snapshots.
#             
#
# RETURNS
#    0           if all goes well
#    1           if there's a problem
#    
proc snapshot { param_file } {
  global debug

  if { $debug > 0 } {
    puts "entering snapshot"
  }

  # get information from the param file
  if { [file exists $param_file] != 1 } {
    puts stderr "snapshot: can't open param file $param_file"
    return 1
  }
  set input_dir [get_param $param_file "input_dir"]
  set output_dir [get_param $param_file "output_dir"]
  set snapshot_draw_stars [get_param $param_file "snapshot_draw_stars"]
  set terse_obsfile [get_param $param_file "terse_obsfile"]
  set verbose_obsfile [get_param $param_file "verbose_obsfile"]
  set image_info_file [get_param $param_file "image_info_file"]

  # these parameters allow us to set the zero-point of the magnitude
  #   scale properly for our synthetic photometry
  set snapshot_vega_spectrum [get_param $param_file "snapshot_vega_spectrum"]
  set snapshot_model_spectrum [get_param $param_file "snapshot_model_spectrum"]
  set snapshot_vband_index [get_param $param_file "snapshot_vband_index"]

  # this parameter is needed to calculate background sky and 
  #   readnoise and dark current contributions to magnitude measurements
  set snapshot_aper_rad [get_param $param_file "snapshot_aper_rad"]
 
  # this parameter is used in calculating the output instrumental magnitude
  set snapshot_mag_zero [get_param $param_file "snapshot_mag_zero"]

  # should we add random noise to the output magnitude values?
  set snapshot_add_noise [get_param $param_file "snapshot_add_noise"]

  # before we get started, check the value of the "snapshot_draw_stars"
  #   parameter.  If it is 1, so we are supposed to draw each star
  #   as we calculate its magnitude, check to see if that's possible.
  if { [snapshot_check_graphics snapshot_draw_stars ] != 0 } {
    puts stderr "snapshot: snapshot_check_graphics fails"
    return 1
  }

  # get the name of the program which is used to convolve spectra
  #   with filters
  set convolve_program [get_convolve_program]

  
  # open the list of stars we're going to observe (well, those which
  #   happen to fall onto detectors, anyway)
  set output_cat [get_outcat_name]
  if { [file exists $output_cat] != 1 } {
    puts stderr "snapshot: catalog file $output_cat doesn't exist -- quitting"
    return 1
  }

  if { $snapshot_add_noise == 1 } {
    if { [init_random_deviates] != 0 } {
      puts stderr "snapshot: init_random_deviates fails -- quitting "
      return 1
    }
  }


  # cycle through all the requested (RA, Dec) positions and take
  #   pictures at each one
  set sequence_list [get_snapshot_seq_info]
  set start_ra [lindex $sequence_list 0]
  set start_dec [lindex $sequence_list 1]
  set snaps_per_step [lindex $sequence_list 2]
  set step_ra [lindex $sequence_list 3]
  set step_dec [lindex $sequence_list 4]
  set num_steps [lindex $sequence_list 5]
  set exp_base [lindex $sequence_list 6]
  set exp_sig [lindex $sequence_list 7]

  # we will write information about the properites of each image
  #   to this file
  set image_filename [format "%s/%s" $output_dir $image_info_file]
  set image_fileid [open $image_filename "w"]
  if { $image_fileid < 0 } {
    puts stderr "snapshot: cannot open image_info file $image_info_file"
    return 1
  }


  # gather some information that we'll use for all calculations in the
  #   loop below.  More efficient to set some variables just once here,
  #   rather than set them over and over to the same values below.
  # telescope collecting area
  if { [telescope_area_sqcm tel_area_sqcm] != 0 } {
    puts stderr "snapshot: telescope_area_sqcm fails"
    return 1
  }
  if { $debug > 2 } {
    puts "snapshot: tele_area_sqcm is $tel_area_sqcm"
  }
  # we put information on all chips into an array, so that we can 
  #   quickly access information on the chip each star falls on in
  #   the main loop below
  if { [read_chip_info num_chips chip_info] != 0 } {
    puts stderr "snapshot: read_chip_info fails" 
    return 1
  }
 
  # compare a real flux-calibrated spectrum of Vega to the model
  #   spectrum of a Vega-like star in our set of standard model spectra;
  #   figure out how we must correct the synthetic photometry of our
  #   model spectra to bring them to the proper magnitude zero point.
  #   We assume Vega has magnitude 0 in V-band.
  if { [calc_magnitude_zeropoint $input_dir $snapshot_vega_spectrum \
                                 $snapshot_model_spectrum \
                                 $snapshot_vband_index \
                                 $convolve_program \
                                 mag_zp_offset] != 0 } {
    puts stderr "snapshot: calc_magnitude_zeropoint fails "
    return 1
  }
  if { $debug > 2 } {
    puts "mag_zp_offset is $mag_zp_offset"
  }

  # initialize an array full of information about sky value(s)
  #   in this catalog of stars
  if { [init_sky $output_cat sky_num sky_info] != 0 } {
    puts stderr "snapshot: init_sky fails "
    return 1
  }
  if { $debug > 2 } {
    set i 0
    while { $i < $sky_num } {
      puts "  sky $i is $sky_info($i) "
      incr i
    }
  }



  # prepare to run through a double loop: steps from one position to 
  #   the next, and snaps at each position
  set target_ra $start_ra
  set target_dec $start_dec
  # this will be a part of the output file names, unique for every picture
  set snapshot_id 0

  for { set i_step 0 } { $i_step < $num_steps } { incr i_step } {

    # set the position
    if { [set_foc_radec $target_ra $target_dec] != 0 } {
      puts stderr "snapshot: set_foc_radec fails for $target_ra $target_dec "
    }
    if { $debug > 0 } {
      puts "snapshot: step is $i_step, ra $target_ra  dec $target_dec"
    }
 
   
    for { set i_snap 0 } { $i_snap < $snaps_per_step } { incr i_snap } {

      if { $debug > 0 } {
        puts "snapshot: snap is $i_snap "
      }

      # calculate the exposure time for this image
      if { [calc_exptime $exp_base $exp_sig this_exptime] != 0 } {
        puts stderr "snapshot: calc_exptime returns with error -- quitting"
        return 1
      }

  
      # open the reference catalog of stars
      set ref_fileid [open $output_cat "r"]
      if { [catch { set ref_fileid } ] == 1 } {
        puts stderr "snapshot: error opening input file $output_cat -- quitting"
        return 1
      }
    
      # open the "terse" output file, which will contain only RA, Dec, instr. mag
      set full_terse [format "%s/%s.%d" $output_dir $terse_obsfile \
                                            $snapshot_id ]
      set terse_fileid [open $full_terse "w"]
      if { $terse_fileid < 0 } {
        puts stderr "refcat: can't open output file $full_terse"
      }
    
      # open the "verbose" output file, which will contain large amounts of
      #   information on each star (so that we can plot residuals as a function
      #   of just about any quantity)
      set full_verbose [format "%s/%s.%d" $output_dir $verbose_obsfile\
                                            $snapshot_id ]
      set verbose_fileid [open $full_verbose "w"]
      if { $verbose_fileid < 0 } {
        puts stderr "refcat: can't open output file $full_verbose"
      }
    
     
      # Now, we loop through the catalog of stars.  For each star, we
      #    a) read in information from catalog
      #    b) figure out the (x,y) position on focal plane
      #    c) determine which chip, if any, it falls on, and (row, col) therein
      #    c2)  determine the filter starlight strikes
      #    d) get telescope QE factor
      #    e) get chip readnoise, dark current, etc. 
      #    f) determine angle at which starlight strikes filter
      #    f2)  pick appropriate sky magnitude (if any in catalog)
      #    g) convolve spectrum with filter (w/ angle factor) to get num photons
      #    h) apply QE factors 
      #    i) convert photons to electrons of signal 
      #    i2)  figure out contributions to noise
      #    i3)  calculate instrumental magnitude and uncertainty
      #    j) write output line to output file(s)
      #    k) (if desired) draw star on focal plane
      # 
     
      while { [gets $ref_fileid line] > 0 } {
    
        if { $debug > 3 } {
          puts "next line is ..$line.."
        }
    
        # skip lines starting with the "#" character
        if { [regexp ^# $line] == 1 } {
          continue
        }
    
        if { $debug > 5 } {
          puts "next line is ..$line.."
        }
          
        # Step a. read information from the line: RA, Dec, spec_type, zeropoint
        #   and then expected magnitude in each of the filters.
        #   Ignore any line which doesn't look right ...
        if { [read_outcat_info $line ra dec sptype zp mag_list] != 0 } {
          puts stderr "snapshot: read_outcat_info fails on line ..$line.. "
          continue
        }
        # we are going to use the snapshot_vband_index'th  magnitude 
        #   in the mag_list to scale the synthetic photometry 
        set star_mag [lindex $mag_list $snapshot_vband_index]
    
        # Step b.  Figure out the (x,y) position on focal plane of this star
        if { [radec_to_focxy $ra $dec x y] != 0 } {
          puts stderr "snapshot: radec_to_focxy fails for $ra $dec "
          return 1
        }
        if { $debug > 0 } {
          puts [format "   %12.5f %12.5f  falls at %8.3f %8.3f " $ra $dec $x $y]
        }
    
        # Step c.  Figure out which chip, and (row, col) on chip.
        #   If we get back values of -1 for chip, row, col, then the star
        #   doesn't fall on any chip.  In that case, we go to next star
        if { [focxy_to_chip $x $y chip row col] != 0 } {
          puts stderr "snapshot: focxy_to_chip fails on $x $y "
          return 1
        }
        if { $debug > 0 } {
          puts [format "   chip %3d  row %4.0f  col %4.0f " $chip $row $col]
        }
        if { ($chip == -1) } {
          if { $debug > 0 } {
            puts "       star doesn't fall on any chip -- skip it"
          }
          continue
        }
    
        # Step c2. Figure out which filter the light strikes
        if { [chip_to_filter $chip $row $col filter_index filter_file] != 0 } {
          puts stderr "snapshot: chip_to_filter fails on $chip $row $col"
          return 1
        }
    
        # Step d.  Get the telescope QE factor at this location
        if { [get_telescope_qe $x $y tel_qe] != 0 } {
          puts stderr "snapshot: get_telescope_qe fails on $x $y "
          return 1
        }
        if { $debug > 0 } {
          puts [format "   filter index $filter_index  filter $filter_file  tel_qe %6.2f  " $tel_qe]
        }
    
        # Step e.  we should get filter/chip properties here ...
        #   make a sanity check to ensure that the given chip ID lies
        #   within the range of the "chip_info" array we filled before
        #   entering the loop.  We'll pass the appropriate member of 
        #   the chip_info array to the routine which calculates magnitude ...
        if { ($chip < 0) || ($chip >= $num_chips) } {
          puts stderr "snapshot: invalid chip ID $chip ?!"
          return 1
        }
        if { $debug > 1 } {
          puts "snapshot: chip_info is $chip_info($chip) "
        }
        # read the parameters which tell us how sensitivity 
        #   varies across this chip
        set chip_var_params [lindex $chip_info($chip) 20]
        if { $debug > 2 } {
          puts "chip_var_params is ..$chip_var_params.."
        }
        if { [calc_intrachip_factor $chip_var_params $row $col \
                                             chip_var_factor] != 0 } {
          puts stderr "snapshot_calc_mag: calc_intrachip_factor fails?!"
          puts stderr "snapshot_calc_mag: chip_var_params is ..$chip_var_params.."
          return 1
        }
        if { $debug > 2 } {
          puts "chip_var_factor is ..$chip_var_factor.."
        }
  
    
        # Step f.  Figure angle (away from normal) at which 
        #          starlight strikes filter
        if { [focxy_to_incidence $x $y incident_angle] != 0 } {
          puts stderr "snapshot: focxy_to_incidence fails for $x $y "
          return 1
        }

        # Step f2. Pick the "best" sky value for this star.
        #          (It may not be a sophisticated choice)
        #          If there are no sky entries, we end up
        #          with an empty {} best_sky variable
        if { [pick_best_sky $ra $dec $sky_num sky_info best_sky] != 0 } {
          puts stderr "snapshot: pick_best_sky fails "
          return 1
        }
        # we are going to use the snapshot_vband_index'th  magnitude 
        #   in the mag_list to scale the synthetic photometry 
        #   of the sky ... so, pick it out of the "best_sky" 
        if { $best_sky == {} } {
          set best_sky_mag 99.0
        } else {
          set best_sky_mag [lindex [lindex $best_sky 4] $snapshot_vband_index]
        }

    
        # Step g.  Convolve the spectrum with the appropriate filter
        #          (and should include angle factor, too)
        #          The result is the "observed magnitude", or "obs_mag"
        #       
        #          We also calculate the brightness of the sky background
        #          and other factor needed to determine the uncertainty 
        if { [snapshot_calc_mag $input_dir $output_dir $sptype $star_mag \
                       $this_exptime $tel_qe $tel_area_sqcm \
                       $filter_file $incident_angle $convolve_program \
                       $chip_info($chip) $mag_zp_offset \
                       $chip_var_factor \
                       $snapshot_aper_rad $best_sky $best_sky_mag \
                       $snapshot_mag_zero $snapshot_add_noise \
                       obs_mag obs_uncert] != 0 } {
          puts stderr "snapshot: snapshot_calc_mag fails "
          return 1
        }
        if { $debug > 2 } {
          puts "obs_mag ..$obs_mag..  obs_uncert ..$obs_uncert.. "
        }

    
        # Step h. apply QE factors to the instrumental mag
        # added 3/25/2004 -- just a test XXX
        # set obs_mag [expr $obs_mag + $i_step*0.01]
        # set obs_mag [expr $obs_mag + $filter_index*0.01]
    
        # Step i. convert to desired final formag?
    
        # Step j1. write this line to the terse output file
        set terse_line [format " %9.5f %9.5f  %7.4f %7.4f" $ra $dec \
                                       $obs_mag $obs_uncert]
        puts $terse_fileid $terse_line
    
        # Step j2. Calculate additional quantities needed for verbose output
        #          file, and then print everything to the verbose file 
        set verbose_line [format " %9.5f %9.5f %7s %6.3f " $ra $dec $sptype $zp]
        set verbose_line [format "%s %s  " $verbose_line $mag_list]
        set verbose_line [format "%s %3d %7.2f %7.2f " $verbose_line $chip $row $col]
        set verbose_line [format "%s %7.2f %7.2f " $verbose_line $x $y]
        # this is the radial distance from optical center, in mm
        set r [expr sqrt($x*$x + $y*$y)]
        set verbose_line [format "%s %7.2f %7.5f " $verbose_line $r $incident_angle]
        #
        set fiducial_index [get_filter_index_to_fiducial_index $filter_index]
        # Now that we know the index of the fiducial filter, we can pick 
        #   out the synthetic (expected) magnitude in this filter
        set synth_mag [lindex $mag_list $fiducial_index]
        set verbose_line [format "%s %7.4f %3d %3d " $verbose_line $synth_mag \
                                    $filter_index $fiducial_index]
        #
        set delta_mag [expr $obs_mag - $synth_mag]
        set verbose_line [format "%s %7.4f %7.4f  %7.4f " \
                              $verbose_line $obs_mag $obs_uncert $delta_mag]
        puts $verbose_fileid $verbose_line
    
        # Step k. (if desired) plot this star on focal plane
        if { $snapshot_draw_stars == 1 } {
          if { [draw_star $ra $dec $obs_mag] != 0 } {
            puts stderr "snapshot: draw_star fails on $ra $dec $obs_mag"
            return 1
          }
        }
    
        #   puts "hit return to continue " ; set aa [gets stdin]
           
      }

      # export the canvas, in case the user chose to draw stars on it
      if { $snapshot_draw_stars == 1 } {
        if { [export_canvas] != 0 } {
          puts stderr "snapshot: export_canvas returns with error"
          return 1
        }
      }
    
      # close all files we've opened
      close $terse_fileid
      close $verbose_fileid
      close $ref_fileid

      # write information about this image into the image-information file
      if { [write_image_info $image_fileid $snapshot_id \
                       $target_ra $target_dec $this_exptime] != 0 } {
        puts stderr "snapshot: write_image_info fails -- quitting "
        return 1
      }

      # now increment the snapshot ID number, so that we will write the next
      #   bunch of output to a file with a different name ...
      incr snapshot_id

    # end of loop over snaps at a given position (i_snap)
    } 
 
    # time to move to the next position
    set target_ra [expr $target_ra+$step_ra]
    if { $target_ra < 0 } {
      set target_ra [expr $target_ra+360.0]
    }
    if { $target_ra >= 360 } {
      set target_ra [expr $target_ra-360.0]
    }
    set target_dec [expr $target_dec+$step_dec]
    if { $target_dec < -90 } {
      set target_dec -90.0
    }
    if { $target_dec > 90 } {
      set target_dec 90.0
    }

  # end of loop over changing positions (i_step)
  }

  # close the file which holds information on properties of each image
  close $image_fileid
  

  return 0
}


############################################################################
# PROCEDURE: calc_magnitude
# 
# DESCRIPTION: Convolve a spectrum with a particular passband to get 
#              a synthetic magnitude.
#
#                  input_dir         directory with spectral data
#                  output_dir        directory with actual filter data
#                  sptype            name of stellar spectral type      
#                  star_mag          the magnitude of this star in V-like band
#                  exptime           exposure time in seconds
#                  tel_qe            a factor to account for vignetting
#                                         in optics.  Value should be 1.0
#                                         for perfect optics, 0.0 for no light
#                  tel_area_sqcm     telescope area in square centimeters
#                  filter_file       name of actual filter file
#                  incident_angle    angle of light away from normal 
#                  convolve_program  full name of executable which convolves
#                                         spectrum with passband
#                  chip_info         a list which include readnoise, gain,
#                                         and other properties of detector
#                  mag_zp_offset     step 1 in turning synth mag into real mag
#                                         (see below)
#                  chip_var_factor   accounts for variation in sensitivity 
#                                         across the chip; ranges from 
#                                         0.0 (chip detects no light here) to
#                                         1.0 (chip detects all photons here)
#                  synth_mag         the synthetic mag for star in this filter
#                                         (set on output)
#                  synth_uncert      the uncertainty in mag for star 
#                                         in this filter
#                                         (set on output)
#                                         (special value 99.0 for "no data")
#
#             Now, after we convolve our model spectrum with the actual
#             passband, we need to make two shifts to get an accurate
#             value for the number of photons detected:
#
#                       1. add "mag_zp_offset" to synthetic mag -- will 
#                                produce the number of photons from a star
#                                of magnitude V=0 of this spectral class.
#                                This does proper flux calibration of our
#                                model spectra ...
#                       2. add the star's magnitude in fiducial passband
#                                (the first {index,mag} pair) to get
#                                number of photons from a star of proper
#                                magnitude.  This takes our model spectrum
#                                to the actual magnitude of the star.
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc snapshot_calc_mag { input_dir output_dir sptype star_mag \
                             exptime tel_qe tel_area_sqcm \
                             filter_file incident_angle \
                             convolve_program  chip_info mag_zp_offset \
                             chip_var_factor \
                             aper_rad sky_info sky_mag \
                             mag_zero add_noise \
                             o_synth_mag o_synth_uncert } {
  global debug 
  upvar $o_synth_mag synth_mag 
  upvar $o_synth_uncert synth_uncert 

  if { $debug > 0 } {
    puts "entering snapshot_calc_mag "
  }
 
  # get the full path name of spectrum file 
  if { [get_spectrum_file $input_dir $sptype spectrum_file] != 0 } {
    puts stderr "snapshot_calc_mag: get_spectrum_file fails for type $sptype "
    return 1
  }
  
  # verify that the spectrum file exists (a sanity check)
  if { [file exists $spectrum_file] != 1 } {
    puts stderr "snapshot_calc_mag: file $spectrum_file doesn't exist?"
    return 1
  }

  # build the name of the filter file, and verify that it exists
  set filter_file [format "%s/%s" $output_dir $filter_file]
  if { [file exists $filter_file] != 1 } {
    puts stderr "snapshot_calc_mag: file $filter_file doesn't exist?"
    return 1
  }

  # get the fiducial angle of incidence 
  set ll [get_incidence_info]
  set fiducial_angle [lindex $ll 1]

   


  # use the convolution program to calculate the synthetic magnitude
  set cmd [format "%s %s %s %lf %lf " $convolve_program $spectrum_file \
                      $filter_file $fiducial_angle $incident_angle]
  if { $debug > 4 } {
    puts "snapshot_calc_mag: cmd is ..$cmd.."
  }
  set retcode [catch { set result_line [eval exec $cmd] } ]
  if { $retcode != 0 } {
    puts stderr "snapshot_calc_mag: error exec'ing ..$cmd.."
    return 1
  } 
  set synth_nphoton [lindex $result_line 3]
  set synth_mag [lindex $result_line 5]
  if { $debug > 4 } {
    puts "snapshot_calc_mag: result is ..$result_line.."
  }

  # we will make two corrections to the synthetic magnitude: 
  #    - one to take synth_mag to a star of magnitude zero
  #    - one to shift further, to the actual magnitude of star
  # so we should shift the synthetic magnitude by this much to get
  #    the actual magnitude ...
  set shift_mag [expr $mag_zp_offset+$star_mag]
  if { $debug > 4 } {
    puts "snapshot_calc_mag: mag_zp_offset is $mag_zp_offset ...."
    puts "snapshot_calc_mag: shift by $shift_mag ...."
  }
  
  # .... which is equivalent to multiplying the number of photons per
  #    square cm per second -- also produced by the synthetic photometry
  #    program -- by this factor
  set scale_factor [expr pow(10.0, -0.4*$shift_mag)]
  if { $debug > 4 } {
    puts "snapshot_calc_mag: scale by $scale_factor ..."
  }
  # so calculate the number of photons per sq.cm. per second from this star 
  set actual_nphoton [expr $synth_nphoton*$scale_factor]
  if { $debug > 4 } {
    puts [format "snapshot_calc_mag: actual_nphoton is %12.7f " \
                             $actual_nphoton]
  }

  # a correction for any vignetting or other imperfections in optics
  set actual_nphoton [expr $actual_nphoton*$tel_qe]
  if { $debug > 4 } {
    puts [format "snapshot_calc_mag: after tel_qe, actual_nphoton is %12.7f " \
                             $actual_nphoton]
  }

  # a correction for any variations in sensitivity across this chip
  set actual_nphoton [expr $actual_nphoton*$chip_var_factor]
  if { $debug > 4 } {
    puts [format "snapshot_calc_mag: after chip_var_factor, actual_nphoton is %12.7f " \
                             $actual_nphoton]
  }


  # now calculate the total number of photons gathered by the actual
  #   collecting area of the telescope each second 
  set actual_nphoton [expr $actual_nphoton*$tel_area_sqcm]
  if { $debug > 4 } {
    puts [format "snapshot_calc_mag: after area, actual_nphoton is %12.7f " \
                             $actual_nphoton]
  }

  # multiply photons per second by the exposure time, in seconds
  set actual_nphoton [expr $actual_nphoton*$exptime]
  if { $debug > 4 } {
    puts [format "snapshot_calc_mag: after exptime, actual_nphoton is %12.7f " \
                             $actual_nphoton]
  }
  # finally, we use the number of detected photons (= number of electrons)
  #   to calculate an instrumental magnitude for output
  if { $actual_nphoton < 1 } {
    set synth_mag $mag_zero
  } else {
    set synth_mag [expr $mag_zero - (2.5*log10($actual_nphoton))]
  }
  if { $debug > 4 } {
    puts [format "snapshot_calc_mag: actual_nphoton %12.7f  synth_mag %8.4f" \
                             $actual_nphoton $synth_mag]
  }

  # temporary output line for some debugging purposes
  set chip_id [lindex $chip_info 0]
  if { $debug > 2 } {
    puts [format "QQ  sp %4s  filt %s  chip %3d  nphot %12.5f " \
                    $sptype $filter_file $chip_id $actual_nphoton ]
  }


  # 
  # Okay, now we repeat a bunch of the steps above in order to calculate
  #   the number of photons we detect from the sky ..
  #
  if { $debug > 4 } {
    puts "snapshot_calc_mag: sky_info is ..$sky_info.. "
  }

  # check -- if there is NO sky information given, we assume that
  #   the sky contributes no light to our measurement.  In other words,
  #   set the number of sky photons inside aperture to zero and move on.
  if { $sky_info == {} } {
    if { $debug > 2 } {
      puts "snapshot_calc_mag: given no sky information, use zero contribution"
    }
    set sky_photon_sqarcsec 0

  } else {

    set sky_sptype [lindex $sky_info 2]
    if { $debug > 2 } {
      puts "snapshot_calc_mag: sky_sptype $sky_sptype "
    }

    # get the full path name of (sky) spectrum file 
    if { [get_spectrum_file $input_dir $sky_sptype sky_spectrum_file] != 0 } {
      puts stderr "snapshot_calc_mag: get_spectrum_file fails for sky $sky_sptype "
      return 1
    }
    # verify that the spectrum file exists (a sanity check)
    if { [file exists $sky_spectrum_file] != 1 } {
      puts stderr "snapshot_calc_mag: file $sky_spectrum_file doesn't exist?"
      return 1
    }

    # use the convolution program to calculate the synthetic magnitude
    set cmd [format "%s %s %s %lf %lf " $convolve_program $sky_spectrum_file \
                        $filter_file $fiducial_angle $incident_angle]
    if { $debug > 4 } {
      puts "snapshot_calc_mag: cmd is ..$cmd.."
    }
    set retcode [catch { set result_line [eval exec $cmd] } ]
    if { $retcode != 0 } {
      puts stderr "snapshot_calc_mag: error exec'ing ..$cmd.."
      return 1
    }   
    set sky_synth_nphoton [lindex $result_line 3]
    set sky_synth_mag [lindex $result_line 5]
    if { $debug > 4 } {
      puts "snapshot_calc_mag: result is ..$result_line.."
    }

    # we will make two corrections to the synthetic magnitude: 
    #    - one to take synth_mag to a star of magnitude zero
    #    - one to shift further, to the actual magnitude of star
    # so we should shift the synthetic magnitude by this much to get
    #    the actual magnitude ...
    set sky_shift_mag [expr $mag_zp_offset+$sky_mag]
    if { $debug > 2 } {
      puts "snapshot_calc_mag: sky shift by $sky_shift_mag ...."
    }
  
    # .... which is equivalent to multiplying the number of photons per
    #    square cm per second -- also produced by the synthetic photometry
    #    program -- by this factor
    set sky_scale_factor [expr pow(10.0, -0.4*$sky_shift_mag)]
    if { $debug > 4 } {
      puts "snapshot_calc_mag: scale sky by $sky_scale_factor ..."
    }
    # so calculate the number of photons per sq.cm. per second from this star 
    #    per square arcsecond -- that's how the sky values are given as input
    set sky_photon_sqarcsec [expr $sky_synth_nphoton*$sky_scale_factor]
    if { $debug > 4 } {
      puts [format "snapshot_calc_mag: sky_photon_sqarcsec is %12.7f " \
                             $sky_photon_sqarcsec]
    }

    # a correction for any vignetting or other imperfections in optics
    set sky_photon_sqarcsec [expr $sky_photon_sqarcsec*$tel_qe]
    if { $debug > 4 } {
      puts [format "snapshot_calc_mag: after tel_qe, sky_photon_sqarcsec is %12.7f " \
                             $sky_photon_sqarcsec]
    }

    # a correction for any variations in sensitivity across this chip
    set sky_photon_sqarcsec [expr $sky_photon_sqarcsec*$chip_var_factor]
    if { $debug > 4 } {
      puts [format "snapshot_calc_mag: after chip_var_factor, sky_photon_sqarcsec is %12.7f " \
                               $sky_photon_sqarcsec]
    }


    # now calculate the total number of photons gathered by the actual
    #   collecting area of the telescope each second 
    set sky_photon_sqarcsec [expr $sky_photon_sqarcsec*$tel_area_sqcm]
    if { $debug > 4 } {
      puts [format "snapshot_calc_mag: after area, sky_photon_sqarcsec is %12.7f " \
                             $sky_photon_sqarcsec]
    }

    # multiply photons per second by the exposure time, in seconds
    set sky_photon_sqarcsec [expr $sky_photon_sqarcsec*$exptime]
    if { $debug > 4 } {
      puts [format "snapshot_calc_mag: after exptime, sky_photon_sqarcsec is %12.7f " \
                             $sky_photon_sqarcsec]
    }

    # temporary output line for some debugging purposes
    set chip_id [lindex $chip_info 0]
    if { $debug > 2 } {
      puts [format "QQ sky %4s  filt %s  chip %3d  nphot %12.5f " \
                    $sky_sptype $filter_file $chip_id $sky_photon_sqarcsec ]
    }

  } 
  # end of "else" clause for calculating number of sky photons

  # we need to know how big the synthetic aperture is, so that we
  #   can calculate the amount of sky light and detector noise inside
  #   the aperture
  if { [calc_aperture_properties $aper_rad $chip_info \
                       aperture_npix aperture_sqarcsec] != 0 } {
    puts "snapshot_calc_magnitude: calc_aperture_properties fails "
    return 1
  }
  if { $debug > 4 } {
    puts "snapshot_calc_magnitude: aperture_npix is $aperture_npix"
    puts "snapshot_calc_magnitude: aperture_sqarcsec is $aperture_sqarcsec"
  }

  # now we can calculate the number of sky photons falling within
  #   the aperture during the exposure
  set sky_nphoton [expr $sky_photon_sqarcsec*$aperture_sqarcsec]
  if { $debug > 2 } {
    puts [format "QQ sky nphot %12.5f " $sky_nphoton]
  }

  # follow Howell, PASP 101, 616 (1989), in calculating the signal
  #   and the noise.  All our calculations so far have given us
  #   "number of detected photons".  We now assume that 
  #              1 detected photon = 1 electron
  #   and switch to thinking in terms of electrons.  We need
  #   to know
  # 
  #             N_star      number of electrons produced by star 
  #                                  (inside aperture)
  #             N_sky       number of electrons produced by sky
  #                                  (inside aperture)
  #             N_dark      number of electrons produced by dark current
  #                                  (inside aperture)
  #             N_read      variance of electrons produced by readnoise
  #                                  (inside aperture)
  #
  set N_star $actual_nphoton
  set N_sky  $sky_nphoton
  if { $debug > 4 } {
    puts "snapshot_calc_mag: N_star $actual_nphoton"
    puts "snapshot_calc_mag: N_sky  $sky_nphoton "
  }

  set dark_per_sec_per_pix [lindex $chip_info 15]
  set N_dark [expr $dark_per_sec_per_pix*$aperture_npix*$exptime]
  if { $debug > 4 } {
    puts "snapshot_calc_mag: dark_per_sec_per_pix $dark_per_sec_per_pix"
    puts "snapshot_calc_mag: N_dark $N_dark"
  }

  set readnoise_per_pix [lindex $chip_info 16]
  set N_read [expr ($readnoise_per_pix*$readnoise_per_pix)*$aperture_npix]
  if { $debug > 4 } {
    puts "snapshot_calc_mag: readnoise_per_pix $readnoise_per_pix"
    puts "snapshot_calc_mag: N_read $N_read"
  }

  # we can now calculate the signal, noise, and signal-to-noise ratio
  #   following Howell's "CCD equation".
  #   The S/N ratio is simply the inverse of the uncertainty, at least
  #   for large S/N.  
  #   We use the special value "99.0" for the uncertainty in a magnitude
  #   when it is formally indeterminate
  set signal $N_star
  set noise [expr sqrt($N_star + $N_sky + $N_dark + $N_read)]
  if { $noise > 0 } {
    set signal_to_noise [expr $signal/$noise]
    set synth_uncert [expr 1.0/$signal_to_noise]
  } else {
    set signal_to_noise 0.0
    set synth_uncert 99.0
  }
  if { $debug > 2 } {
    puts [format "snapshot_calc_mag: S/N %10.3f   uncert %8.3f " \
                       $signal_to_noise $synth_uncert ]
  }

  # now, if the user wants noise added to the magnitude measurements,
  #   add a random gaussian amount of noise here
  if { $add_noise == 1 } {
    set deviate [get_random_deviate]
    set noise_mag [expr $deviate*$synth_uncert]
    set synth_mag_with_noise [expr $synth_mag+$noise_mag]
    if { $debug > 4 } {
      puts [format "snapshot_calc_mag: true mag %8.4f  with noise %8.4f " \
                         $synth_mag $synth_mag_with_noise]
    }
    set synth_mag $synth_mag_with_noise
  }


  return 0
}


############################################################################
# PROCEDURE: calc_instrumental_magnitude
# 
# DESCRIPTION: Given a synthetic magnitude (based on convolution of
#              spectrum with some passband), a filter index, and
#              an exposure time, figure out how many electrons would
#              be knocked free by photons in the given exposure time.
#              Calculate an instrumental magnitude based on the
#              number of detected electrons.
#
#              In addition, calculate an uncertainty in the 
#              instrumental magnitude, including noise from the
#              star itself, background light, dark current, and 
#              read noise.
#
#                  input_dir         directory with spectral data
#                  output_dir        directory with actual filter data
#                  synth_mag         the synthetic mag for star in this filter
#                  fiducial_index    fiducial filter index
#                  exp_time          exposure time (seconds)
#                  instr_mag         (output) instrumental mag
#                  instr_mag_sig     (output) uncertainty in instr_mag
#
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc calc_instrumental_magnitude { input_dir output_dir \
                             synth_mag fiducial_index exp_time \
                             o_instr_mag o_instr_mag_sig } {
  global debug 
  upvar $o_instr_mag instr_mag $o_instr_mag_sig instr_mag_sig

  if { $debug > 0 } {
    puts "entering calc_instrumental_magnitude "
  }

  # for now, we assume a bunch of stuff -- make some of this 
  #    come from arguments or parameter file eventually
  set qe 1.0
  set num_pix 50
  set read_noise 5
  set dark_rate 1
  set num_zero 800000
  
 
  # how many electrons should we get from the star?
  set num_star [expr $num_zero*pow(10.0, -0.4*$synth_mag)*$exp_time*$qe]

  # how many electrons should we get from the sky?

  # how many electrons should we get from dark current?

  # how many electrons should we get from readnoise

  # what is the overall signal?

  # what is the overall noise

  # what is the signal-to-noise ratio -- and hence uncertainty in magnitude?



  

  return 0
}



##############################################################################
# PROCEDURE: read_outcat_info
# 
# DESCRIPTION: Read information from a line in the stellar catalog.
#              We will parse it and pick out
#
#                  RA              Right Ascension in decimal degrees (J2000)
#                  Dec             Declination in decimal degrees (J2000)
#                  sptype          guess at a spectral type. the format
#                                          is like "a0v" or "f6iii".
#                  zp              the zero point of this star,
#                                          in fundamental fiducial filter
#                  maglist         a list of the magnitudes of this star
#                                          through each fiducial filter;
#                                          sorted in order of fiducial
#                                          filter index
#
#              These six values are placed into the output args.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc read_outcat_info { line o_ra o_dec o_sptype o_zp o_maglist } {
  global debug 
  upvar $o_ra ra $o_dec dec $o_sptype sptype
  upvar $o_zp zp $o_maglist maglist

  # we expect a line like this:
  #
  #       ra dec sptype zp fnum mag fnum mag ... 
  #
  # where there is one "fnum mag" pair of values for each fiducial
  # filter
  #
  set line_length [llength $line]
  if { $line_length < 6 } {
    puts "read_outcat_info: fewer than 6 words in line ..$line.."
    return 1
  }
  set ra [lindex $line 0]
  set dec [lindex $line 1]
  set sptype [lindex $line 2]
  set zp [lindex $line 3]

  # how many pairs of "index mag" are there?
  set num_pairs [expr ($line_length-4)/2]
  # create a list which contains the filter index values only ...
  set index_values [list [lindex $line 4]]
  for { set i 1 } { $i < $num_pairs } { incr i } {
    lappend index_values [lindex $line [expr 4+($i*2)]]
  }
  if { $debug > 4 } {
    puts "index values is ..$index_values.. "
  }
  # okay. now we pick out the magnitudes, ordered by their filter index
  for { set i 0 } { $i < $num_pairs } { incr i } {
    set position_in_list [lsearch $index_values $i]
    set mag [lindex $line [expr 5+($position_in_list*2)]]
    if { $debug > 4 } {
      puts "   filter $i is at pos $position_in_list mag is $mag " 
    }
    if { $i == 0 } {
      set maglist [list $mag]
    } else {
      lappend maglist $mag
    }
  }
  
  if { $debug > 2 } {
    puts "about to return maglist ..$maglist.."
  }

  return 0
}




##############################################################################
# PROCEDURE: calc_exptime
# 
# DESCRIPTION: Given a base exposure time (seconds) and a "scatter"
#              in exposure time (also seconds), calculate the exposure
#              time for the next image.  
#
#              At the moment, we simply pick a random number
#              in the range [-scatter, +scatter] and add it to the
#              base exposure time.
#
#              Place the exposure time into the final argument.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc calc_exptime { base scatter o_exptime } {
  global debug 
  upvar $o_exptime exptime

  set exptime [expr $base + (2.0*$scatter*(0.5 - rand()))]

  if { $debug > 0 } {
    puts [format "calc_exptime: base %6.1f  scatter %5.2f  exptime %7.2f " \
                  $base $scatter $exptime]
  }

  return 0
}



##############################################################################
# PROCEDURE: calc_aperture_properties
# 
# DESCRIPTION: Given the radius of the aperture for synthetic photometry
#              (in units of arcseconds), calculate some properties
#              of the aperture for use in future calculations:
#
#                  aperture_npix            number of pixels inside aperture
#
#                  aperture_sqarcsec        area of aperture in square 
#                                                   arcseconds
#
#              Place these values into the output arguments.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc calc_aperture_properties { aperture_radius chip_info \
                                 o_aperture_npix o_aperture_sqarcsec } {
  global debug 
  upvar $o_aperture_npix aperture_npix 
  upvar $o_aperture_sqarcsec aperture_sqarcsec 

  if { $debug > 0 } {
    puts "entering calc_aperture_properties"
  }

  # the area of the aperture is easy: it's a circle
  set aperture_sqarcsec [expr 3.14159*$aperture_radius*$aperture_radius]

  # the number of pixels is harder, since we need to know the size
  #   of each pixel in units of arcseconds 
  # first, we get the size of each pixel in mm
  set pixsize [lindex $chip_info 3]

  # next, we calculate how many mm per arcsec on focal plane
  # this is the scale factor on focal plane, mm per radian
  set foc_list [get_foc_info]
  set scale_factor [lindex $foc_list 0]
  # convert to mm per arcsec
  set mm_per_arcsec [expr $scale_factor/206264.8]

  # using the pixel size (in mm), we can convert to pixels per arcsec
  set pix_per_arcsec [expr $mm_per_arcsec/$pixsize]
  # and now we can calculate the number of pixels within the aperture,
  #   since we know the area in square arcseconds
  set aperture_npix [expr $aperture_sqarcsec*($pix_per_arcsec*$pix_per_arcsec)]
 
  if { $debug > 2 } {
    puts [format "calc_aperture_properties: sqarcsec %8.3f npix %8.2f " \
                          $aperture_sqarcsec $aperture_npix]
  }
  

  return 0
}




##############################################################################
# PROCEDURE: write_image_info
# 
# DESCRIPTION: Given an already-opened fileid, plus some information
#              describing the properties of one snapshot, write
#              a single line in the form of keyed lists, like these:
#
#                   { id 3 } { exptime 10.20 }
#
#              into the given fileid.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc write_image_info { fileid id ra dec exptime } {
  global debug 

  puts $fileid [format "{ id %d } { ra %10.6f } { dec %10.6f } { exptime %8.4f } " \
                      $id $ra $dec $exptime ]

  return 0
}



##############################################################################
# PROCEDURE: read_chip_info
# 
# DESCRIPTION: Read information on the electronic properties of each
#              chip from a datafile, and store the properties in
#              an array for quick reference.  Place the number of chips
#              into the "num_chips" argument.
#         
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc read_chip_info { o_num_chips o_chip_info } {
  global debug 
  upvar $o_num_chips num_chips 
  upvar $o_chip_info chip_info 

  if { $debug > 0 } {
    puts "entering read_chip_info"
  }

  set chipfile_name [get_chipfile_name]
  if { $chipfile_name == {} } {
    puts stderr "read_chip_info: get_chipfile_name returns NULL"
    return 1
  }
  if { $debug > 1 } {
    puts "read_chip_info: chip filename is $chipfile_name"
  }
  set chip_id 0
  set done 0
  while { $done == 0 } {

    # read information about chip number $chip_id from file
    set name [format "chip_%03d" $chip_id]
    set ll [get_param $chipfile_name $name]
    if { $ll == {} } { 
      set done 1
    } else {
     set chip_info($chip_id) [lindex $ll 0]
     if { $debug > 1 } {
       set ci $chip_info($chip_id)
       puts "chip_info ( $chip_id ) is ..$ci.. "
     }
     incr chip_id
     
    }

  }
  set num_chips $chip_id
  if { $debug > 1 } {
    puts "read_chip_info: last good chip_id was $chip_id"
  }

  return 0
}





##############################################################################
# PROCEDURE: calc_magnitude_zeropoint
# 
# DESCRIPTION: Given a real flux-calibrated spectrum of Vega, and a 
#              fiducial (V-like) passband index, calculate the actual
#              number of photons per sq.cm. per sec from a star
#              of magnitude zero.  Then, use the spectrum of a vega-like 
#              star in our set of model spectra -- which are NOT 
#              flux-calibrated -- to calculate the number of "synthetic"
#              photons per sq.cm. per sec.  We calculate a difference
#              in magnitude from the ratio of these values.  By adding
#              that difference to every synthetic magnitude, we are
#              essentially flux-calibrating our synthetic spectra.
#         
#              Place the magnitude offset into the output argument.
#              We'll use it later to calculate the actual number of
#              photons we ought to detect from a star ...
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc calc_magnitude_zeropoint { input_dir vega_spectrum model_spectrum \
                                 vband_index convolve_program o_mag_zp  } {
  global debug 
  upvar $o_mag_zp mag_zp 

  if { $debug > 0 } {
    puts "entering calc_magnitude_zeropoint"
  }

  # sanity checks: create full file names and verify that the input
  #   spectrum datafiles exist
  set vega_filename [format "%s/%s" $input_dir $vega_spectrum]
  if { [file exists $vega_filename] != 1 } {
    puts stderr "calc_magnitude_zeropoint: vega file $vega_filename does not exist?"
    return 1
  }
  set model_filename [format "%s/%s" $input_dir $model_spectrum]
  if { [file exists $model_filename] != 1 } {
    puts stderr "calc_magnitude_zeropoint: model file $model_filename does not exist?"
    return 1
  }
  # and check that the fiducial passband/filter file exists
  set filter_filename [get_filter_fiducial_name $vband_index]
  if { [file exists $filter_filename] != 1 } {
    puts stderr "calc_magnitude_zeropoint: filter file $filter_filename does not exist?"
    return 1
  }
  # make sure that the convolving program exists
  if { [file exists $convolve_program] != 1 } {
    puts stderr "calc_magnitude_zeropoint: $convolve_program does not exist?"
    return 1
  }

  # okay, we are ready to go.  First, convolve Vega with filter
  set fiducial_angle 0
  set incident_angle 0
  set cmd [format "%s %s %s %lf %lf " $convolve_program $vega_filename \
                      $filter_filename $fiducial_angle $incident_angle]
  set retcode [catch { set result_line [eval exec $cmd] } ]
  if { $retcode != 0 } {
    puts stderr "calc_magnitude_zeropoint: error exec'ing ..$cmd.."
    return 1
  } 
  set vega_synth_mag [lindex $result_line 5]

  # Next, convolve our model vega-like spectrm with the same filter
  set cmd [format "%s %s %s %lf %lf " $convolve_program $model_filename \
                      $filter_filename $fiducial_angle $incident_angle]
  set retcode [catch { set result_line [eval exec $cmd] } ]
  if { $retcode != 0 } {
    puts stderr "calc_magnitude_zeropoint: error exec'ing ..$cmd.."
    return 1
  } 
  set model_synth_mag [lindex $result_line 5]

  # and now, find the difference in magnitudes.  If we add this value
  #   to all our synthetic photometry, we get the proper value for
  #   a star of mag zero.  We can then add the actual stellar mag
  #   to _that_ to get the proper synthetic magnitude.
  set mag_zp [expr $vega_synth_mag-$model_synth_mag]
  
  return 0
}




##############################################################################
# PROCEDURE: init_sky
# 
# DESCRIPTION: This routine runs once before each snapshot.  We
#              look through the catalog of stars for all objects
#              with a type of "sky".  We read in information about
#              each one, storing it in an array for quick access
#              later on.
#
#              If there are no sky entries in the catalog file, we
#              print a warning message and place "0" into the 
#              "sky_num" output argument
#
#              Args are:
#
#                 catalog_file         name of file with star and sky entries
#                 
#                 sky_num              (output) how many sky entries we
#                                               read from the file
#
#                 sky_info             (output) array of lists, one list per
#                                               sky entry with its info
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc init_sky { catalog_file o_sky_num o_sky_info} {
  global debug 
  upvar $o_sky_num sky_num $o_sky_info sky_info

  if { $debug > 0 } {
    puts "entering init_sky"
  }
 
  set sky_num 0

  # sanity check -- verify that the catalog file does exist
  if { [file exists $catalog_file] == 0 } {
    puts stderr "init_sky: catalog_file ..$catalog_file.. does not exist!?"
    return 1
  } 

  # open the reference catalog of stars
  set init_sky_fileid [open $catalog_file "r"]
  if { [catch { set init_sky_fileid } ] == 1 } {
    puts stderr "init_sky: error opening file $catalog_file -- quitting"
    return 1
  }

  while { [gets $init_sky_fileid line] > 0 } {
    # skip lines starting with the "#" character
    if { [regexp ^# $line] == 1 } {
      continue
    }
    if { $debug > 0 } {
      puts "init_sky: next line is ..$line.."
    }

    # Read information from the line: RA, Dec, spec_type, zeropoint
    #   and then expected magnitude in each of the filters.
    #   Ignore any line which doesn't look right ...
    if { [read_outcat_info $line ra dec sptype zp mag_list] != 0 } {
      puts stderr "init_sky: read_outcat_info fails on line ..$line.. "
      continue
    }
    # if this isn't a sky entry in the catalog (i.e. it is a star), 
    #   we ignore it
    if { [string compare $sptype "sky"] != 0 } {
      continue
    }

    set sky_info($sky_num) [list $ra $dec $sptype $zp $mag_list]
    if { $debug > 0 } {
      puts "init_sky: sky_info $sky_num is ..$sky_info($sky_num).. "
    }
    incr sky_num
    
  }
  close $init_sky_fileid

  # if there was no sky entry in the catalog, we print a warning
  #   but continue
  if { $sky_num == 0 } {
    puts "init_sky: no sky entry in $catalog_file ... "
  }
  return 0
}


##############################################################################
# PROCEDURE: pick_best_sky
# 
# DESCRIPTION: Given the RA and Dec position of a star, we want to 
#              find a sky value for it.  In theory, we'd pick the
#              sky entry closest to the star.  
#
#              In practice, we currently just pick the first sky 
#              entry there is.
#
#              We place a list of information about the sky into the 
#              final output arg:
#
#                  {  ra dec "sky" magzp  { mag0 mag1 ... magN }   }
#
#              If there is no "best" sky value, we place an empty list
#              into the output arg.
#
#              Args are:
#
#                 ra                   Right Ascension of star 
#                                                      (decimal degrees)
#
#                 dec                  Declination of star (decimal degrees)
#
#                 sky_num              total number of sky entries from which
#                                               we must pick the best
#
#                 i_sky_info           array of sky entries from which
#                                               we must pick the best; each
#                                               entry in array is a list
#                 
#                 o_best_sky           (output) the list with info about 
#                                               the best sky entry for this
#                                               (RA, Dec) position
#
# RETURNS:
#    0              if all goes well
#    1              if not
#
proc pick_best_sky { ra dec sky_num i_sky_info o_best_sky } {
  global debug 
  upvar $i_sky_info sky_info
  upvar $o_best_sky best_sky

  if { $debug > 0 } {
    puts "entering pick_best_sky"
  }

  # start by making the "best_sky" arg empty
  set best_sky {}
 
  # if there are no entries, we simply return now
  if { $sky_num < 1 } {
    if { $debug > 1 } {
      puts "pick_best_sky: no sky entries"
    }
    return 0
  } 

  # our current "best" sky entry is simply the first entry in the array
  #   (change later as needed)
  set best_sky $sky_info(0)
  if { $debug > 1 } {
     puts "pick_best_sky: best_sky is ..$best_sky.."
  }

  return 0
}


#########################################################################
# PROCEDURE: snapshot_check_graphics
# 
# DESCRIPTION: If the user has set the "snapshot_draw_stars" parameter
#              to 1, check to make sure that we will be able to draw
#              stars: we must be running under "wish" (not tclsh),
#              _and_ we must already have created a canvas.
#
#              If we won't be able to draw stars, then print a warning
#              message and re-set the "snapshot_draw_stars" variable
#              to zero.
#
# RETURNS:
#              0     if all goes well
#              1     if an error occurs
#
proc snapshot_check_graphics { o_snapshot_draw_stars } {
  global debug
  upvar $o_snapshot_draw_stars snapshot_draw_stars

  if { $debug > 0 } {
    puts "entering snapshot_check_graphics"
  }

  # if the user doesn't want to draw stars, then there's nothing to check.
  #    we are done
  if { $snapshot_draw_stars == 0 } {
    return 0
  }

  if { [is_graphics_enabled] == 0 } {
    puts stderr "snapshot_check_graphics: graphics not enabled"
    puts stderr "snapshot_check_graphics: resetting snapshot_draw_stars to zero"
    set snapshot_draw_stars 0
  }

  # Also check to make sure that we have already drawn a canvas of the
  #   focal plane (in the "draw_focal" step of the pipeline).  If we 
  #   have not, we can't draw the stars, so, again, print an error message,
  #   reset the "snapshot_draw_stars" variable to zero, but continue
  #   processing.
  if { [get_canvas_info c_name c_size c_magnify] != 0 } {
    puts stderr "snapshot_check_graphics: get_canvas_info fails "
    return 1
  }
  if { $c_name == {} } {
    puts stderr "snapshot_check_graphics: no canvas on which to draw"
    puts stderr "snapshot_check_graphics: resetting snapshot_draw_stars to zero"
    set snapshot_draw_stars 0
  }

  return 0
}




#########################################################################
# PROCEDURE: calc_intrachip_factor
# 
# DESCRIPTION: We want to calculate the factor by which a chip is
#              more or less sensitive at the position of a particular star.
#              This factor will always be <= 1.0.  
#
#              Our model involves a polynomial: 
#                   row0, col0          central (row, col) of variation
#                   a, b, c, d, e, f    coefficients of polynomial for RQE
#
#              First, we calculate a Reciprocal Quantum Efficiency (RQE)
#                  
#                   RQE  =  a + b*x + c*y + d*x*x + e*x*y + f*y*y
#
#              where    x = (row - row0)
#                       y = (col - col0)
#
#
#              Next, we calculate a relative sensitivity as
#
#                       sensitivity_factor =  1.0 / RQE
#
#              So values of the RQE must always be >= 1.0.
#
# RETURNS:
#              0     if all goes well
#              1     if an error occurs
#
proc calc_intrachip_factor { chip_params row col o_sensitivity_factor } {
  global debug
  upvar $o_sensitivity_factor sensitivity_factor

  if { $debug > 0 } {
    puts "entering calc_intrachip_factor"
  }

  set row0 [lindex $chip_params 0]
  set col0 [lindex $chip_params 1]
  set a [lindex $chip_params 2]
  set b [lindex $chip_params 3]
  set c [lindex $chip_params 4]
  set d [lindex $chip_params 5]
  set e [lindex $chip_params 6]
  set f [lindex $chip_params 7]

  set x [expr $row - $row0]
  set y [expr $col - $col0]
  if { $debug > 5 } {
    puts "calc_intrachip_factor: x is $x  y is $y "
    puts "calc_intrachip_factor: coeffs $a $b $c $d $e $f "
  }
  set rqe [expr $a + $b*$x + $c*$y + $d*$x*$x + $e*$x*$y + $f*$y*$y]
  set sensitivity_factor [expr 1.0/$rqe]

  return 0
}



namespace eval Snapshot {
  namespace export init_snapshot get_snapshot_seq_info \
                       init_random_deviates get_random_deviate
  variable Snapshot_Start_RA {}
  variable Snapshot_Start_Dec {}
  variable Snapshot_Snaps_Per_Step {}
  variable Snapshot_Step_RA {}
  variable Snapshot_Step_Dec {}
  variable Snapshot_Num_Steps {}
  variable Snapshot_Exp_Base {}
  variable Snapshot_Exp_Sig {}
  variable Snapshot_Num_Deviates {}
  variable Snapshot_Deviate_List {}
  variable Snapshot_Deviate_Index {}


#######################################################################
# Initialize values which control the number of and location of the
#   snapshots we take.
#
proc init_snapshot { param_file } {
  global debug
  variable Snapshot_Start_RA 
  variable Snapshot_Start_Dec 
  variable Snapshot_Snaps_Per_Step 
  variable Snapshot_Step_RA 
  variable Snapshot_Step_Dec 
  variable Snapshot_Num_Steps 
  variable Snapshot_Exp_Base
  variable Snapshot_Exp_Sig

  if { $debug > 0 } {
    puts "entering init_snapshot "
  }
  set Snapshot_Start_RA       [get_param $param_file "snapshot_start_ra"]
  set Snapshot_Start_Dec      [get_param $param_file "snapshot_start_dec"]
  set Snapshot_Snaps_Per_Step [get_param $param_file "snapshot_snaps_per_step"]
  set Snapshot_Step_RA        [get_param $param_file "snapshot_step_ra"]
  set Snapshot_Step_Dec       [get_param $param_file "snapshot_step_dec"]
  set Snapshot_Num_Steps      [get_param $param_file "snapshot_num_steps"]
  set Snapshot_Exp_Base       [get_param $param_file "snapshot_exp_base"]
  set Snapshot_Exp_Sig        [get_param $param_file "snapshot_exp_sig"]

  return 0
}

######################################################################
# Return a list of information describing the sequence of pictures to
#   be taken
#
#     { start_ra start_dec  snaps_per_step  step_ra step_dec  num_steps 
#                        exp_base exp_sig }
#
proc get_snapshot_seq_info { } {
  global debug
  variable Snapshot_Start_RA 
  variable Snapshot_Start_Dec 
  variable Snapshot_Snaps_Per_Step 
  variable Snapshot_Step_RA 
  variable Snapshot_Step_Dec 
  variable Snapshot_Num_Steps 
  variable Snapshot_Exp_Base 
  variable Snapshot_Exp_Sig

  if { $debug > 0 } {
    puts "entering get_snapshot_seq_info"
  }
  return [list $Snapshot_Start_RA $Snapshot_Start_Dec \
                    $Snapshot_Snaps_Per_Step $Snapshot_Step_RA  \
                    $Snapshot_Step_Dec $Snapshot_Num_Steps  \
                    $Snapshot_Exp_Base $Snapshot_Exp_Sig ]
}


#######################################################################
# Initialize a list of numbers which look like a set of random Gaussian
#   deviatess, normalized to have mean 0.0 and variance 1.0.  We use
#   a method based on the Central Limit Theorem (thanks to Matt Donadio
#   for his clear explanation).  We can then, later on, pick values
#   from this list very quickly and scale them to generate random-ish
#   noise.
#
#   Sets the variables 
#
#         Snapshot_Num_Deviates            number of values in array
#         Snapshot_Deviate_List            a list of randomish values
#         Snapshot_Deviate_Index           use this to pick items out
#                                               of the list
#
# The user can call "get_random_deviate" to pick one at random
#
# RETURNS:
#     0      if all goes well
#     1      otherwise
#
proc init_random_deviates { } {
  global debug
  variable Snapshot_Num_Deviates
  variable Snapshot_Deviate_List
  variable Snapshot_Deviate_Index

  if { $debug > 0 } {
    puts "entering init_random_deviates "
  }

  # we'll generate this many values, then cycle through them as needed
  set Snapshot_Num_Deviates  200
  set Snapshot_Deviate_Index 0

  # we generate the random deviates here using central limit theorem
  set n 500
  for { set iter 0 } { $iter < $Snapshot_Num_Deviates } { incr iter } {
    set x 0
    for { set i 0 } { $i < $n } { incr i } {
      set x [expr $x+rand()]
    }
    # move mean value to zero
    set x [expr $x-($n/2.0)]
    # normalize variance to 1.0
    set x [expr $x*sqrt(12.0/$n)]

    if { $debug > 5 } {
      puts [format "init_random_deviates: array elem %6d is %8.4f "  \
                   $iter $x]
    }

    lappend Snapshot_Deviate_List $x
  }
  
  return 0
}



#######################################################################
# Pick one element from the list of random gaussian deviates.  
#   Return it.
#
# Uses the variable Snapshot_Deviate_Index to walk through the
#   list "Snapshot_Deviate_List" and then return to the beginning
#   when necessary ... so actually these aren't really random
#   values.  We could pick a random element of the list, I suppose ... 
#
# RETURNS:
#     the random deviate value, with mean 0.0 and variance 1.0
#
proc get_random_deviate { } {
  global debug
  variable Snapshot_Num_Deviates
  variable Snapshot_Deviate_List
  variable Snapshot_Deviate_Index

  if { $debug > 0 } {
    puts "entering get_random_deviate "
  }

  # sanity check to make sure init_random_deviates has already been called
  if { [catch { if { $Snapshot_Deviate_Index >= $Snapshot_Num_Deviates } {} } ] != 0 } {
    puts stderr "get_random_deviates: didn't call init_random_deviates ?!"
    puts stderr "get_random_deviates: quitting!"
    exit
  }

  if { $Snapshot_Deviate_Index >= $Snapshot_Num_Deviates } {
    set Snapshot_Deviate_Index 0
  }
  set x [lindex $Snapshot_Deviate_List $Snapshot_Deviate_Index]
  incr Snapshot_Deviate_Index
 
  return $x
}



  # end of namespace "Snapshot"
} 


