
	/*
	 * Given a N >= 2 datafiles with some values (transmission, or
	 *   reflectivity, or QE) as a function of wavelength,
	 *
	 *         wavelength (Angstroms)      value
	 *         wavelength (Angstroms)      value
	 *         wavelength (Angstroms)      value
	 *
	 *   convolve them to produce a single output with the same format
	 *
	 *         wavelength (Angstroms)      value
	 *
	 * We use simple linear interpolation between all data points.
	 *
	 * By default, output goes to stdout, but the user can redirect it into
	 * a file by giving the "-o outfile" option.
	 * We skip all lines at the beginning and the end of the wavelength
	 * range which have values of zero; that is, we print out only
	 * the non-zero range.
	 *
	 *
	 * Usage:
	 *          spec_convolve  file1 file2 ...  [-o outfile] 
	 *
	 * MWR 5/10/2004
	 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#undef DEBUG
#undef DEBUG2

#define BUFLEN  256
#define COMMENT_CHAR  '#'


/* width of bins used for numerical integration (Angstroms) */
#define BIN_WIDTH 1.0

/* max allowed data points in a single file */
#define MAXPTS  100000

/* max allowed number of input data files */
#define MAXFILES  10


int    g_num_files;
int    g_num_data[MAXFILES];
double g_data_lam[MAXFILES][MAXPTS], g_data_val[MAXFILES][MAXPTS];
int    g_num_output;
double g_output_lam[MAXPTS], g_output_val[MAXPTS];


/* subroutines */
static int read_data(int file_index, char *infile);
static double get_interp_value(int file_index, double lambda);
static int    do_integration(void);
static int    print_output(FILE *out_fp);



	/* 
	 * Usage: spec_convolve  file1 file2 ... [-o outfile] 
	 */


int 
main
	(
	int argc,
	char *argv[]
	)
{
	int i;
	char infile[MAXFILES][BUFLEN];
	char time_stringval[BUFLEN];
	FILE *out_fp;
	time_t time_intval;
#ifdef DEBUG
	double lambda;
	double flux;
#endif


	if (argc < 3) {
		fprintf(stderr, "usage: spec_convolve file1 file2 ... [-o outfile] \n");
		exit(1);
	}
#ifdef DEBUG
	{
	int i;
	for (i = 0; i < argc; i++) {
		printf("  arg %2d is ..%s.. \n", i, argv[i]);
	}
	}
#endif

   /* parse args, figure out if we have an output file or will use stdout */
	if (strncmp("-o", argv[argc - 2], 2) == 0) {
		/* 
		 * yes, the penultimate arg is a flag "-o", and ultimate arg
		 * is therefore a file name for the output
		 */
		if ((out_fp = fopen(argv[argc - 1], "w")) == NULL) {
			fprintf(stderr, "spec_convolve: fail to fopen file %s for writing\n",
					argv[argc - 1]);
			exit(1);
		}
		g_num_files = argc - 3;
	}
	else {
		/* 
		 * nope, no output file name given.  We'll use stdout instead.
		 * All the args are input file names.
		 */
		out_fp = stdout;
		g_num_files = argc - 1;
	}

	if (g_num_files >= MAXFILES) {
		fprintf(stderr, "spec_convolve: %d is too many input files, max %d \n",
				g_num_files, MAXFILES - 1);
		exit(1);
	}


	/*
	 * we are ready to read in the data from each input file, 
	 *   storing it in a pair of global data arrays for future usage.
	 */
	for (i = 0; i < g_num_files; i++) {
		strncpy(infile[i], argv[i + 1], BUFLEN);
		infile[i][BUFLEN - 1] = '\0';
		if (read_data(i, infile[i]) != 0) {
			fprintf(stderr, "spec_convolve: read_data fails for file %s \n",
					infile[i]);
			exit(1);
		}
	}


#ifdef DEBUG2
   {
		int index;
		double lambda, value;

		printf(" gimme lambda: ");
		scanf("%lf", &lambda);
		for (i = 0; i < g_num_files; i++) {
			value = get_interp_value(i, lambda);
			printf("  file %d = %9.4f \n", i, value);
		}
	}
#endif

	/* 
	 * print a couple of lines with COMMENT_CHAR in first position,
	 * describing the provenance of the coming data ...
	 */
	fprintf(out_fp, "%c ", COMMENT_CHAR);
	for (i = 0; i < argc; i++) {
		fprintf(out_fp, "%s ", argv[i]);
	}
	fprintf(out_fp, "\n");
	/* the time we created this file  */
	time_intval = time((time_t *) NULL);
	strncpy(time_stringval, ctime((const time_t *) &time_intval), BUFLEN - 1);
	fprintf(out_fp, "%c %s", COMMENT_CHAR, time_stringval);


	/* okay, we're ready for the integration */
	if (do_integration() != 0) {
		fprintf(stderr, "spec_convolve: do_integration returns with error\n");
		exit(1);
	}

	/* print the results */
	if (print_output(out_fp) != 0) {
		fprintf(stderr, "spec_convolve: print_output returns with error\n");
		exit(1);
	}

	return(0);
}



/***************************************************************************
 * PROCEDURE: read_data
 *
 * DESCRIPTION: reads the data about a passband from the given file.
 *              The file is assumed to contain two columns of data:
 *
 *                  lambda    value
 *
 *              where
 *                    lambda       wavelength (in Angstroms)
 *
 *                    value        could be transmission, or QE, or ...
 *
 *              The values are placed into the global array
 *              g_data_val[][] and g_data_lam[][].  The number
 *              of data values is placed into g_ndata[] variable.
 * RETURNS:
 *    0        if all goes well
 *    1        if an error occurs
 */

static int
read_data
	(
	int file_index,             /* I: index of data file in list of all files */
	char *filename              /* I: name of file */
	)
{
	int i, ndata;
	char line[BUFLEN+1];
	FILE *fp;

	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "read_data: can't open file %s\n", filename);
		return(1);
	}

	/* 
	 * go line-by-line through the data file.
	 * Skip lines starting with "#", or blank lines.
	 */
	ndata = 0;
	while (fgets(line, BUFLEN, fp) != NULL) {

		/* check to see if we've run out of space. If so, quit with error */
		if (ndata >= MAXPTS) {
			fprintf(stderr, "read_data: too many data points\n");
			return(1);
		}

		/* skip lines starting with comment char */
		if (line[0] == COMMENT_CHAR) {
			continue;
		}

		/* skip blank lines, too */
		for (i = 0; isblank(line[i]); i++) {
			;
		}
		if (line[i] == '\n') {
			continue;
		}

#ifdef DEBUG
		printf("read_data: next good line is ..\n%s", line);
#endif

		if (sscanf(line, "%lf %lf", &(g_data_lam[file_index][ndata]),
				&(g_data_val[file_index][ndata])) != 2) {
			fprintf(stderr, "read_data: bad data line follows; quitting\n");
			fprintf(stderr, "%s\n", line);
			return(1);
		}

#ifdef DEBUG
		printf(" line %4d  lam %8.2f  data %8.4f \n", ndata, 
				g_data_lam[file_index][ndata], g_data_val[file_index][ndata]);
#endif

		ndata++;
	}

	fclose(fp);
	g_num_data[file_index] = ndata;



	return(0);
}



/*****************************************************************************
 * PROCEDURE: get_interp_value
 *
 * DESCRIPTION: Given a wavelength and a datafile index, return the value
 *              of the given data at the given wavelength.
 *              Use simple linear interpolation between the two nearest
 *              wavelengths.
 *
 *              If the given wavelength is outside the range of wavelengths
 *              with known transmission values, return zero.
 *
 * RETURNS:
 *    interpolated value (or zero)
 *
 */

static double
get_interp_value
	(
	int file_index,           /* I: index of input file */
	double lambda             /* I: wavelength of interest */
	)
{
	int i;
	double low_lam, high_lam;
	double low_trans, high_trans;
	double frac;
	double trans;

	/* check for out of bounds */
	if (lambda <= g_data_lam[file_index][0]) {
		return(0);
	}
	if (lambda >= g_data_lam[file_index][g_num_data[file_index]-1]) {
		return(0);
	}

	for (i = 1; i < g_num_data[file_index]; i++) {
		if (g_data_lam[file_index][i] > lambda) {

			/* interpolate linearly between "i" and "i-1" */
			low_lam = g_data_lam[file_index][i - 1];
			high_lam = g_data_lam[file_index][i];
			/* sanity check */
			if (high_lam <= low_lam) {
				fprintf(stderr, "get_interp_value: bad lambdas\n");
				fprintf(stderr, "  %5d %8.2f  %5d %8.2f \n", i-1, low_lam, i, high_lam);
				exit(1);
			}
						
			low_trans = g_data_val[file_index][i - 1];
			high_trans = g_data_val[file_index][i];

			frac = (lambda - low_lam)/(high_lam - low_lam);
			trans = low_trans + frac*(high_trans - low_trans);

			return(trans);

		}
	}

	/* should never get here! */
	fprintf(stderr, "get_interp_value: dropped off end?!\n");
	exit(1);


}


/***************************************************************************
 * PROCEDURE: do_integration
 *
 * DESCRIPTION: performs the convolution/integration over the passband.
 *              Calculates the number of photons per cm^2 per second.
 *              Prints out the number of photons,
 *              plus a "magnitude" based upon it with an arbitrary
 *              zero point.
 *
 * RETURNS:
 *   0        if all goes well
 *   1        if an error occurs
 */

static int
do_integration
	(
	void
	)
{
	int i;
	double start_lam, end_lam, lam, delta_lam;
	double x, product;

	/* 
	 * figure out the limits of integration --- based on the extrema
	 *   of the files.  We go from the shortest wavelength in any file
	 *   to the longest wavelength in any file.
	 */
	start_lam = g_data_lam[0][0];
	end_lam = g_data_lam[0][0];
	for (i = 0; i < g_num_files; i++) {
		if (g_data_lam[i][0] < start_lam) {
			start_lam = g_data_lam[i][0];
		}
		if (g_data_lam[i][g_num_data[i]-1] > end_lam) {
			end_lam = g_data_lam[i][g_num_data[i]-1];
		}
	}
#ifdef DEBUG
	printf("do_integration: integration limits %9.2f %9.2f \n", 
			start_lam, end_lam);
#endif

	/*
	 * now, we walk from the start to the end in steps of size BIN_WIDTH
	 *   Angstroms.  At each step, we 
	 *     - find the interpolated value of each file
	 *     - multiply together all these interpolated values
	 *     - print a line with the wavelength and this product
	 */
	g_num_output = 0;
	delta_lam = BIN_WIDTH;
	for (lam = start_lam; lam <= end_lam; lam += delta_lam) {
		product = 1.0;
		for (i = 0; i < g_num_files; i++) {
			x = get_interp_value(i, lam);
#ifdef DEBUG2
			printf("  file %d has val %10.4f  so prod %10.4f -> %10.4f \n",
					i, x, product, product*x);
#endif
			product *= x;
		}
		g_output_lam[g_num_output] = lam;
		g_output_val[g_num_output] = product;
		g_num_output++;
	}

	return(0);
}



/***************************************************************************
 * PROCEDURE: print_output
 *
 * DESCRIPTION: prints the results to the given FILE pointer.
 *              We truncate the range of wavlengths to the 
 *              interesting values (skip all zeroes at start and
 *              all zeroes at end).
 *
 * RETURNS:
 *   0        if all goes well
 *   1        if an error occurs
 */

static int
print_output
	(
	FILE *out_fp        /* I: send output here */
	)
{
	int i, start_i, end_i;

	/* 
	 * find the interesting limits in wavelength: the point at which
	 *   data is first non-zero, and the point at which data is last
	 *   non-zero.  We'll only print output within this range
	 *   of wavelengths.
	 */
	start_i = 0;
	/* first, scan for the starting point */
	for (i = 0; i < g_num_output; i++) {
		if (g_output_val[i] == 0) {
			if (((i + 1) < g_num_output) && (g_output_val[i + 1] == 0)) {
				start_i = i;
			}
			else {
				break;
			}
		}
	}
	/* now, scan for the ending point */
	end_i = g_num_output - 1;
	for (i = g_num_output - 1; i > 0; i--) {
		if (g_output_val[i] == 0) {
			if (((i - 1) > 0) && (g_output_val[i - 1] == 0)) {
				end_i = i;
			}
			else {
				break;
			}
		}
	}
#ifdef DEBUG
	printf(" starting index %d = %10.4f   end index %d = %10.4f \n", 
			start_i, g_output_lam[start_i], end_i, g_output_lam[end_i]);
#endif

	/* okay, print out the interesting data, one line per wavelength */
	for (i = start_i; i <= end_i; i++) {
		fprintf(out_fp, "%10.4f %10.4f \n", g_output_lam[i], g_output_val[i]);
	}

	return(0);
}



