#
# This TCL script works with the telescope information in the 
#   SNAP pipeline
#
# MWR 8/19/2003  
#    


if { [catch { set debug } ] == 1 } {
  set debug 2
}

source param.tcl


###########################################################################
# PROCEDURE: get_telescope_qe
#
# DESCRIPTION: Given an (x,y) position on the focal plane (in mm), 
#              return a value of the telescope's relative sensitivity
#              at that location.  All values must range between 0.0 and 1.0.
#
#              We could have arbitrarily complicated code here to account
#              for vignetting, blocking by particular optical elements,
#              imperfect coatings, etc.
#
#              For now (8/19/2003), simple polynomial as func of x,y.
#
#              Place the QE in the final output arg.
#
# RETURNS
#    0           if all goes well
#    1           if there's a problem
#    
proc get_telescope_qe { x y o_qe } {
  global debug
  upvar $o_qe qe

  if { $debug > 0 } {
    puts "entering get_telescope_qe"
  }

  if { [telescope_coeffs coeffs] != 0 } {
    puts stderr "get_telescope_qe: can't get coeffs "
    return 1
  }

  # simple polynomial expression for Reciprocal QE
  #
  #     RQE =  a + b*x + c*y + d*x*x + e*x*y + f*y*y
  # 
  # and then
  #
  #     qe(x, y)   =    1.0 / RQE
  #

  set a [lindex $coeffs 0]
  set b [lindex $coeffs 1]
  set c [lindex $coeffs 2]
  set d [lindex $coeffs 3]
  set e [lindex $coeffs 4]
  set f [lindex $coeffs 5]
  set rqe [expr $a + $b*$x + $c*$y + $d*$x*$x + $e*$x*$y + $f*$y*$y]
 
  # sanity check here: a value of rqe=1 means perfect reflectivity, etc.,
  #   and rqe = infinity means zero reflectivity, etc.
  #   We should never have rqe < 1
  if { $rqe < 1.0 } {
    puts stderr "get_telescope_qe: rqe $rqe is less than 1?!"
    return 1
  } else {
    set qe [expr 1.0/$rqe]
  }

  return 0
}





namespace eval Telescope {
  namespace export init_telescope telescope_coeffs telescope_area_sqcm
  variable Telescope_Coeffs {}
  variable Telescope_Area_Sqcm {}


#######################################################################
# Grab the values for the telescope QE coefficients from the given
#   param file
#
# RETURNS
#    0         if all goes well
#    1         if an error occurs
#
proc init_telescope { param_file } {
  global debug
  variable Telescope_Coeffs
  variable Telescope_Area_Sqcm

  if { $debug > 0 } {
    puts "entering init_telescope "
  }
  
  if { [file exists $param_file] != 1 } {
    puts stderr "init_telescope: can't open param file $param_file"
    return 1
  }
  set telescope_coeffs [get_param $param_file "tel_rqe_coeffs"]
  set Telescope_Coeffs [lindex $telescope_coeffs 0]

  # figure out the collecting area of the telescope, in square centimeters
  set pri_diam [get_param $param_file "tel_primary_diam"]
  set sec_diam [get_param $param_file "tel_secondary_diam"]
  set area_square_meters \
           [expr 3.14159*0.25*(($pri_diam*$pri_diam)-($sec_diam*$sec_diam))]
  set area_square_cm [expr $area_square_meters*100.0*100.0]
  set Telescope_Area_Sqcm $area_square_cm
  

  return 0
}

#######################################################################
# Return a list of the telescope QE coefficients, in the form
#
#        { a b c d e f }
#
# RETURNS
#    0        if all goes well
#    1        if an error occurs
#
proc telescope_coeffs { o_coeffs } {
  global debug
  variable Telescope_Coeffs
  upvar $o_coeffs coeffs

  if { $debug > 0 } {
    puts "entering telescope_coeffs "
  }
  
  set coeffs $Telescope_Coeffs

  return 0
} 

#######################################################################
# Place into the argument the telescope's collecting area,
#    in square centimeters
#
# RETURNS
#    0        if all goes well
#    1        if an error occurs
#
proc telescope_area_sqcm { o_area } {
  global debug
  variable Telescope_Area_Sqcm
  upvar $o_area area

  if { $debug > 0 } {
    puts "entering telescope_area_sqcm "
  }
  
  set area $Telescope_Area_Sqcm

  return 0
} 


  # end of namespace Telescope
}
